#ifndef NUMERICALANALYSIS_LINEARALGEBRA_ARITHMETICOPERATIONS_H
#define NUMERICALANALYSIS_LINEARALGEBRA_ARITHMETICOPERATIONS_H

#include <mkl_cblas.h>
#include <mkl_trans.h>
#include <mkl_vml.h>

#include "NumericVector.h"
#include "NumericMatrix.h"
#include "SupportFunctions.h"


// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// Arithmetic operations for NumericVector

namespace NumericalAnalysis
{
	namespace LinearAlgebra
	{
		inline NumericVector sum(const double firstScalingMultiplier, const NumericVector& nva, const double secondScalingMultiplier, const NumericVector& nvb) noexcept
		{
			NumericVector result = nvb;
			cblas_daxpby(result.size(), firstScalingMultiplier, nva.begin(), 1, secondScalingMultiplier, result.begin(), 1);
			return result;
		}

		template <typename matrix_transform_type = numeric_matrix_identity_tag>
		inline NumericVector multiply(const double scalingMultiplier, const NumericMatrix& nm, const NumericVector& nv) noexcept
		{
			constexpr CBLAS_TRANSPOSE transposeType = translateNumericMatrixTransposeTagToInt<matrix_transform_type>();
			NumericVector result(nv.size());
			cblas_dgemv(CblasColMajor, transposeType, nm.columnSize(), nm.rowSize(), scalingMultiplier, nm.begin(), nm.columnSize(), nv.begin(), 1, 0.0, result.begin(), 1);
			return result;
		}

		inline NumericVector operator+(const NumericVector& nva, const NumericVector& nvb) noexcept
		{
			NumericVector result(nva.size());
			vmdAdd(result.size(), nva.begin(), nvb.begin(), result.begin(), VML_HA);
			return result;
		}

		inline NumericVector operator-(const NumericVector& nva, const NumericVector& nvb) noexcept
		{
			NumericVector result(nva.size());
			vmdSub(result.size(), nva.begin(), nvb.begin(), result.begin(), VML_HA);
			return result;
		}

		inline NumericVector operator*(const NumericVector& nva, const NumericVector& nvb) noexcept
		{
			NumericVector result(nva.size());
			vmdMul(result.size(), nva.begin(), nvb.begin(), result.begin(), VML_HA);
			return result;
		}

		inline NumericVector operator*(const NumericMatrix& nm, const NumericVector& nv) noexcept
		{
			return multiply(1.0, nm, nv);
		}

		inline NumericVector operator*(const double scalingMultiplier, const NumericVector& nv) noexcept
		{
			NumericVector result(nv.size());
			cblas_daxpy(result.size(), scalingMultiplier, nv.begin(), 1, result.begin(), 1);
			return result;
		}

		inline NumericVector operator*(const NumericVector& nv, const double scalingMultiplier) noexcept
		{
			return (scalingMultiplier * nv);
		}

		inline NumericVector operator/(const NumericVector& nv, const double scalingDivider)
		{
			if (std::isnormal(scalingDivider))
			{
				return ((1.0 / scalingDivider) * nv);
			}

			else
			{
				std::string errorMessage = "------------------------------------------------------\n\n";
				errorMessage += "Error   : Invalid argument\n";
				errorMessage += "Location: The function of \"NumericalAnalysis::LinearAlgebra::operator/\".\n";
				errorMessage += "Message : scalingDivider is not a normalized value.\n\n";
				errorMessage += "------------------------------------------------------\n\n";
				throw std::invalid_argument(errorMessage);
			}
		}

		inline NumericVector operator/(const NumericVector& nva, const NumericVector& nvb) noexcept
		{
			NumericVector result(nva.size());
			vmdDiv(result.size(), nva.begin(), nvb.begin(), result.begin(), VML_HA);
			return result;
		}
	}
}

namespace NumericalAnalysis
{
	namespace LinearAlgebra
	{
		inline void rotate(const double cosine, NumericVector& nva, const double sine, NumericVector& nvb)
		{
			cblas_drot(nva.size(), nva.begin(), 1, nvb.begin(), 1, cosine, sine);
		}

		inline double innerProduct(const NumericVector& nva, const NumericVector& nvb)
		{
			if (nva.size() == nvb.size())
			{
				return cblas_ddot(nva.size(), nva.begin(), 1, nvb.begin(), 1);
			}

			else
			{
				std::string errorMessage = "------------------------------------------------------\n\n";
				errorMessage += "Error   : Invalid argument\n";
				errorMessage += "Location: The function of FunctionalAnalysis::LinearAlgebra::innerProduct\n";
				errorMessage += "Message : The size of the two NumericVector is different.\n\n";
				errorMessage += "------------------------------------------------------\n\n";
				throw std::invalid_argument(errorMessage);
			}
		}

		inline NumericVector crossProduct(const NumericVector& nva, const NumericVector& nvb)
		{
			double a = nva[1] * nvb[2] - nva[2] * nvb[1];
			double b = nva[2] * nvb[0] - nva[0] * nvb[2];
			double c = nva[0] * nvb[1] - nva[1] * nvb[0];
			return NumericVector{ a,b,c };
		}
	}
}

// Arithmetic operations for NumericVector
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// Arithmetic operations for NumericMatrix

namespace NumericalAnalysis
{
	namespace LinearAlgebra
	{
		template <typename first_matrix_transform_type = numeric_matrix_identity_tag, typename second_matrix_transform_type = numeric_matrix_identity_tag>
		inline NumericMatrix sum(const double firstScalingMultiplier, const NumericMatrix& nma, const double secondScalingMultiplier, const NumericMatrix& nmb)
		{
			constexpr char firstTransposeType = translateNumericMatrixTransposeTagToChar<first_matrix_transform_type>();
			constexpr char secondTransposeType = translateNumericMatrixTransposeTagToChar<second_matrix_transform_type>();
			NumericMatrix::size_type columnSize = getColumnSize<first_matrix_transform_type>(nma);
			NumericMatrix::size_type rowSize = getRowSize<first_matrix_transform_type>(nma);

			NumericMatrix result(columnSize, rowSize);
			MKL_Domatadd('C', firstTransposeType, secondTransposeType, columnSize, rowSize, firstScalingMultiplier, nma.begin(), nma.columnSize(), secondScalingMultiplier, nmb.begin(), nmb.columnSize(), result.begin(), columnSize);

			return result;
		}

		template <typename first_matrix_transform_type = numeric_matrix_identity_tag, typename second_matrix_transform_type = numeric_matrix_identity_tag>
		inline NumericMatrix multiply(const double scalingMultiplier, const NumericMatrix& nma, const NumericMatrix& nmb)
		{
			constexpr CBLAS_TRANSPOSE firstTransposeType = translateNumericMatrixTransposeTagToInt<first_matrix_transform_type>();
			constexpr CBLAS_TRANSPOSE secondTransposeType = translateNumericMatrixTransposeTagToInt<second_matrix_transform_type>();
			NumericMatrix::size_type columnSize = getColumnSize<first_matrix_transform_type>(nma);
			NumericMatrix::size_type multiplingSize = getRowSize<first_matrix_transform_type>(nma);
			NumericMatrix::size_type rowSize = getRowSize<first_matrix_transform_type>(nmb);

			NumericMatrix result(columnSize, rowSize);
			cblas_dgemm(CblasColMajor, firstTransposeType, secondTransposeType, columnSize, rowSize, multiplingSize, scalingMultiplier, nma.begin(), nma.columnSize(), nmb.begin(), nmb.columnSize(), 0.0, result.begin(), columnSize);
			return result;
		}

		inline NumericMatrix operator+(const NumericMatrix& nma, const NumericMatrix& nmb)
		{
			return sum(1.0, nma, 1.0, nmb);
		}

		inline NumericMatrix operator-(const NumericMatrix& nma, const NumericMatrix& nmb)
		{
			return sum(1.0, nma, -1.0, nmb);
		}

		inline NumericMatrix operator*(const NumericMatrix& nma, const NumericMatrix& nmb)
		{
			return multiply(1.0, nma, nmb);
		}

		inline NumericMatrix operator*(const double scalingMultiplier, const NumericMatrix& nm)
		{
			constexpr char transposeType = translateNumericMatrixTransposeTagToChar<numeric_matrix_identity_tag>();
			NumericMatrix::size_type columnSize = getColumnSize<numeric_matrix_identity_tag>(nm);
			NumericMatrix::size_type rowSize = getRowSize<numeric_matrix_identity_tag>(nm);

			NumericMatrix result(columnSize, rowSize);
			MKL_Domatcopy('C', transposeType, columnSize, rowSize, scalingMultiplier, nm.begin(), nm.columnSize(), result.begin(), columnSize);
			return result;
		}

		inline NumericMatrix operator*(const NumericMatrix& nm, const double scalingMultiplier)
		{
			return (scalingMultiplier * nm);
		}

		inline NumericMatrix operator/(const NumericMatrix& nm, const double scalingDivider)
		{
			if (std::isnormal(scalingDivider))
			{
				return ((1.0 / scalingDivider) * nm);
			}

			else
			{
				std::string errorMessage = "------------------------------------------------------\n\n";
				errorMessage += "Error   : Invalid argument\n";
				errorMessage += "Location: The function of \"NumericalAnalysis::LinearAlgebra::operator/\".\n";
				errorMessage += "Message : scalingDivider is not a normalized value.\n\n";
				errorMessage += "------------------------------------------------------\n\n";
				throw std::invalid_argument(errorMessage);
			}
		}
	}
}

// Arithmetic operations for NumericMatrix
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// Arithmetic operations for NumericVector




#endif // !NUMERICALANALYSIS_LINEARALGEBRA_ARITHMETICOPERATIONS_H
