#ifndef SOLIDSTATEPHYSICS_CRYSTALLIZATION_COMPONENTS_CHEMICALBOND_H
#define SOLIDSTATEPHYSICS_CRYSTALLIZATION_COMPONENTS_CHEMICALBOND_H

#include <string>
#include <stdexcept>
#include <functional>

#include "MinMaxConstrainter.h"
#include "ConstraintOwners.h"


namespace SolidStatePhysics
{
	namespace Crystallization
	{
		namespace Components
		{
			class ChemicalBond :public SolidStatePhysics::Crystallization::Calculators::MinMaxConstrainter
			{
// **********************************************************************************************************************************************************************************************************************************************************************************************
			// Protected constructors, public destructor, and static creation method

			public:
				ChemicalBond(ChemicalBondOwner& realOwner, const ChemicalBondOwnerAdaptor& attachedOwner);
				ChemicalBond(ChemicalBondOwner& realOwner, ChemicalBondOwnerAdaptor&& attachedOwner);
				explicit ChemicalBond(const std::shared_ptr<ChemicalBond>& twinPtr);

				virtual ~ChemicalBond();

			// Protected constructors, public destructor, and static creation method
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
			// Static methods

				static void create(ChemicalBondOwner& realOwner, const ChemicalBondOwnerAdaptor& attachedOwner);
				static void erase(std::shared_ptr<ChemicalBond>&& chemicalBondPtr);

				static bool shouldCreate(const InteractingAtom& ownerAtom, const AtomicAdaptor& attachedAtom) noexcept;
				static bool shouldCreate(const std::shared_ptr<InteratomicConnector>&) noexcept;

				static double getBondingEnergy(const double ownerBondingAffinity, const double attachedBondingAffinity) noexcept;
				static double getBondingEnergy(const InteractingAtom& ownerAtom, const AtomicAdaptor& attachedAtom) noexcept;

			// Static methods
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
			// Public methods

				void renew() override;
				bool isFeasible(const double maxBondingLengthErrorRate) const noexcept;
				double bondingEnergy() const noexcept;

				ChemicalBondOwner& realOwner() const noexcept;

			// Public methods
// **********************************************************************************************************************************************************************************************************************************************************************************************

			private:
				std::reference_wrapper<ChemicalBondOwner> m_realOwner;
				ChemicalBondOwnerAdaptor m_attachedOwner;
				double m_bondingEnergy;

				static double m_maxBondingLengthExtensionRate;


			private:
				ChemicalBond() = delete;
				ChemicalBond(const ChemicalBond&) = delete;
				ChemicalBond(ChemicalBond&&) noexcept = delete;
				ChemicalBond& operator=(const ChemicalBond&) = delete;
				ChemicalBond& operator=(ChemicalBond&&) noexcept = delete;
			};
		}
	}
}

// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// Static methods

inline void SolidStatePhysics::Crystallization::Components::ChemicalBond::create(ChemicalBondOwner& realOwner, const ChemicalBondOwnerAdaptor& attachedOwner)
{
	InteratomicConnector::createInteratomicComponent(realOwner, attachedOwner);
}

inline void SolidStatePhysics::Crystallization::Components::ChemicalBond::erase(std::shared_ptr<ChemicalBond>&& chemicalBondPtr)
{
	InteratomicConnector::eraseInteratomicComponent(std::move(chemicalBondPtr));
}

// Static methods
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// Public methods

inline void SolidStatePhysics::Crystallization::Components::ChemicalBond::renew()
{
	const double minimumDistance = ownerAtom().minimumIonicRadius() + attachedAtom().entity().minimumIonicRadius();
	const double maximumDistance = ownerAtom().maximumIonicRadius() + attachedAtom().entity().maximumIonicRadius();
	renewMinMaxDistance(minimumDistance, maximumDistance);
}

inline bool SolidStatePhysics::Crystallization::Components::ChemicalBond::isFeasible(const double maxBondingLengthErrorRate) const noexcept
{
	double maxBondLength = (1.0 + maxBondingLengthErrorRate) * maximumDistance();

	if (maxBondLength < interatomicDistance())
		return false;
	
	else
	{
		double minBondLength = (1.0 - maxBondingLengthErrorRate) * minimumDistance();

		if (interatomicDistance() < minBondLength)
			return false;
		else
			return true;
	}
}

inline double SolidStatePhysics::Crystallization::Components::ChemicalBond::bondingEnergy() const noexcept
{
	return m_bondingEnergy;
}

inline SolidStatePhysics::Crystallization::Components::ChemicalBondOwner& SolidStatePhysics::Crystallization::Components::ChemicalBond::realOwner() const noexcept
{
	return m_realOwner.get();
}

// Public methods
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************

#endif // !SOLIDSTATEPHYSICS_CRYSTALLIZATION_COMPONENTS_CHEMICALBOND_H
