#ifndef SOLIDSTATEPHYSICS_MATERIALS_CRYSTALLATTICE_H
#define SOLIDSTATEPHYSICS_MATERIALS_CRYSTALLATTICE_H

#include <stdexcept>
#include <string>

#include "LinearAlgebra.h"
#include "MathematicalConstants.h"


namespace SolidStatePhysics
{
	namespace Materials
	{
		struct CrystalLatticeParameters final
		{
			struct Length
			{
				double a;
				double b;
				double c;
			};

			struct Angle
			{
				double alpha;
				double beta;
				double gamma;
			};

			Length length;
			Angle angle;
		};

		class CrystalLattice final
		{
			using NumericVector = NumericalAnalysis::LinearAlgebra::NumericVector;
			using NumericMatrix = NumericalAnalysis::LinearAlgebra::NumericMatrix;

		public:
			using size_type = std::size_t;

// **********************************************************************************************************************************************************************************************************************************************************************************************
		// Constructors, destructors, and operators

		public:
			CrystalLattice();
			explicit CrystalLattice(const NumericMatrix& primitiveLatticeVectors);
			explicit CrystalLattice(NumericMatrix&& primitiveLatticeVectors);
			~CrystalLattice() = default;

			CrystalLattice(const CrystalLattice&) = default;
			CrystalLattice(CrystalLattice&&) noexcept = default;
			CrystalLattice& operator=(const CrystalLattice&) = default;
			CrystalLattice& operator=(CrystalLattice&&) noexcept = default;

		// Constructors, destructors, and operators
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
		// Access methods

			const NumericMatrix& primitiveLatticeVectors() const noexcept;
			const NumericMatrix& inversePrimitiveLatticeVectors() const noexcept;
			NumericMatrix primitiveReciprocalVectors() const noexcept;

			void resetPrimitiveLatticeVectors(const NumericMatrix& primitiveLatticeVectors);
			void resetPrimitiveLatticeVectors(NumericMatrix&& primitiveLatticeVectors);

			const NumericMatrix& transformationTensor() const noexcept;
			void addTransformationTensor(const double scalingMulitiplier, const NumericMatrix& additiveTransformationTensor);
			void scaleTransformationTensor(const double scalingMulitiplier);
			void resetTransformationTensor(const NumericMatrix& transformationTensor = NumericMatrix{});

		// Access methods
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
		// Utility

			void transform();
			void transform(const NumericMatrix& transformationTensor);

			NumericVector toFractionalCoordinate(const NumericVector& cartesianCoordinate) const noexcept;
			NumericVector toCartesianCoordinate(const NumericVector& fractionalCoordinate) const noexcept;

			void normalize() noexcept;
			void clear() noexcept;

		// Utility
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
		// Crystal Information

			bool hasNormalCrystalLattice() const noexcept;
			bool hasTooLongLatticeVector(const double maxLengthRatio) const noexcept;
			double getPrimitiveCellVolume() const noexcept;

			CrystalLatticeParameters getCrystalLatticeParameters() const noexcept;

		// Crystal Information
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
		// Private methods

		private:
			bool hasPositiveVolume() const noexcept;
			bool hasNormalCrystalLatticeParameters() const noexcept;
			bool hasStandingCrystalLattice() const noexcept;

			void normalizeLatticeVector(const NumericVector& basisLatticeVector, NumericVector& optimizedLatticeVector) noexcept;

		// Crystal Information
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
		// Private methods for supports

			double getAngle(const NumericVector&, const NumericVector&) const noexcept;
			double getAngle(NumericVector&&, NumericVector&&) const noexcept;

		// Private methods for supports
// **********************************************************************************************************************************************************************************************************************************************************************************************

		private:
			NumericMatrix m_primitiveLatticeVectors;
			NumericMatrix m_inversePrimitiveLatticeVectors;
			NumericMatrix m_transformationTensor;
		};
	}
}

// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// Access methods

inline const NumericalAnalysis::LinearAlgebra::NumericMatrix& SolidStatePhysics::Materials::CrystalLattice::primitiveLatticeVectors() const noexcept
{
	return m_primitiveLatticeVectors;
}

inline const NumericalAnalysis::LinearAlgebra::NumericMatrix& SolidStatePhysics::Materials::CrystalLattice::inversePrimitiveLatticeVectors() const noexcept
{
	return m_inversePrimitiveLatticeVectors;
}

inline NumericalAnalysis::LinearAlgebra::NumericMatrix SolidStatePhysics::Materials::CrystalLattice::primitiveReciprocalVectors() const noexcept
{
	return ((2.0 * NumericalAnalysis::MathematicalConstants::getPi()) * m_inversePrimitiveLatticeVectors);
}

inline void SolidStatePhysics::Materials::CrystalLattice::resetPrimitiveLatticeVectors(const NumericMatrix& primitiveLatticeVectors)
{
	if (primitiveLatticeVectors.columnSize() == 3 && primitiveLatticeVectors.rowSize() == 3)
	{
		m_primitiveLatticeVectors = primitiveLatticeVectors;
		m_inversePrimitiveLatticeVectors = m_primitiveLatticeVectors.inverseMatrix();
		m_transformationTensor = 0.0;
	}

	else
	{
		std::string errorMessage = "------------------------------------------------------\n\n";
		errorMessage += "Error   : Invalid argument\n";
		errorMessage += "Location: The member function of \"resetPrimitiveLatticeVectors\" in the class of SolidStatePhysics::Materials::CrystalLattice.\n";
		errorMessage += "Message : The sizes of the column and/or row of \"primitiveLatticeVectors\" are not three.\n\n";
		errorMessage += "------------------------------------------------------\n\n";
		throw std::invalid_argument(errorMessage);
	}
}

inline void SolidStatePhysics::Materials::CrystalLattice::resetPrimitiveLatticeVectors(NumericMatrix&& primitiveLatticeVectors)
{
	if (primitiveLatticeVectors.columnSize() == 3 && primitiveLatticeVectors.rowSize() == 3)
	{
		m_primitiveLatticeVectors = std::move(primitiveLatticeVectors);
		m_inversePrimitiveLatticeVectors = m_primitiveLatticeVectors.inverseMatrix();
		m_transformationTensor = 0.0;
	}

	else
	{
		std::string errorMessage = "------------------------------------------------------\n\n";
		errorMessage += "Error   : Invalid argument\n";
		errorMessage += "Location: The member function of \"resetPrimitiveLatticeVectors\" in the class of SolidStatePhysics::Materials::CrystalLattice.\n";
		errorMessage += "Message : The sizes of the column and/or row of \"primitiveLatticeVectors\" are not three.\n\n";
		errorMessage += "------------------------------------------------------\n\n";
		throw std::invalid_argument(errorMessage);
	}
}

inline const NumericalAnalysis::LinearAlgebra::NumericMatrix& SolidStatePhysics::Materials::CrystalLattice::transformationTensor() const noexcept
{
	return m_transformationTensor;
}

inline void SolidStatePhysics::Materials::CrystalLattice::addTransformationTensor(const double scalingMulitiplier, const NumericMatrix& additiveTransformationTensor)
{
	m_transformationTensor.add(scalingMulitiplier, additiveTransformationTensor);
}

inline void SolidStatePhysics::Materials::CrystalLattice::scaleTransformationTensor(const double scalingMulitiplier)
{
	m_transformationTensor *= scalingMulitiplier;
}

inline void SolidStatePhysics::Materials::CrystalLattice::resetTransformationTensor(const NumericMatrix& transformationTensor)
{
	if (transformationTensor.columnSize() == 3 && transformationTensor.rowSize() == 3)
		m_transformationTensor = transformationTensor;

	else
	{
		if (transformationTensor.empty())
			m_transformationTensor = 0.0;

		else
		{
			std::string errorMessage = "------------------------------------------------------\n\n";
			errorMessage += "Error   : Invalid argument\n";
			errorMessage += "Location: The member function of \"resetTransformationTensor\" in the class of SolidStatePhysics::Materials::CrystalLattice.\n";
			errorMessage += "Message : The sizes of the column and/or row of \"transformationTensor\" are not three.\n\n";
			errorMessage += "------------------------------------------------------\n\n";
			throw std::invalid_argument(errorMessage);
		}
	}
}

// Access methods
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// Utility

inline void SolidStatePhysics::Materials::CrystalLattice::transform()
{
	if (m_transformationTensor.columnSize() == 3 && m_transformationTensor.rowSize() == 3)
	{
		m_primitiveLatticeVectors += m_transformationTensor;
		m_inversePrimitiveLatticeVectors = m_primitiveLatticeVectors.inverseMatrix();
		m_transformationTensor = 0.0;
	}

	else
	{
		std::string errorMessage = "------------------------------------------------------\n\n";
		errorMessage += "Error   : Logic error\n";
		errorMessage += "Location: The member function of \"transform\" in the class of SolidStatePhysics::Materials::CrystalLattice.\n";
		errorMessage += "Message : The sizes of the column and/or row of \"m_transformationTensor\" are not three.\n\n";
		errorMessage += "------------------------------------------------------\n\n";
		throw std::logic_error(errorMessage);
	}
}

inline void SolidStatePhysics::Materials::CrystalLattice::transform(const NumericMatrix& transformationTensor)
{
	if (transformationTensor.columnSize() == 3 && transformationTensor.rowSize() == 3)
	{
		m_primitiveLatticeVectors += transformationTensor;
		m_inversePrimitiveLatticeVectors = m_primitiveLatticeVectors.inverseMatrix();
	}

	else
	{
		std::string errorMessage = "------------------------------------------------------\n\n";
		errorMessage += "Error   : Invalid argument\n";
		errorMessage += "Location: The member function of \"transform\" in the class of SolidStatePhysics::Materials::CrystalLattice.\n";
		errorMessage += "Message : The sizes of the column and/or row of \"transformationTensor\" are not three.\n\n";
		errorMessage += "------------------------------------------------------\n\n";
		throw std::invalid_argument(errorMessage);
	}
}

inline NumericalAnalysis::LinearAlgebra::NumericVector SolidStatePhysics::Materials::CrystalLattice::toFractionalCoordinate(const NumericVector& cartesianCoordinate) const noexcept
{
	return (m_inversePrimitiveLatticeVectors * cartesianCoordinate);
}

inline NumericalAnalysis::LinearAlgebra::NumericVector SolidStatePhysics::Materials::CrystalLattice::toCartesianCoordinate(const NumericVector& fractionalCoordinate) const noexcept
{
	return (m_primitiveLatticeVectors * fractionalCoordinate);
}

inline void SolidStatePhysics::Materials::CrystalLattice::normalize() noexcept
{
	std::vector<NumericVector> latticeVectors{ m_primitiveLatticeVectors[0], m_primitiveLatticeVectors[1], m_primitiveLatticeVectors[2] };
	{
		normalizeLatticeVector(latticeVectors[0], latticeVectors[1]);
		normalizeLatticeVector(latticeVectors[1], latticeVectors[0]);

		normalizeLatticeVector(latticeVectors[1], latticeVectors[2]);
		normalizeLatticeVector(latticeVectors[2], latticeVectors[1]);

		normalizeLatticeVector(latticeVectors[2], latticeVectors[0]);
		normalizeLatticeVector(latticeVectors[0], latticeVectors[2]);
	}

	m_primitiveLatticeVectors = NumericMatrix{ latticeVectors[0], latticeVectors[1], latticeVectors[2] };
	m_inversePrimitiveLatticeVectors = m_primitiveLatticeVectors.inverseMatrix();
	m_transformationTensor = 0.0;
}

inline void SolidStatePhysics::Materials::CrystalLattice::clear() noexcept
{
	m_primitiveLatticeVectors = 0.0;
	m_primitiveLatticeVectors(0, 0) = 1.0;
	m_primitiveLatticeVectors(1, 1) = 1.0;
	m_primitiveLatticeVectors(2, 2) = 1.0;

	m_inversePrimitiveLatticeVectors = 0.0;
	m_inversePrimitiveLatticeVectors(0, 0) = 1.0;
	m_inversePrimitiveLatticeVectors(1, 1) = 1.0;
	m_inversePrimitiveLatticeVectors(2, 2) = 1.0;

	m_transformationTensor = 0.0;
}

// Utility
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// Crystal Information

inline bool SolidStatePhysics::Materials::CrystalLattice::hasNormalCrystalLattice() const noexcept
{
	if (hasPositiveVolume())
	{
		if (hasNormalCrystalLatticeParameters())
			return hasStandingCrystalLattice();
		else
			return false;
	}

	else
		return false;
}

inline double SolidStatePhysics::Materials::CrystalLattice::getPrimitiveCellVolume() const noexcept
{
	using NumericalAnalysis::LinearAlgebra::innerProduct;
	using NumericalAnalysis::LinearAlgebra::crossProduct;
	return innerProduct(m_primitiveLatticeVectors[0], crossProduct(m_primitiveLatticeVectors[1], m_primitiveLatticeVectors[2]));
}

inline SolidStatePhysics::Materials::CrystalLatticeParameters SolidStatePhysics::Materials::CrystalLattice::getCrystalLatticeParameters() const noexcept
{
	double alpha = getAngle(m_primitiveLatticeVectors[1], m_primitiveLatticeVectors[2]);
	double beta = getAngle(m_primitiveLatticeVectors[2], m_primitiveLatticeVectors[0]);
	double gamma = getAngle(m_primitiveLatticeVectors[0], m_primitiveLatticeVectors[1]);

	CrystalLatticeParameters::Length length{ m_primitiveLatticeVectors[0].norm(), m_primitiveLatticeVectors[1].norm(), m_primitiveLatticeVectors[2].norm() };
	CrystalLatticeParameters::Angle angle{ alpha, beta, gamma };
	return CrystalLatticeParameters{ length, angle };
}

// Crystal Information
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// Private methods

inline bool SolidStatePhysics::Materials::CrystalLattice::hasPositiveVolume() const noexcept
{
	if (0.0 < getPrimitiveCellVolume())
		return true;
	else
		return false;
}

inline bool SolidStatePhysics::Materials::CrystalLattice::hasNormalCrystalLatticeParameters() const noexcept
{
	constexpr double minAngle = 20.0;
	constexpr double maxAngle = 160.0;
	CrystalLatticeParameters crystalLatticeParameters = getCrystalLatticeParameters();

	if (crystalLatticeParameters.angle.alpha < minAngle || maxAngle < crystalLatticeParameters.angle.alpha)
		return false;
	if (crystalLatticeParameters.angle.beta < minAngle || maxAngle < crystalLatticeParameters.angle.beta)
		return false;
	if (crystalLatticeParameters.angle.gamma < minAngle || maxAngle < crystalLatticeParameters.angle.gamma)
		return false;

	return true;
}

inline bool SolidStatePhysics::Materials::CrystalLattice::hasStandingCrystalLattice() const noexcept
{
	constexpr double maxAngle = 70.0;
	using NumericalAnalysis::LinearAlgebra::crossProduct;

	if (70.0 < getAngle(crossProduct(m_primitiveLatticeVectors[0], m_primitiveLatticeVectors[1]), m_primitiveLatticeVectors[2]))
		return false;
	if (70.0 < getAngle(crossProduct(m_primitiveLatticeVectors[1], m_primitiveLatticeVectors[2]), m_primitiveLatticeVectors[0]))
		return false;
	if (70.0 < getAngle(crossProduct(m_primitiveLatticeVectors[2], m_primitiveLatticeVectors[0]), m_primitiveLatticeVectors[1]))
		return false;

	return true;
}

// Private methods
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// Private methods for supports

inline double SolidStatePhysics::Materials::CrystalLattice::getAngle(const NumericVector& vecA, const NumericVector& vecB) const noexcept
{
	constexpr double unitChangeFactor = 57.295779513082320876798154814105;

	double angle = 0.0;
	{
		double cosine = NumericalAnalysis::LinearAlgebra::innerProduct(vecA, vecB);
		cosine /= vecA.norm();
		cosine /= vecB.norm();

		angle = std::acos(cosine);
		angle *= unitChangeFactor;
	}

	return angle;
}

inline double SolidStatePhysics::Materials::CrystalLattice::getAngle(NumericVector&& vecA, NumericVector&& vecB) const noexcept
{
	constexpr double unitChangeFactor = 57.295779513082320876798154814105;

	double angle = 0.0;
	{
		double cosine = NumericalAnalysis::LinearAlgebra::innerProduct(vecA, vecB);
		cosine /= vecA.norm();
		cosine /= vecB.norm();

		angle = std::acos(cosine);
		angle *= unitChangeFactor;
	}

	return angle;
}

// Private methods for supports
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************

#endif // !SOLIDSTATEPHYSICS_MATERIALS_CRYSTALLATTICE_H
