#ifndef SOLIDSTATEPHYSICS_CRYSTALLIZATION_SIMULATORS_CRYSTALLIZATIONPLAN_H
#define SOLIDSTATEPHYSICS_CRYSTALLIZATION_SIMULATORS_CRYSTALLIZATIONPLAN_H

#include <stdexcept>
#include <string>
#include <set>
#include <utility>

#include "CrystallizationVariables.h"


namespace SolidStatePhysics
{
	namespace Crystallization
	{
		namespace Simulators
		{
			class CrystallizationPlan
			{
				using size_type = std::size_t;

// **********************************************************************************************************************************************************************************************************************************************************************************************
			// Constructors, destructor, and operators

			public:
				CrystallizationPlan();
				virtual ~CrystallizationPlan() = default;

				CrystallizationPlan(const CrystallizationPlan&) = default;
				CrystallizationPlan(CrystallizationPlan&&) noexcept = default;
				CrystallizationPlan& operator=(const CrystallizationPlan&) = default;
				CrystallizationPlan& operator=(CrystallizationPlan&&) noexcept = default;

			// Constructors, destructor, and operators
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
			// Access methods

				size_type maxGeometricalOptimizing() const noexcept;
				double maxBondingLengthErrorRate() const noexcept;
				double minPackingFraction() const noexcept;
				double pressure() const noexcept;
				double attractiveForceConstant() const noexcept;
				double repulsiveForceConstant() const noexcept;
				
				const std::set<std::pair<std::string, std::string>>& vertexSharingCationicPairs() const noexcept;
				const std::set<std::pair<std::string, std::string>>& edgeSharingCationicPairs() const noexcept;
				const std::set<std::pair<std::string, std::string>>& faceSharingCationicPairs() const noexcept;
				const CrystallizationVariables& globalOptimizationVariables() const noexcept;
				const CrystallizationVariables& localOptimizationVariables() const noexcept;

				void resetMaxBondingLengthErrorRate(const double maxBondingLengthErrorRate);
				void resetMinPackingFraction(const double minPackingFraction);
				void resetPressure(const double pressure);
				void resetAttractiveForceConstant(const double attractiveForceConstant);
				void resetRepulsiveForceConstant(const double repulsiveForceConstant);

				void resetVertexSharingCationicPairs(const std::set<std::pair<std::string, std::string>>&) noexcept;
				void resetVertexSharingCationicPairs(std::set<std::pair<std::string, std::string>>&&) noexcept;
				void resetEdgeSharingCationicPairs(const std::set<std::pair<std::string, std::string>>&) noexcept;
				void resetEdgeSharingCationicPairs(std::set<std::pair<std::string, std::string>>&&) noexcept;
				void resetFaceSharingCationicPairs(const std::set<std::pair<std::string, std::string>>&) noexcept;
				void resetFaceSharingCationicPairs(std::set<std::pair<std::string, std::string>>&&) noexcept;
				void resetGlobalOptimizationVariables(const CrystallizationVariables&) noexcept;
				void resetGlobalOptimizationVariables(CrystallizationVariables&&) noexcept;
				void resetLocalOptimizationVariables(const CrystallizationVariables&) noexcept;
				void resetLocalOptimizationVariables(CrystallizationVariables&&) noexcept;

			// Access methods
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
			// Utility

				void clear() noexcept;

			// Utility
// **********************************************************************************************************************************************************************************************************************************************************************************************

			private:
				size_type m_maxGeomatricalOptimizing;
				double m_maxBondingLengthErrorRate;
				double m_minPackingFraction;
				double m_pressure;
				double m_attractiveForceConstant;
				double m_repulsiveForceConstant;

				std::set<std::pair<std::string, std::string>> m_vertexSharingCationicPairs;
				std::set<std::pair<std::string, std::string>> m_edgeSharingCationicPairs;
				std::set<std::pair<std::string, std::string>> m_faceSharingCationicPairs;
				CrystallizationVariables m_globalOptimizationVariables;
				CrystallizationVariables m_localOptimizationVariables;
			};
		}
	}
}

// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// Access methods

inline SolidStatePhysics::Crystallization::Simulators::CrystallizationPlan::size_type SolidStatePhysics::Crystallization::Simulators::CrystallizationPlan::maxGeometricalOptimizing() const noexcept
{
	return m_maxGeomatricalOptimizing;
}

inline double SolidStatePhysics::Crystallization::Simulators::CrystallizationPlan::maxBondingLengthErrorRate() const noexcept
{
	return m_maxBondingLengthErrorRate;
}

inline double SolidStatePhysics::Crystallization::Simulators::CrystallizationPlan::minPackingFraction() const noexcept
{
	return m_minPackingFraction;
}

inline double SolidStatePhysics::Crystallization::Simulators::CrystallizationPlan::pressure() const noexcept
{
	return m_pressure;
}

inline double SolidStatePhysics::Crystallization::Simulators::CrystallizationPlan::attractiveForceConstant() const noexcept
{
	return m_attractiveForceConstant;
}

inline double SolidStatePhysics::Crystallization::Simulators::CrystallizationPlan::repulsiveForceConstant() const noexcept
{
	return m_repulsiveForceConstant;
}

inline const std::set<std::pair<std::string, std::string>>& SolidStatePhysics::Crystallization::Simulators::CrystallizationPlan::vertexSharingCationicPairs() const noexcept
{
	return m_vertexSharingCationicPairs;
}

inline const std::set<std::pair<std::string, std::string>>& SolidStatePhysics::Crystallization::Simulators::CrystallizationPlan::edgeSharingCationicPairs() const noexcept
{
	return m_edgeSharingCationicPairs;
}

inline const std::set<std::pair<std::string, std::string>>& SolidStatePhysics::Crystallization::Simulators::CrystallizationPlan::faceSharingCationicPairs() const noexcept
{
	return m_faceSharingCationicPairs;
}

inline const SolidStatePhysics::Crystallization::Simulators::CrystallizationVariables& SolidStatePhysics::Crystallization::Simulators::CrystallizationPlan::globalOptimizationVariables() const noexcept
{
	return m_globalOptimizationVariables;
}

inline const SolidStatePhysics::Crystallization::Simulators::CrystallizationVariables& SolidStatePhysics::Crystallization::Simulators::CrystallizationPlan::localOptimizationVariables() const noexcept
{
	return m_localOptimizationVariables;
}

inline void SolidStatePhysics::Crystallization::Simulators::CrystallizationPlan::resetMaxBondingLengthErrorRate(const double maxBondingLengthErrorRate)
{
	if (0.0 < maxBondingLengthErrorRate)
		m_maxBondingLengthErrorRate = maxBondingLengthErrorRate;

	else
	{
		std::string errorMessage = "------------------------------------------------------\n\n";
		errorMessage += "Error   : Invalid argument\n";
		errorMessage += "Location: The member function of \"resetMaxBondingLengthErrorRate\" in the class of SolidStatePhysics::Crystallization::Simulators::CrystallizationPlan.\n";
		errorMessage += "Message : \"maxBondingLengthErrorRate\" is not more than 0.0.\n\n";
		errorMessage += "------------------------------------------------------\n\n";
		throw std::invalid_argument(errorMessage);
	}
}

inline void SolidStatePhysics::Crystallization::Simulators::CrystallizationPlan::resetMinPackingFraction(const double minPackingFraction)
{
	if (0.0 < minPackingFraction)
	{
		if (minPackingFraction < 1.0)
			m_minPackingFraction = minPackingFraction;

		else
		{
			std::string errorMessage = "------------------------------------------------------\n\n";
			errorMessage += "Error   : Invalid argument\n";
			errorMessage += "Location: The member function of \"resetMinPackingFraction\" in the class of SolidStatePhysics::Crystallization::CrystallizationPlan.\n";
			errorMessage += "Message : \"minPackingFraction\" is not less than 1.0.\n\n";
			errorMessage += "------------------------------------------------------\n\n";
			throw std::invalid_argument(errorMessage);
		}
	}

	else
	{
		std::string errorMessage = "------------------------------------------------------\n\n";
		errorMessage += "Error   : Invalid argument\n";
		errorMessage += "Location: The member function of \"resetMinPackingFraction\" in the class of SolidStatePhysics::Crystallization::CrystallizationPlan.\n";
		errorMessage += "Message : \"minPackingFraction\" is not more than 0.0.\n\n";
		errorMessage += "------------------------------------------------------\n\n";
		throw std::invalid_argument(errorMessage);
	}
}

inline void SolidStatePhysics::Crystallization::Simulators::CrystallizationPlan::resetPressure(const double pressure)
{
	if (pressure < 0.0)
	{
		std::string errorMessage = "------------------------------------------------------\n\n";
		errorMessage += "Error   : Invalid argument\n";
		errorMessage += "Location: The function of \"resetPressure\" in the class of SolidStatePhysics::Crystallization::Simulators::CrystallizationPlan.\n";
		errorMessage += "Message : \"pressure\" is less than zero.\n\n";
		errorMessage += "------------------------------------------------------\n\n";
		throw std::invalid_argument(errorMessage);
	}

	else
		m_pressure = pressure;
}

inline void SolidStatePhysics::Crystallization::Simulators::CrystallizationPlan::resetAttractiveForceConstant(const double attractiveForceConstant)
{
	if (0.0 < attractiveForceConstant)
		m_attractiveForceConstant = attractiveForceConstant;

	else
	{
		std::string errorMessage = "------------------------------------------------------\n\n";
		errorMessage += "Error   : Invalid argument\n";
		errorMessage += "Location: The function of \"resetAttractiveForceConstant\" in the class of SolidStatePhysics::Crystallization::Simulators::CrystallizationPlan.\n";
		errorMessage += "Message : \"attractiveForceConstant\" is not more than zero.\n\n";
		errorMessage += "------------------------------------------------------\n\n";
		throw std::invalid_argument(errorMessage);
	}
}

inline void SolidStatePhysics::Crystallization::Simulators::CrystallizationPlan::resetRepulsiveForceConstant(const double repulsiveForceConstant)
{
	if (repulsiveForceConstant < 0.0)
		m_repulsiveForceConstant = repulsiveForceConstant;

	else
	{
		std::string errorMessage = "------------------------------------------------------\n\n";
		errorMessage += "Error   : Invalid argument\n";
		errorMessage += "Location: The function of \"resetRepulsiveForceConstant\" in the class of SolidStatePhysics::Crystallization::Simulators::CrystallizationPlan.\n";
		errorMessage += "Message : \"repulsiveForceConstant\" is not less than zero.\n\n";
		errorMessage += "------------------------------------------------------\n\n";
		throw std::invalid_argument(errorMessage);
	}
}

inline void SolidStatePhysics::Crystallization::Simulators::CrystallizationPlan::resetVertexSharingCationicPairs(const std::set<std::pair<std::string, std::string>>& vertexSharingCationicPairs) noexcept
{
	m_vertexSharingCationicPairs = vertexSharingCationicPairs;
}

inline void SolidStatePhysics::Crystallization::Simulators::CrystallizationPlan::resetVertexSharingCationicPairs(std::set<std::pair<std::string, std::string>>&& vertexSharingCationicPairs) noexcept
{
	m_vertexSharingCationicPairs = std::move(vertexSharingCationicPairs);
}

inline void SolidStatePhysics::Crystallization::Simulators::CrystallizationPlan::resetEdgeSharingCationicPairs(const std::set<std::pair<std::string, std::string>>& edgeSharingCationicPairs) noexcept
{
	m_edgeSharingCationicPairs = edgeSharingCationicPairs;
}

inline void SolidStatePhysics::Crystallization::Simulators::CrystallizationPlan::resetEdgeSharingCationicPairs(std::set<std::pair<std::string, std::string>>&& edgeSharingCationicPairs) noexcept
{
	m_edgeSharingCationicPairs = std::move(edgeSharingCationicPairs);
}

inline void SolidStatePhysics::Crystallization::Simulators::CrystallizationPlan::resetFaceSharingCationicPairs(const std::set<std::pair<std::string, std::string>>& faceSharingCationicPairs) noexcept
{
	m_faceSharingCationicPairs = faceSharingCationicPairs;
}

inline void SolidStatePhysics::Crystallization::Simulators::CrystallizationPlan::resetFaceSharingCationicPairs(std::set<std::pair<std::string, std::string>>&& faceSharingCationicPairs) noexcept
{
	m_faceSharingCationicPairs = std::move(faceSharingCationicPairs);
}

inline void SolidStatePhysics::Crystallization::Simulators::CrystallizationPlan::resetGlobalOptimizationVariables(const CrystallizationVariables& globalOptimizationVariables) noexcept
{
	m_globalOptimizationVariables = globalOptimizationVariables;
	m_maxGeomatricalOptimizing = m_globalOptimizationVariables.necessaryGeometricalOptimizing();
}

inline void SolidStatePhysics::Crystallization::Simulators::CrystallizationPlan::resetGlobalOptimizationVariables(CrystallizationVariables&& globalOptimizationVariables) noexcept
{
	m_globalOptimizationVariables = std::move(globalOptimizationVariables);
	m_maxGeomatricalOptimizing = m_globalOptimizationVariables.necessaryGeometricalOptimizing();
}

inline void SolidStatePhysics::Crystallization::Simulators::CrystallizationPlan::resetLocalOptimizationVariables(const CrystallizationVariables& localOptimizationVariables) noexcept
{
	m_localOptimizationVariables = localOptimizationVariables;
}

inline void SolidStatePhysics::Crystallization::Simulators::CrystallizationPlan::resetLocalOptimizationVariables(CrystallizationVariables&& localOptimizationVariables) noexcept
{
	m_localOptimizationVariables = std::move(localOptimizationVariables);
}

// Access methods
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************

#endif // !SOLIDSTATEPHYSICS_CRYSTALLIZATION_SIMULATORS_CRYSTALLIZATIONPLAN_H
