#ifndef SOLIDSTATEPHYSICS_CRYSTALLIZATION_SIMULATORS_CRYSTALLIZER_H
#define SOLIDSTATEPHYSICS_CRYSTALLIZATION_SIMULATORS_CRYSTALLIZER_H

#include <string>
#include <filesystem>
#include <stdexcept>
#include <mutex>

#include "IntelligentLiquid.h"

#include "CrystallizationPlan.h"
#include "MonitoringPlan.h"
#include "ChemicalBondingRecorder.h"


namespace SolidStatePhysics
{
	namespace Crystallization
	{
		namespace Simulators
		{
			class Crystallizer
			{
				using size_type = std::size_t;
				using IntelligentLiquid = SolidStatePhysics::Crystallization::Components::IntelligentLiquid;

// **********************************************************************************************************************************************************************************************************************************************************************************************
			// Constructors and destructor

			public:
				Crystallizer() = default;
				Crystallizer(const CrystallizationPlan&, const MonitoringPlan&);
				Crystallizer(CrystallizationPlan&&, MonitoringPlan&&);

				virtual ~Crystallizer() = default;

			// Constructors and destructor
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
			// Public methods

				void initialize(const std::filesystem::path& workDirectoryPath);
				void finalize() noexcept;
				void run(IntelligentLiquid&);

				void resetCrystallizationPlan(const CrystallizationPlan&) noexcept;
				void resetCrystallizationPlan(CrystallizationPlan&&) noexcept;
				void resetMonitoringPlan(const MonitoringPlan&) noexcept;
				void resetMonitoringPlan(MonitoringPlan&&) noexcept;

				size_type elapsedTime() const noexcept;
				size_type localOptimizationCount() const noexcept;
				size_type remodellingCount() const noexcept;

			// Public methods
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
			// Private methods

			private:
				void runGeometricalOptimizers(IntelligentLiquid&);

				void overCreateDatFile(const IntelligentLiquid&);
				void overCreateCifFile(const IntelligentLiquid&);
				void overCreateMdFile(const IntelligentLiquid&);

				void linkChemicalBondingRecorder();

			// Private methods
// **********************************************************************************************************************************************************************************************************************************************************************************************

			private:
				CrystallizationPlan m_crystallizationPlan;
				MonitoringPlan m_monitoringPlan;

				size_type m_elapsedTime;
				size_type m_localOptimizationCount;
				size_type m_remodellingCount;
				std::filesystem::path m_workDirectoryPath;
				ChemicalBondingRecorder m_chemicalBondingRecorder;

			private:
				Crystallizer(const Crystallizer&) = delete;
				Crystallizer(Crystallizer&&) noexcept = delete;
				Crystallizer& operator=(const Crystallizer&) = delete;
				Crystallizer& operator=(Crystallizer&&) noexcept = delete;
			};
		}
	}
}

// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// Public methods

inline void SolidStatePhysics::Crystallization::Simulators::Crystallizer::initialize(const std::filesystem::path& workDirectoryPath)
{
	m_elapsedTime = 0;
	m_localOptimizationCount = 0;
	m_remodellingCount = 0;
	m_workDirectoryPath = workDirectoryPath;
	{
		if (!(std::filesystem::is_directory(m_workDirectoryPath)))
		{
			finalize();

			std::string errorMessage = "------------------------------------------------------\n\n";
			errorMessage += "Error   : Invalid argument\n";
			errorMessage += "Location: The member function of \"initialize\" in the class of SolidStatePhysics::Crystallization::Simulators::Crystallizer.\n";
			errorMessage += "Message : \"m_workDirectoryPath\" is not a path to a directory.\n\n";
			errorMessage += "------------------------------------------------------\n\n";
			throw std::invalid_argument(errorMessage);
		}
	}

	if (m_monitoringPlan.needChemicalBondingRecord())
		linkChemicalBondingRecorder();
}

inline void SolidStatePhysics::Crystallization::Simulators::Crystallizer::finalize() noexcept
{
	m_elapsedTime = 0;
	m_localOptimizationCount = 0;
	m_remodellingCount = 0;
	m_workDirectoryPath.clear();
	m_chemicalBondingRecorder.releaseFile();
}

inline void SolidStatePhysics::Crystallization::Simulators::Crystallizer::resetCrystallizationPlan(const CrystallizationPlan& crystallizationPlan) noexcept
{
	m_crystallizationPlan = crystallizationPlan;
}

inline void SolidStatePhysics::Crystallization::Simulators::Crystallizer::resetCrystallizationPlan(CrystallizationPlan&& crystallizationPlan) noexcept
{
	m_crystallizationPlan = std::move(crystallizationPlan);
}

inline void SolidStatePhysics::Crystallization::Simulators::Crystallizer::resetMonitoringPlan(const MonitoringPlan& monitoringPlan) noexcept
{
	m_monitoringPlan = monitoringPlan;
}

inline void SolidStatePhysics::Crystallization::Simulators::Crystallizer::resetMonitoringPlan(MonitoringPlan&& monitoringPlan) noexcept
{
	m_monitoringPlan = std::move(monitoringPlan);
}

inline SolidStatePhysics::Crystallization::Simulators::Crystallizer::size_type SolidStatePhysics::Crystallization::Simulators::Crystallizer::elapsedTime() const noexcept
{
	return m_elapsedTime;
}

inline SolidStatePhysics::Crystallization::Simulators::Crystallizer::size_type SolidStatePhysics::Crystallization::Simulators::Crystallizer::localOptimizationCount() const noexcept
{
	return m_localOptimizationCount;
}

inline SolidStatePhysics::Crystallization::Simulators::Crystallizer::size_type SolidStatePhysics::Crystallization::Simulators::Crystallizer::remodellingCount() const noexcept
{
	return m_remodellingCount;
}

// Public methods
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************

#endif // !SOLIDSTATEPHYSICS_CRYSTALLIZATION_SIMULATORS_CRYSTALLIZER_H
