#ifndef SOLIDSTATEPHYSICS_CRYSTALLIZATION_CALCULATORS_DYNAMICSCONTROLLER_H
#define SOLIDSTATEPHYSICS_CRYSTALLIZATION_CALCULATORS_DYNAMICSCONTROLLER_H

#include "CrystalLattice.h"
#include "IntelligentAtom.h"


namespace SolidStatePhysics
{
	namespace Crystallization
	{
		namespace Calculators
		{
			class DynamicsController
			{
				using CrystalLattice = SolidStatePhysics::Materials::CrystalLattice;
				using IntelligentAtom = SolidStatePhysics::Crystallization::Components::IntelligentAtom;
				using AtomicArrangement = std::vector<IntelligentAtom>;

// **********************************************************************************************************************************************************************************************************************************************************************************************
			// Constructors, destructor, and operators

			public:
				DynamicsController();
				DynamicsController(const double maxDisplacement, const double coolingRate = 1.0);
				virtual ~DynamicsController() = default;

				DynamicsController(const DynamicsController&) = default;
				DynamicsController(DynamicsController&&) noexcept = default;
				DynamicsController& operator=(const DynamicsController&) = default;
				DynamicsController& operator=(DynamicsController&&) noexcept = default;

			// Constructors, destructor, and operators
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
			// Public methods

				void initialize() noexcept;
				void update();
				void clear() noexcept;

				double maxDisplacement() const noexcept;
				double coolingRate() const noexcept;
				double maxLatticeTransformation() const noexcept;

				void resetMaxDisplacement(const double maxDisplacement);
				void resetCoolingRate(const double coolingRate);
				void resetMaxLatticeTransformation(const AtomicArrangement&);

			// Public methods
// **********************************************************************************************************************************************************************************************************************************************************************************************

			private:
				double m_maxDisplacement;
				double m_coolingRate;
				double m_maxLatticeTransformation;
			};
		}
	}
}

// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// Public methods

inline void SolidStatePhysics::Crystallization::Calculators::DynamicsController::update()
{
	m_maxDisplacement *= m_coolingRate;
	m_maxLatticeTransformation *= m_coolingRate;
}

inline double SolidStatePhysics::Crystallization::Calculators::DynamicsController::maxDisplacement() const noexcept
{
	return m_maxDisplacement;
}

inline double SolidStatePhysics::Crystallization::Calculators::DynamicsController::coolingRate() const noexcept
{
	return m_coolingRate;
}

inline double SolidStatePhysics::Crystallization::Calculators::DynamicsController::maxLatticeTransformation() const noexcept
{
	return m_maxLatticeTransformation;
}

inline void SolidStatePhysics::Crystallization::Calculators::DynamicsController::resetMaxDisplacement(const double maxDisplacement)
{
	if (maxDisplacement < 0.0)
	{
		std::string errorMessage = "------------------------------------------------------\n\n";
		errorMessage += "Error   : Invalid argument\n";
		errorMessage += "Location: The function of \"reseMaxDisplacement\" in the class of SolidStatePhysics::Crystallization::Calculators::DynamicsController.\n";
		errorMessage += "Message : \"maxDisplacement\" is less than zero.\n\n";
		errorMessage += "------------------------------------------------------\n\n";
		throw std::invalid_argument(errorMessage);
	}

	else
		m_maxDisplacement = maxDisplacement;
}

inline void SolidStatePhysics::Crystallization::Calculators::DynamicsController::resetCoolingRate(const double coolingRate)
{
	if (coolingRate < 0.0)
	{
		std::string errorMessage = "------------------------------------------------------\n\n";
		errorMessage += "Error   : Invalid argument\n";
		errorMessage += "Location: The function of \"resetCoolingRate\" in the class of SolidStatePhysics::Crystallization::Calculators::DynamicsController.\n";
		errorMessage += "Message : \"coolingRate\" is less than zero.\n\n";
		errorMessage += "------------------------------------------------------\n\n";
		throw std::invalid_argument(errorMessage);
	}

	else
	{
		if (1.0 < coolingRate)
		{
			std::string errorMessage = "------------------------------------------------------\n\n";
			errorMessage += "Error   : Invalid argument\n";
			errorMessage += "Location: The function of \"resetCoolingRate\" in the class of SolidStatePhysics::Crystallization::Calculators::DynamicsController.\n";
			errorMessage += "Message : \"coolingRate\" is more than one.\n\n";
			errorMessage += "------------------------------------------------------\n\n";
			throw std::invalid_argument(errorMessage);
		}

		else
			m_coolingRate = coolingRate;
	}
}

inline void SolidStatePhysics::Crystallization::Calculators::DynamicsController::resetMaxLatticeTransformation(const AtomicArrangement& atomicArrangement)
{
	constexpr double coefficient = (1.0 / 4000.0);

	double sphericalVolume = 0.0;
	{
		constexpr double sphericalVolumeCoef = (4.0 * NumericalAnalysis::MathematicalConstants::getPi() / 3.0);

		for (const auto& intelligentAtom : atomicArrangement)
			sphericalVolume += sphericalVolumeCoef * intelligentAtom.maximumIonicRadius() * intelligentAtom.maximumIonicRadius() * intelligentAtom.maximumIonicRadius();
	}

	m_maxLatticeTransformation = coefficient * sphericalVolume * m_maxDisplacement;
}

// Public methods
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************

#endif // !SOLIDSTATEPHYSICS_CRYSTALLIZATION_CALCULATORS_DYNAMICSCONTROLLER_H
