#ifndef SOLIDSTATEPHYSICS_CRYSTALLIZATION_COMPONENTS_INTERACTINGATOM_H
#define SOLIDSTATEPHYSICS_CRYSTALLIZATION_COMPONENTS_INTERACTINGATOM_H

#include <string>
#include <stdexcept>
#include <map>
#include <unordered_map>
#include <memory>
#include <utility>
#include <vector>

#include "TranslatingAdaptor.h"

#include "AtomicNucleus.h"
#include "ConstraintOwners.h"


namespace SolidStatePhysics
{
	namespace Crystallization
	{
		namespace Components
		{
			class InteractingAtom :public AtomicNucleus
			{
				using size_type = std::size_t;
				using NumericVector = NumericalAnalysis::LinearAlgebra::NumericVector;
				using NumericMatrix = NumericalAnalysis::LinearAlgebra::NumericMatrix;
				using CrystalLattice = SolidStatePhysics::Materials::CrystalLattice;

				using AtomicAdaptor = SolidStatePhysics::Crystallization::Calculators::TranslatingAdaptor<InteractingAtom>;

// **********************************************************************************************************************************************************************************************************************************************************************************************
			// Constructors, destructor, and operator

			public:
				InteractingAtom(const size_type atomicNumber);
				InteractingAtom(const size_type atomicNumber, const NumericVector& cartesianCoordinate);
				virtual ~InteractingAtom() = default;

				InteractingAtom(const InteractingAtom&);
				InteractingAtom(InteractingAtom&&) noexcept;
				InteractingAtom& operator=(const InteractingAtom&);
				InteractingAtom& operator=(InteractingAtom&&) noexcept;

			// Constructors, destructor, and operator
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
			// Access methods

				double minimumIonicRadius() const noexcept;
				double maximumIonicRadius() const noexcept;
				double metallicRadius() const noexcept;

				size_type getCoordinationNumber() const noexcept;
				size_type countMaxCommonBridgingAtomsSharing() const noexcept;
				double getMaxEquilibriumBondLength() const noexcept;
				double getMinEquilibriumBondLength() const noexcept;

				void resetMinMaxIonicRadius(const double minimumIonicRadius, const double maximumIonicRadius);

			// Access methods
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
			// Public methods

				void initialize() noexcept override;
				void normalizeFractionalCoordinate(const CrystalLattice& crystalLattice);

				bool hasChemicalBond(const AtomicAdaptor&) const noexcept;
				bool hasIonicConstrainter(const AtomicAdaptor&) const noexcept;
				bool hasMetallicConstrainter(const AtomicAdaptor&) const noexcept;
				bool hasPolyhedralConstrainter(const AtomicAdaptor&) const noexcept;
				bool hasSecondNeighborConstrainter(const AtomicAdaptor&) const noexcept;

				void clearOwingChemicalBonds() noexcept;
				void clearOwingIonicConstrainters() noexcept;
				void clearOwingMetallicConstrainters() noexcept;
				void clearOwingPolyhedralConstrainters() noexcept;
				void clearOwingSecondNeighborConstrainters() noexcept;

				void eraseChemicalBond(const AtomicAdaptor&) noexcept;
				void eraseIonicConstrainter(const AtomicAdaptor&) noexcept;
				void eraseMetallicConstrainter(const AtomicAdaptor&) noexcept;
				void erasePolyhedralConstrainter(const AtomicAdaptor&) noexcept;
				void eraseSecondNeighborConstrainter(const AtomicAdaptor&) noexcept;

				void eraseInactiveConstraintOwners();

			// Public methods
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
			// Customize methods

				bool configureChemicalBonds(PeriodicLiquid& ownerPeriodicLiquid);
				void configurePolyhedralConstrainters(PeriodicLiquid& ownerPeriodicLiquid);

				void createIonicConstrainter(const AtomicAdaptor& attachedAtom, PeriodicLiquid& ownerPeriodicLiquid);
				void createMetallicConstrainter(const AtomicAdaptor& attachedAtom, PeriodicLiquid& ownerPeriodicLiquid);
				void createSecondNeighborConstrainter(const AtomicAdaptor& attachedAtom, PeriodicLiquid& ownerPeriodicLiquid);

				void eraseInfeasibleChemicalBonds(const double maxBondingLengthErrorRate, PeriodicLiquid& ownerPeriodicLiquid);

			// Customize methods
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
			// Pure virtual functions

				virtual bool hasIdealCoordinationNumber() const noexcept = 0;
				virtual bool isAnion() const noexcept = 0;
				virtual bool isCation() const noexcept = 0;
				virtual size_type maxCommonBridgingAtomsSharing() const noexcept = 0;
				virtual std::pair<size_type, double> bondingMotivation() const noexcept = 0;

			// Pure virtual functions
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
			// Protected methods

			protected:
				void createChemicalBond(const AtomicAdaptor& attachedAtom, PeriodicLiquid& ownerPeriodicLiquid);
				void createPolyhedralConstrainter(const AtomicAdaptor& attachedAtom, PeriodicLiquid& ownerPeriodicLiquid, const AtomicAdaptor& bridgingAtom);

				std::string getChemicalBondsKey() const noexcept;
				std::string getCommonBridgingAtomsSharingsKey() const noexcept;
				std::string getIonicConstraintsKey(const double maxBondingLengthErrorRate) const noexcept;

			// Protected methods
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
			// Private methods

			private:
				bool tryConfigureChemicalBonds(const std::map<double, AtomicAdaptor>& promisingAtomicAdaptors, const size_type additionalCoordinationNumber, PeriodicLiquid& ownerPeriodicLiquid);

				template <typename CO>
				bool hasConstrainter(const AtomicAdaptor&, const std::unordered_map<AtomicAdaptor, CO, AtomicAdaptor::Hasher>& forceOwners) const noexcept;

				template <typename CO>
				void createConstrainter(const AtomicAdaptor&, std::unordered_map<AtomicAdaptor, CO, AtomicAdaptor::Hasher>& forceOwners, std::unordered_map<AtomicAdaptor, CO, AtomicAdaptor::Hasher>& twinForceOwners, PeriodicLiquid& ownerPeriodicLiquid);

				template <typename CO>
				void eraseConstrainter(const AtomicAdaptor&, std::unordered_map<AtomicAdaptor, CO, AtomicAdaptor::Hasher>& forceOwners) noexcept;

				template <typename CO>
				void eraseInactiveConstraintOwners(std::unordered_map<AtomicAdaptor, CO, AtomicAdaptor::Hasher>& forceOwners) noexcept;

				template <typename CO>
				size_type countActiveConstraintOwners(const std::unordered_map<AtomicAdaptor, CO, AtomicAdaptor::Hasher>& forceOwners) const noexcept;

			// Private methods
// **********************************************************************************************************************************************************************************************************************************************************************************************

			private:
				double m_minimumIonicRadius;
				double m_maximumIonicRadius;
				double m_metallicRadius;
				std::unordered_map<AtomicAdaptor, ChemicalBondOwner, AtomicAdaptor::Hasher> m_chemicalBondOwners;
				std::unordered_map<AtomicAdaptor, IonicConstraintOwner, AtomicAdaptor::Hasher> m_ionicConstraintOwners;
				std::unordered_map<AtomicAdaptor, MetallicConstraintOwner, AtomicAdaptor::Hasher> m_metallicConstraintOwners;
				std::unordered_map<AtomicAdaptor, PolyhedralConstraintOwner, AtomicAdaptor::Hasher> m_polyhedralConstraintOwners;
				std::unordered_map<AtomicAdaptor, SecondNeighborConstraintOwner, AtomicAdaptor::Hasher> m_secondNeighborConstraintOwners;
			};
		}
	}
}

// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// Access methods

inline double SolidStatePhysics::Crystallization::Components::InteractingAtom::minimumIonicRadius() const noexcept
{
	return m_minimumIonicRadius;
}

inline double SolidStatePhysics::Crystallization::Components::InteractingAtom::maximumIonicRadius() const noexcept
{
	return m_maximumIonicRadius;
}

inline double SolidStatePhysics::Crystallization::Components::InteractingAtom::metallicRadius() const noexcept
{
	return m_metallicRadius;
}

inline SolidStatePhysics::Crystallization::Components::InteractingAtom::size_type SolidStatePhysics::Crystallization::Components::InteractingAtom::getCoordinationNumber() const noexcept
{
	size_type chemicalBonding = 0;
	{
		for (const auto& chemicalBondOwner : m_chemicalBondOwners)
		{
			if (chemicalBondOwner.second.isActive())
				++chemicalBonding;
		}
	}

	return chemicalBonding;
}

inline void SolidStatePhysics::Crystallization::Components::InteractingAtom::resetMinMaxIonicRadius(const double minimumIonicRadius, const double maximumIonicRadius)
{
	if (0.0 < minimumIonicRadius && 0.0 < maximumIonicRadius)
	{
		if (maximumIonicRadius < minimumIonicRadius)
		{
			std::string errorMessage = "------------------------------------------------------\n\n";
			errorMessage += "Error   : Invalid argument\n";
			errorMessage += "Location: The member function of \"resetMinMaxIonicRadius\" in the class of SolidStatePhysics::Crystallization::Components::InteractingAtom.\n";
			errorMessage += "Message : \"maximumIonicRadius\" is less than \"minimumIonicRadius\".\n\n";
			errorMessage += "------------------------------------------------------\n\n";
			throw std::logic_error(errorMessage);
		}

		else
		{
			m_minimumIonicRadius = minimumIonicRadius;
			m_maximumIonicRadius = maximumIonicRadius;
		}
	}

	else
	{
		std::string errorMessage = "------------------------------------------------------\n\n";
		errorMessage += "Error   : Invalid argument\n";
		errorMessage += "Location: The member function of \"resetMinMaxIonicRadius\" in the class of SolidStatePhysics::Crystallization::Components::InteractingAtom.\n";
		errorMessage += "Message : \"minimumIonicRadius\" and/or \"maximumIonicRadius\" are not more than zero.\n\n";
		errorMessage += "------------------------------------------------------\n\n";
		throw std::logic_error(errorMessage);
	}
}

// Access methods
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// Public methods

inline bool SolidStatePhysics::Crystallization::Components::InteractingAtom::hasChemicalBond(const AtomicAdaptor& atomicAdaptor) const noexcept
{
	return hasConstrainter(atomicAdaptor, m_chemicalBondOwners);
}

inline bool SolidStatePhysics::Crystallization::Components::InteractingAtom::hasIonicConstrainter(const AtomicAdaptor& atomicAdaptor) const noexcept
{
	return hasConstrainter(atomicAdaptor, m_ionicConstraintOwners);
}

inline bool SolidStatePhysics::Crystallization::Components::InteractingAtom::hasMetallicConstrainter(const AtomicAdaptor& atomicAdaptor) const noexcept
{
	return hasConstrainter(atomicAdaptor, m_metallicConstraintOwners);
}

inline bool SolidStatePhysics::Crystallization::Components::InteractingAtom::hasPolyhedralConstrainter(const AtomicAdaptor& atomicAdaptor) const noexcept
{
	return hasConstrainter(atomicAdaptor, m_polyhedralConstraintOwners);
}

inline bool SolidStatePhysics::Crystallization::Components::InteractingAtom::hasSecondNeighborConstrainter(const AtomicAdaptor& atomicAdaptor) const noexcept
{
	return hasConstrainter(atomicAdaptor, m_secondNeighborConstraintOwners);
}

inline void SolidStatePhysics::Crystallization::Components::InteractingAtom::clearOwingChemicalBonds() noexcept
{
	m_chemicalBondOwners.clear();
}

inline void SolidStatePhysics::Crystallization::Components::InteractingAtom::clearOwingIonicConstrainters() noexcept
{
	m_ionicConstraintOwners.clear();
}

inline void SolidStatePhysics::Crystallization::Components::InteractingAtom::clearOwingMetallicConstrainters() noexcept
{
	m_metallicConstraintOwners.clear();
}

inline void SolidStatePhysics::Crystallization::Components::InteractingAtom::clearOwingPolyhedralConstrainters() noexcept
{
	m_polyhedralConstraintOwners.clear();
}

inline void SolidStatePhysics::Crystallization::Components::InteractingAtom::clearOwingSecondNeighborConstrainters() noexcept
{
	m_secondNeighborConstraintOwners.clear();
}

inline void SolidStatePhysics::Crystallization::Components::InteractingAtom::eraseChemicalBond(const AtomicAdaptor& atomicAdaptor) noexcept
{
	eraseConstrainter(atomicAdaptor, m_chemicalBondOwners);
}

inline void SolidStatePhysics::Crystallization::Components::InteractingAtom::eraseIonicConstrainter(const AtomicAdaptor& atomicAdaptor) noexcept
{
	eraseConstrainter(atomicAdaptor, m_ionicConstraintOwners);
}

inline void SolidStatePhysics::Crystallization::Components::InteractingAtom::eraseMetallicConstrainter(const AtomicAdaptor& atomicAdaptor) noexcept
{
	eraseConstrainter(atomicAdaptor, m_metallicConstraintOwners);
}

inline void SolidStatePhysics::Crystallization::Components::InteractingAtom::erasePolyhedralConstrainter(const AtomicAdaptor& atomicAdaptor) noexcept
{
	eraseConstrainter(atomicAdaptor, m_polyhedralConstraintOwners);
}

inline void SolidStatePhysics::Crystallization::Components::InteractingAtom::eraseSecondNeighborConstrainter(const AtomicAdaptor& atomicAdaptor) noexcept
{
	eraseConstrainter(atomicAdaptor, m_secondNeighborConstraintOwners);
}

inline void SolidStatePhysics::Crystallization::Components::InteractingAtom::eraseInactiveConstraintOwners()
{
	eraseInactiveConstraintOwners(m_chemicalBondOwners);
	eraseInactiveConstraintOwners(m_ionicConstraintOwners);
	eraseInactiveConstraintOwners(m_metallicConstraintOwners);
	eraseInactiveConstraintOwners(m_polyhedralConstraintOwners);
	eraseInactiveConstraintOwners(m_secondNeighborConstraintOwners);
}

// Public methods
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// Private methods

template <typename CO>
inline bool SolidStatePhysics::Crystallization::Components::InteractingAtom::hasConstrainter(const AtomicAdaptor& atomicAdaptor, const std::unordered_map<AtomicAdaptor, CO, AtomicAdaptor::Hasher>& forceOwners) const noexcept
{
	auto iter = forceOwners.find(atomicAdaptor);

	if (iter == forceOwners.end())
		return false;

	else
	{
		if (iter->second.isActive())
			return true;
		else
			return false;
	}
}

template <typename CO>
inline void SolidStatePhysics::Crystallization::Components::InteractingAtom::createConstrainter(const AtomicAdaptor& atomicAdaptor, std::unordered_map<AtomicAdaptor, CO, AtomicAdaptor::Hasher>& forceOwners, std::unordered_map<AtomicAdaptor, CO, AtomicAdaptor::Hasher>& twinForceOwners, PeriodicLiquid& ownerPeriodicLiquid)
{
	if (hasConstrainter(atomicAdaptor, forceOwners))
	{
		std::string errorMessage = "------------------------------------------------------\n\n";
		errorMessage += "Error   : Logic error\n";
		errorMessage += "Location: The member function of \"createConstrainter\" in the class of SolidStatePhysics::Crystallization::Components::InteractingAtom.\n";
		errorMessage += "Message : This has already created \"Constrainter\" with \"atomicAdaptor\".\n\n";
		errorMessage += "------------------------------------------------------\n\n";
		throw std::logic_error(errorMessage);
	}

	else
	{
		AtomicAdaptor twinAtomicAdaptor{ *this, atomicAdaptor };
		twinAtomicAdaptor.reverseLatticePoint();

		if (atomicAdaptor.entity().hasConstrainter(twinAtomicAdaptor, twinForceOwners))
		{
			std::string errorMessage = "------------------------------------------------------\n\n";
			errorMessage += "Error   : Logic error\n";
			errorMessage += "Location: The member function of \"createConstrainter\" in the class of SolidStatePhysics::Crystallization::Components::InteractingAtom.\n";
			errorMessage += "Message : \"atomicAdaptor\" has already created \"Constrainter\" with this.\n\n";
			errorMessage += "------------------------------------------------------\n\n";
			throw std::logic_error(errorMessage);
		}

		else
		{
			auto emplaceResult = forceOwners.emplace(atomicAdaptor, CO{ *this });
			auto twinEmplaceResult = twinForceOwners.emplace(twinAtomicAdaptor, CO{ atomicAdaptor.entity() });
			CO::Constrainter::create(emplaceResult.first->second, SolidStatePhysics::Crystallization::Calculators::TranslatingAdaptor<CO>{ twinEmplaceResult.first->second, atomicAdaptor });
			emplaceResult.first->second.registerOwingConstrainter(ownerPeriodicLiquid);
		}
	}
}

template <typename CO>
inline void SolidStatePhysics::Crystallization::Components::InteractingAtom::eraseConstrainter(const AtomicAdaptor& atomicAdaptor, std::unordered_map<AtomicAdaptor, CO, AtomicAdaptor::Hasher>& forceOwners) noexcept
{
	auto iter = forceOwners.find(atomicAdaptor);

	if (iter != forceOwners.end())
		forceOwners.erase(iter);
}

template <typename CO>
inline void SolidStatePhysics::Crystallization::Components::InteractingAtom::eraseInactiveConstraintOwners(std::unordered_map<AtomicAdaptor, CO, AtomicAdaptor::Hasher>& forceOwners) noexcept
{
	for (auto iter = forceOwners.begin(); iter != forceOwners.end();)
	{
		if (iter->second.isActive())
			++iter;
		else
			iter = forceOwners.erase(iter);
	}
}

template <typename CO>
inline SolidStatePhysics::Crystallization::Components::InteractingAtom::size_type SolidStatePhysics::Crystallization::Components::InteractingAtom::countActiveConstraintOwners(const std::unordered_map<AtomicAdaptor, CO, AtomicAdaptor::Hasher>& forceOwners) const noexcept
{
	size_type activeForceOwnersCount = 0;
	{
		for (auto iter = forceOwners.begin(); iter != forceOwners.end(); ++iter)
		{
			if (iter->second.isActive())
				++activeForceOwnersCount;
		}
	}

	return activeForceOwnersCount;
}

// Private methods
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************

#endif // !SOLIDSTATEPHYSICS_CRYSTALLIZATION_COMPONENTS_INTERACTINGATOM_H
