#ifndef SOLIDSTATEPHYSICS_CRYSTALLIZATION_CALCULATORS_INTERATOMICCONNECTOR_H
#define SOLIDSTATEPHYSICS_CRYSTALLIZATION_CALCULATORS_INTERATOMICCONNECTOR_H

#include <memory>
#include <functional>

#include "LinearAlgebra.h"
#include "CrystalLattice.h"

#include "TranslatingAdaptor.h"
#include "InteractingAtom.h"


namespace SolidStatePhysics
{
	namespace Crystallization
	{
		namespace Calculators
		{
			class InteratomicConnector
			{
			protected:
				using size_type = std::size_t;
				using NumericVector = NumericalAnalysis::LinearAlgebra::NumericVector;
				using CrystalLattice = SolidStatePhysics::Materials::CrystalLattice;
				using InteractingAtom = SolidStatePhysics::Crystallization::Components::InteractingAtom;
				using AtomicAdaptor = TranslatingAdaptor<SolidStatePhysics::Crystallization::Components::InteractingAtom>;

// **********************************************************************************************************************************************************************************************************************************************************************************************
			// Protected constructors, protected support function, and public destructor

			protected:
				InteratomicConnector(InteractingAtom& ownerAtom, AtomicAdaptor&& attachedAtom);
				explicit InteratomicConnector(const std::shared_ptr<InteratomicConnector>& twinPtr);

				void resetTwinInteratomicConnectorPtr(const std::shared_ptr<InteratomicConnector>& twinPtr) noexcept;

			public:
				virtual ~InteratomicConnector() = default;

			// Protected constructors, protected support function, and public destructor
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
			// Static methods

			protected:
				template <typename CO>
				static void createInteratomicComponent(CO& realOwner, const TranslatingAdaptor<CO>& attachedOwner);

				template <typename IC>
				static void eraseInteratomicComponent(std::shared_ptr<IC>&& interatomicComponentPtr);

			// Static methods
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
			// Access methods

			public:
				CrystalLattice& crystalLattice() const noexcept;
				InteractingAtom& ownerAtom() const noexcept;
				const AtomicAdaptor& attachedAtom() const noexcept;
				AtomicAdaptor& attachedAtom() noexcept;

				const NumericVector& displacement() const noexcept;
				double interatomicDistanceSquare() const noexcept;
				double interatomicDistance() const noexcept;

			// Access methods
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
			// Public methods

				virtual void update();

			// Public methods
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
			// Protected methods

			protected:
				void synchronizeDisplacement();

				std::shared_ptr<InteratomicConnector> twinPtr() const noexcept;
				void resetTwinPtr() noexcept;
				void updateDisplacement() noexcept;

				void resetOwnerAtom(InteractingAtom& ownerAtom);
				void resetAttachedAtom(const AtomicAdaptor& attachedAtom);
				void resetAttachedAtom(AtomicAdaptor&& attachedAtom);
				void resetAttachedAtomEntity(InteractingAtom& attachedAtomEntity);

			// Protected methods
// **********************************************************************************************************************************************************************************************************************************************************************************************

			private:
				std::reference_wrapper<CrystalLattice> m_crystalLattice;
				std::reference_wrapper<InteractingAtom> m_ownerAtom;
				AtomicAdaptor m_attachedAtom;
				std::weak_ptr<InteratomicConnector> m_twinPtr;

				NumericVector m_displacement;
				double m_interatomicDistance;

			private:
				InteratomicConnector() = delete;
				InteratomicConnector(const InteratomicConnector&) = delete;
				InteratomicConnector(InteratomicConnector&&) = delete;
				InteratomicConnector& operator=(const InteratomicConnector&) = delete;
				InteratomicConnector& operator=(InteratomicConnector&&) = delete;
			};
		}
	}
}

// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// Support method

inline void SolidStatePhysics::Crystallization::Calculators::InteratomicConnector::resetTwinInteratomicConnectorPtr(const std::shared_ptr<InteratomicConnector>& twinPtr) noexcept
{
	m_twinPtr = twinPtr;
}

// Support method
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// Static methods

template <typename CO>
inline void SolidStatePhysics::Crystallization::Calculators::InteratomicConnector::createInteratomicComponent(CO& realOwner, const TranslatingAdaptor<CO>& attachedOwner)
{
	std::shared_ptr<typename CO::Constrainter> constrainterPtr{ new typename CO::Constrainter(realOwner, attachedOwner) };
	std::shared_ptr<typename CO::Constrainter> twinConstrainterPtr{ new typename CO::Constrainter(constrainterPtr) };
	constrainterPtr->resetTwinInteratomicConnectorPtr(std::static_pointer_cast<InteratomicConnector>(twinConstrainterPtr));

	realOwner.join(std::move(constrainterPtr));
	attachedOwner.entity().join(std::move(twinConstrainterPtr));
}

template <typename IC>
inline void SolidStatePhysics::Crystallization::Calculators::InteratomicConnector::eraseInteratomicComponent(std::shared_ptr<IC>&& interatomicComponentPtr)
{
	if (2 < interatomicComponentPtr.use_count())
	{
		std::string errorMessage = "------------------------------------------------------\n\n";
		errorMessage += "Error   : Logic error\n";
		errorMessage += "Location: The static member function of \"erase\" in the class of SolidStatePhysics::Crystallization::Calculators::ForceCarrier.\n";
		errorMessage += "Message : The number of the users of \"interatomicComponentPtr\" is not two.\n\n";
		errorMessage += "------------------------------------------------------\n\n";
		throw std::logic_error(errorMessage);
	}

	else
	{
		interatomicComponentPtr->realOwner().disjoin();
		interatomicComponentPtr.reset();
	}
}

// Static method
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// Access methods

inline SolidStatePhysics::Materials::CrystalLattice& SolidStatePhysics::Crystallization::Calculators::InteratomicConnector::crystalLattice() const noexcept
{
	return m_crystalLattice.get();
}

inline SolidStatePhysics::Crystallization::Components::InteractingAtom& SolidStatePhysics::Crystallization::Calculators::InteratomicConnector::ownerAtom() const noexcept
{
	return m_ownerAtom.get();
}

inline const SolidStatePhysics::Crystallization::Calculators::InteratomicConnector::AtomicAdaptor& SolidStatePhysics::Crystallization::Calculators::InteratomicConnector::attachedAtom() const noexcept
{
	return m_attachedAtom;
}

inline SolidStatePhysics::Crystallization::Calculators::InteratomicConnector::AtomicAdaptor& SolidStatePhysics::Crystallization::Calculators::InteratomicConnector::attachedAtom() noexcept
{
	return m_attachedAtom;
}

inline const NumericalAnalysis::LinearAlgebra::NumericVector& SolidStatePhysics::Crystallization::Calculators::InteratomicConnector::displacement() const noexcept
{
	return m_displacement;
}

inline double SolidStatePhysics::Crystallization::Calculators::InteratomicConnector::interatomicDistanceSquare() const noexcept
{
	return m_displacement.normSquare();
}

inline double SolidStatePhysics::Crystallization::Calculators::InteratomicConnector::interatomicDistance() const noexcept
{
	return m_interatomicDistance;
}

// Access methods
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// Public methods

inline void SolidStatePhysics::Crystallization::Calculators::InteratomicConnector::update()
{
	updateDisplacement();
	synchronizeDisplacement();
}

// Public methods
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// Protected methods

inline void SolidStatePhysics::Crystallization::Calculators::InteratomicConnector::synchronizeDisplacement()
{
	auto twinPtr = m_twinPtr.lock();
	twinPtr->m_displacement = ((-1.0) * m_displacement);
	twinPtr->m_interatomicDistance = m_interatomicDistance;
}

inline std::shared_ptr<SolidStatePhysics::Crystallization::Calculators::InteratomicConnector> SolidStatePhysics::Crystallization::Calculators::InteratomicConnector::twinPtr() const noexcept
{
	return m_twinPtr.lock();
}

inline void SolidStatePhysics::Crystallization::Calculators::InteratomicConnector::resetTwinPtr() noexcept
{
	m_twinPtr.reset();
}

inline void SolidStatePhysics::Crystallization::Calculators::InteratomicConnector::updateDisplacement() noexcept
{
	if (m_attachedAtom.isInOriginalCell())
	{
		m_displacement = (m_attachedAtom.entity().cartesianCoordinate() - m_ownerAtom.get().cartesianCoordinate());
		m_interatomicDistance = m_displacement.norm();
	}

	else
	{
		m_displacement = (m_attachedAtom.entity().cartesianCoordinate() + m_attachedAtom.translationVector());
		m_displacement -= m_ownerAtom.get().cartesianCoordinate();
		m_interatomicDistance = m_displacement.norm();
	}
}

inline void SolidStatePhysics::Crystallization::Calculators::InteratomicConnector::resetOwnerAtom(InteractingAtom& ownerAtom)
{
	m_ownerAtom = std::ref(ownerAtom);
}

inline void SolidStatePhysics::Crystallization::Calculators::InteratomicConnector::resetAttachedAtom(const AtomicAdaptor& attachedAtom)
{
	m_attachedAtom = attachedAtom;
}

inline void SolidStatePhysics::Crystallization::Calculators::InteratomicConnector::resetAttachedAtom(AtomicAdaptor&& attachedAtom)
{
	m_attachedAtom = std::move(attachedAtom);
}

inline void SolidStatePhysics::Crystallization::Calculators::InteratomicConnector::resetAttachedAtomEntity(InteractingAtom& attachedAtomEntity)
{
	m_attachedAtom.resetEntity(attachedAtomEntity);
}

// Protected methods
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************

#endif // !SOLIDSTATEPHYSICS_CRYSTALLIZATION_CALCULATORS_INTERATOMICCONNECTOR_H
