#ifndef CONTROLCENTER_ADMINISTERING_MATERIALSEXPLORER_H
#define CONTROLCENTER_ADMINISTERING_MATERIALSEXPLORER_H

#include <string>
#include <filesystem>
#include <vector>

#include "CrystallizationPlan.h"
#include "MonitoringPlan.h"
#include "RequestedFeedstocks.h"


namespace ControlCenter
{
	namespace Administering
	{
		class MaterialsExplorer
		{
			using size_type = std::size_t;
			using ChemicalConstituents = SolidStatePhysics::Materials::ChemicalConstituents;
			using CrystallizationPlan = SolidStatePhysics::Crystallization::Simulators::CrystallizationPlan;
			using MonitoringPlan = SolidStatePhysics::Crystallization::Simulators::MonitoringPlan;

// **********************************************************************************************************************************************************************************************************************************************************************************************
		// Constructors and destructor

		public:
			MaterialsExplorer() = default;
			MaterialsExplorer(const CrystallizationPlan&, const MonitoringPlan&);
			MaterialsExplorer(CrystallizationPlan&&, MonitoringPlan&&);

			virtual ~MaterialsExplorer() = default;

		// Constructors and destructor
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
		// Access methods

			const CrystallizationPlan& crystallizationPlan() const noexcept;
			const MonitoringPlan& monitoringPlan() const noexcept;

			void resetCrystallizationPlan(const CrystallizationPlan&);
			void resetCrystallizationPlan(CrystallizationPlan&&);
			void resetMonitoringPlan(const MonitoringPlan&);
			void resetMonitoringPlan(MonitoringPlan&&);

		// Access methods
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
		// Public methods

			void executePeriodicSolidsGeneration(RequestedFeedstocks&&);
			void executePeriodicSolidsReheating();

		// Public methods
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
		// Private methods

		private:
			void initializeGeometricalConstraints();

		// Public methods
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
		// Private methods for generations of structures

			void operatePeriodicSolidsGenerator();
			void initializeFinalSimulationsReporter(const std::filesystem::path& chemicalConstituentsDirectoryPath);
			void assignChemicalConstituentsToMpiProcess(const RequestedFeedstocks& requestedFeedstocks);
			void tryInitializeChemicalSystemDirectory(const std::string& chemicalSystemName);

		// Private methods generations of structures
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
		// Private methods for reoptimizing

			void operatePeriodicSolidsReheater();
			void assignReheatingChemicalConstituentsDirectoriesToMpiProcess();
			bool hasDatFile(const std::filesystem::path& workDirectoryPath);
			void createReheatingChemicalConstituentsDirectoriesFile();
			void removeReheatingChemicalConstituentsDirectoriesFile();

			std::string getDirectoryName(std::filesystem::path directoryPath) const noexcept;

		// Private methods for reoptimizing
// **********************************************************************************************************************************************************************************************************************************************************************************************

		private:
			CrystallizationPlan m_crystallizationPlan;
			MonitoringPlan m_monitoringPlan;

			mutable size_type m_feedstocksPerChemicalConstituents{ 0 };
			mutable std::vector<ChemicalConstituents> m_targetChemicalConstituents{};
			mutable std::filesystem::path m_chemicalSystemDirectoryPath{};

			mutable std::vector<std::filesystem::path> m_reheatingChemicalConstituentsDirectories{};

		private:
			MaterialsExplorer(const MaterialsExplorer&) = delete;
			MaterialsExplorer(MaterialsExplorer&&) noexcept = delete;
			MaterialsExplorer& operator=(const MaterialsExplorer&) = delete;
			MaterialsExplorer& operator=(MaterialsExplorer&&) noexcept = delete;
		};
	}
}

// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// Access methods

inline const SolidStatePhysics::Crystallization::Simulators::CrystallizationPlan& ControlCenter::Administering::MaterialsExplorer::crystallizationPlan() const noexcept
{
	return m_crystallizationPlan;
}

inline const SolidStatePhysics::Crystallization::Simulators::MonitoringPlan& ControlCenter::Administering::MaterialsExplorer::monitoringPlan() const noexcept
{
	return m_monitoringPlan;
}

inline void ControlCenter::Administering::MaterialsExplorer::resetCrystallizationPlan(const CrystallizationPlan& crystallizationPlan)
{
	m_crystallizationPlan = crystallizationPlan;
}

inline void ControlCenter::Administering::MaterialsExplorer::resetCrystallizationPlan(CrystallizationPlan&& crystallizationPlan)
{
	m_crystallizationPlan = std::move(crystallizationPlan);
}

inline void ControlCenter::Administering::MaterialsExplorer::resetMonitoringPlan(const MonitoringPlan& monitoringPlan)
{
	m_monitoringPlan = monitoringPlan;
}

inline void ControlCenter::Administering::MaterialsExplorer::resetMonitoringPlan(MonitoringPlan&& monitoringPlan)
{
	m_monitoringPlan = std::move(monitoringPlan);
}

// Access methods
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// Public methods

inline void ControlCenter::Administering::MaterialsExplorer::executePeriodicSolidsGeneration(RequestedFeedstocks&& requestedFeedstocks)
{
	initializeGeometricalConstraints();
	m_feedstocksPerChemicalConstituents = requestedFeedstocks.feedstocksPerChemicalConstituents();
	
	assignChemicalConstituentsToMpiProcess(requestedFeedstocks);
	requestedFeedstocks.clear();
	operatePeriodicSolidsGenerator();

	m_feedstocksPerChemicalConstituents = 0;
	m_targetChemicalConstituents.clear();
	m_chemicalSystemDirectoryPath.clear();
}

inline void ControlCenter::Administering::MaterialsExplorer::executePeriodicSolidsReheating()
{
	initializeGeometricalConstraints();

	assignReheatingChemicalConstituentsDirectoriesToMpiProcess();
	createReheatingChemicalConstituentsDirectoriesFile();
	operatePeriodicSolidsReheater();
	removeReheatingChemicalConstituentsDirectoriesFile();

	m_reheatingChemicalConstituentsDirectories.clear();
}

// Public methods
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// Private methods

inline std::string ControlCenter::Administering::MaterialsExplorer::getDirectoryName(std::filesystem::path directoryPath) const noexcept
{
	std::string name;
	{
		std::string directoryPathName = directoryPath.generic_string();

		if (directoryPathName.back() == '/')
		{
			directoryPathName.pop_back();
			directoryPath = std::filesystem::path{ directoryPathName };
		}

		name = directoryPath.filename().string();
	}

	return name;
}

// Private methods
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************

#endif // !CONTROLCENTER_ADMINISTERING_MATERIALSEXPLORER_H
