#ifndef SOLIDSTATEPHYSICS_MATERIALS_IO_MDFILEOWNER_H
#define SOLIDSTATEPHYSICS_MATERIALS_IO_MDFILEOWNER_H

#include <stdexcept>
#include <string>
#include <filesystem>

#include "FileOwner.h"

#include "LinearAlgebra.h"
#include "UnitConversion.h"

#include "CrystalStructure.h"


namespace SolidStatePhysics
{
	namespace Materials
	{
		namespace IO
		{
			class MdFileOwner :public ControlCenter::System::IO::FileOwner
			{
				using size_type = std::size_t;
				using NumericVector = NumericalAnalysis::LinearAlgebra::NumericVector;
				using NumericMatrix = NumericalAnalysis::LinearAlgebra::NumericMatrix;

// **********************************************************************************************************************************************************************************************************************************************************************************************
			// Constructors, destructor, and operators

			public:
				MdFileOwner() = default;
				explicit MdFileOwner(const std::filesystem::path& mdFilePath);
				explicit MdFileOwner(std::filesystem::path&& mdFilePath);

				MdFileOwner(MdFileOwner&&) noexcept = default;
				MdFileOwner& operator=(MdFileOwner&&) noexcept = default;

				virtual ~MdFileOwner() = default;

			// Constructors, destructor, and operators
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
			// Utility

				void releaseFile() noexcept override;

				static std::string correctExtension() noexcept;
				static bool isCorrectExtension(const std::filesystem::path& filePath) noexcept;

			// Utility
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
			// Public methods to make animation

				template <typename A, typename AA>
				void drawFrame(const CrystalStructure<A, AA>&, const size_type elapsedTimeCount);

				template <typename A, typename AA>
				void drawFrame(const CrystalStructure<A, AA>&, const double elapsedTime);

				template <typename A, typename AA>
				void drawFrame(const CrystalStructure<A, AA>&, const size_type elapsedTimeCount, const double totalEnergy);

				template <typename A, typename AA>
				void drawFrame(const CrystalStructure<A, AA>&, const double elapsedTime, const double totalEnergy);

			// Public methods to make animation
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
			// Private methods

			private:
				std::string writeHeaderMessage(const std::size_t totalAtomCount, const size_type elapsedTimeCount) const;
				std::string writeHeaderMessage(const std::size_t totalAtomCount, const double elapsedTime) const;
				std::string writeHeaderMessage(const std::size_t totalAtomCount, const size_type elapsedTimeCount, const double totalEnergy) const;
				std::string writeHeaderMessage(const std::size_t totalAtomCount, const double elapsedTime, const double totalEnergy) const;

				template <typename A, typename AA>
				std::string writePrimitiveLatticeVectors(const CrystalStructure<A, AA>&) const;

				template <typename A, typename AA>
				std::string writeAtomicArrangement(const CrystalStructure<A, AA>&) const;

			// Private methods
// **********************************************************************************************************************************************************************************************************************************************************************************************

			private:
				mutable std::size_t m_frameNumber{ 0 };

			private:
				MdFileOwner(const MdFileOwner&) = delete;
				MdFileOwner& operator=(const MdFileOwner&) = delete;
			};
		}
	}
}

// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// Utility

inline void SolidStatePhysics::Materials::IO::MdFileOwner::releaseFile() noexcept
{
	FileOwner::releaseFile();
	m_frameNumber = 0;
}

inline std::string SolidStatePhysics::Materials::IO::MdFileOwner::correctExtension() noexcept
{
	return std::string{ ".md" };
}

inline bool SolidStatePhysics::Materials::IO::MdFileOwner::isCorrectExtension(const std::filesystem::path& filePath) noexcept
{
	if (filePath.extension().string() == correctExtension())
		return true;
	else
		return false;
}

// Utility
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// Public methods to make animation

template <typename A, typename AA>
inline void SolidStatePhysics::Materials::IO::MdFileOwner::drawFrame(const CrystalStructure<A, AA>& crystalStructure, const size_type elapsedTimeCount)
{
	std::string outputString;
	outputString += writeHeaderMessage(crystalStructure.countTotalAtoms(), elapsedTimeCount);
	outputString += writePrimitiveLatticeVectors(crystalStructure);
	outputString += writeAtomicArrangement(crystalStructure);
	write(outputString);

	++m_frameNumber;
}

template <typename A, typename AA>
inline void SolidStatePhysics::Materials::IO::MdFileOwner::drawFrame(const CrystalStructure<A, AA>& crystalStructure, const double elapsedTime)
{
	std::string outputString;
	outputString += writeHeaderMessage(crystalStructure.countTotalAtoms(), elapsedTime);
	outputString += writePrimitiveLatticeVectors(crystalStructure);
	outputString += writeAtomicArrangement(crystalStructure);
	write(outputString);

	++m_frameNumber;
}

template <typename A, typename AA>
inline void SolidStatePhysics::Materials::IO::MdFileOwner::drawFrame(const CrystalStructure<A, AA>& crystalStructure, const size_type elapsedTimeCount, const double totalEnergy)
{
	std::string outputString;
	outputString += writeHeaderMessage(crystalStructure.countTotalAtoms(), elapsedTimeCount, totalEnergy);
	outputString += writePrimitiveLatticeVectors(crystalStructure);
	outputString += writeAtomicArrangement(crystalStructure);
	write(outputString);

	++m_frameNumber;
}

template <typename A, typename AA>
inline void SolidStatePhysics::Materials::IO::MdFileOwner::drawFrame(const CrystalStructure<A, AA>& crystalStructure, const double elapsedTime, const double totalEnergy)
{
	std::string outputString;
	outputString += writeHeaderMessage(crystalStructure.countTotalAtoms(), elapsedTime, totalEnergy);
	outputString += writePrimitiveLatticeVectors(crystalStructure);
	outputString += writeAtomicArrangement(crystalStructure);
	write(outputString);

	++m_frameNumber;
}

// Public methods to make animation
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// Private methods

inline std::string SolidStatePhysics::Materials::IO::MdFileOwner::writeHeaderMessage(const std::size_t totalAtomCount, const size_type elapsedTimeCount) const
{
	std::string outputString;
	{
		outputString += std::to_string(totalAtomCount);
		outputString += "\n   time_count=    ";
		outputString += std::to_string(elapsedTimeCount);
		outputString += " ()    ";
	}

	return outputString;
}

inline std::string SolidStatePhysics::Materials::IO::MdFileOwner::writeHeaderMessage(const std::size_t totalAtomCount, const double elapsedTime) const
{
	using time_cast = SolidStatePhysics::UnitSystem::time_cast<SolidStatePhysics::UnitSystem::TimeUnit::Femtosecond, SolidStatePhysics::UnitSystem::TimeUnit::AtomicUnit>;

	std::string outputString;
	{
		outputString += std::to_string(totalAtomCount);
		outputString += "\n   time=    ";
		outputString += std::to_string(time_cast{}(elapsedTime));
		outputString += " (fs)     ";
	}

	return outputString;
}

inline std::string SolidStatePhysics::Materials::IO::MdFileOwner::writeHeaderMessage(const std::size_t totalAtomCount, const size_type elapsedTimeCount, const double totalEnergy) const
{
	std::string outputString;
	{
		outputString += std::to_string(totalAtomCount);
		outputString += "\n   time_count=    ";
		outputString += std::to_string(elapsedTimeCount);
		outputString += " ()     ";
		outputString += "Energy= ";
		outputString += std::to_string(totalEnergy);
		outputString += " ()  ";
	}

	return outputString;
}

inline std::string SolidStatePhysics::Materials::IO::MdFileOwner::writeHeaderMessage(const std::size_t totalAtomCount, const double elapsedTime, const double totalEnergy) const
{
	using time_cast = SolidStatePhysics::UnitSystem::time_cast<SolidStatePhysics::UnitSystem::TimeUnit::Femtosecond, SolidStatePhysics::UnitSystem::TimeUnit::AtomicUnit>;

	std::string outputString;
	{
		outputString += std::to_string(totalAtomCount);
		outputString += "\n   time=    ";
		outputString += std::to_string(time_cast{}(elapsedTime));
		outputString += " (fs)     ";
		outputString += "Energy= ";
		outputString += std::to_string(totalEnergy);
		outputString += " ()  ";
	}

	return outputString;
}

template <typename A, typename AA>
inline std::string SolidStatePhysics::Materials::IO::MdFileOwner::writePrimitiveLatticeVectors(const CrystalStructure<A, AA>& crystalStructure) const
{
	std::string outputString;
	{
		using length_cast = SolidStatePhysics::UnitSystem::length_cast<SolidStatePhysics::UnitSystem::LengthUnit::Angstrom, SolidStatePhysics::UnitSystem::LengthUnit::AtomicUnit>;

		outputString += "Cell_Vectors=";

		for (std::size_t columnIndex = 0; columnIndex < 3; ++columnIndex)
		{
			outputString += "  ";
			outputString += std::to_string(length_cast{}(crystalStructure.crystalLattice().primitiveLatticeVectors()(0, columnIndex)));
			outputString += "  ";
			outputString += std::to_string(length_cast{}(crystalStructure.crystalLattice().primitiveLatticeVectors()(1, columnIndex)));
			outputString += "  ";
			outputString += std::to_string(length_cast{}(crystalStructure.crystalLattice().primitiveLatticeVectors()(2, columnIndex)));
		}

		outputString += "\n";
	}

	return outputString;
}

template <typename A, typename AA>
inline std::string SolidStatePhysics::Materials::IO::MdFileOwner::writeAtomicArrangement(const CrystalStructure<A, AA>& crystalStructure) const
{
	std::string outputString;
	{
		using length_cast = SolidStatePhysics::UnitSystem::length_cast<SolidStatePhysics::UnitSystem::LengthUnit::Angstrom, SolidStatePhysics::UnitSystem::LengthUnit::AtomicUnit>;
		using speed_cast = SolidStatePhysics::UnitSystem::speed_cast<SolidStatePhysics::UnitSystem::VelocityUnit::MetrePerSecond, SolidStatePhysics::UnitSystem::VelocityUnit::AtomicUnit>;

		for (const auto& atom : crystalStructure.atomicArrangement())
		{
			outputString += "   ";
			outputString += SolidStatePhysics::ChemicalElements::PeriodicTable::replyElementSymbol(atom.atomicNumber());
			outputString += "  ";
			outputString += std::to_string(length_cast{}(atom.cartesianCoordinate()[0]));
			outputString += "  ";
			outputString += std::to_string(length_cast{}(atom.cartesianCoordinate()[1]));
			outputString += "  ";
			outputString += std::to_string(length_cast{}(atom.cartesianCoordinate()[2]));
			outputString += "  ";
			outputString += std::to_string(atom.appliedForce()[0]);
			outputString += "  ";
			outputString += std::to_string(atom.appliedForce()[1]);
			outputString += "  ";
			outputString += std::to_string(atom.appliedForce()[2]);
			outputString += "  ";
			outputString += "0.0";
			outputString += "  ";
			outputString += "0.0";
			outputString += "  ";
			outputString += "0.0";
			outputString += "  0.0  0.0  0.0  0.0\n";
		}
	}

	return outputString;
}

// Private methods
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************

#endif // !SOLIDSTATEPHYSICS_MATERIALS_IO_MDFILEOWNER_H
