#ifndef SOLIDSTATEPHYSICS_CRYSTALLIZATION_CALCULATORS_MINIMUMCONSTRAINTER_H
#define SOLIDSTATEPHYSICS_CRYSTALLIZATION_CALCULATORS_MINIMUMCONSTRAINTER_H

#include <memory>

#include "InteratomicConnector.h"
#include "IConstrainter.h"


namespace SolidStatePhysics
{
	namespace Crystallization
	{
		namespace Calculators
		{
			class MinimumConstrainter :public InteratomicConnector, public IConstrainter
			{
// **********************************************************************************************************************************************************************************************************************************************************************************************
			// Protected constructors and public destructor

			protected:
				MinimumConstrainter(InteractingAtom& ownerAtom, AtomicAdaptor&& attachedAtom);
				explicit MinimumConstrainter(const std::shared_ptr<MinimumConstrainter>& twinPtr);

			public:
				virtual ~MinimumConstrainter() = default;

			// Protected constructors and public destructor
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
			// Static methods

				static double getDefaultRepulsiveForceConstant() noexcept;
				static void resetRepulsiveForceConstant(const double repulsiveForceConstant);

			// Static methods
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
			// Public methods

				void initialize() override;
				double minimumDistance() const noexcept;

				double getPenaltyEnergy() const override;
				void applyForce() override;

			// Public methods
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
			// Private methods

			protected:
				void renewMinimumDistance(const double minimumDistance);

			// Private methods
// **********************************************************************************************************************************************************************************************************************************************************************************************

			private:
				double m_minimumDistance;

				static double m_repulsiveForceConstant;


			private:
				MinimumConstrainter() = delete;
				MinimumConstrainter(const MinimumConstrainter&) = delete;
				MinimumConstrainter(MinimumConstrainter&&) = delete;
				MinimumConstrainter& operator=(const MinimumConstrainter&) = delete;
				MinimumConstrainter& operator=(MinimumConstrainter&&) = delete;
			};
		}
	}
}

// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// Static methods

inline double SolidStatePhysics::Crystallization::Calculators::MinimumConstrainter::getDefaultRepulsiveForceConstant() noexcept
{
	return (-100.0);
}

inline void SolidStatePhysics::Crystallization::Calculators::MinimumConstrainter::resetRepulsiveForceConstant(const double repulsiveForceConstant)
{
	if (repulsiveForceConstant < 0)
		m_repulsiveForceConstant = repulsiveForceConstant;

	else
	{
		std::string errorMessage = "------------------------------------------------------\n\n";
		errorMessage += "Error   : Invalid argument\n";
		errorMessage += "Location: The function of \"resetRepulsiveForceConstant\" in the class of SolidStatePhysics::Crystallization::Calculators::MinimumConstrainter.\n";
		errorMessage += "Message : \"repulsiveForceConstant\" is not less than zero.\n\n";
		errorMessage += "------------------------------------------------------\n\n";
		throw std::invalid_argument(errorMessage);
	}
}

// Static methods
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// Public methods

inline void SolidStatePhysics::Crystallization::Calculators::MinimumConstrainter::initialize()
{
	m_minimumDistance = 0.0;
}

inline double SolidStatePhysics::Crystallization::Calculators::MinimumConstrainter::minimumDistance() const noexcept
{
	return m_minimumDistance;
}

inline double SolidStatePhysics::Crystallization::Calculators::MinimumConstrainter::getPenaltyEnergy() const
{
	if (interatomicDistance() < m_minimumDistance)
		return (m_repulsiveForceConstant * (m_minimumDistance - interatomicDistance()));
	else
		return 0.0;
}

inline void SolidStatePhysics::Crystallization::Calculators::MinimumConstrainter::applyForce()
{
	if (interatomicDistance() < m_minimumDistance)
	{
		double forceScalingMultiplier = m_repulsiveForceConstant / interatomicDistance();

		ownerAtom().receiveForce(forceScalingMultiplier, displacement());
		attachedAtom().entity().receiveForce((-1.0 * forceScalingMultiplier), displacement());

		NumericalAnalysis::LinearAlgebra::NumericMatrix additiveTransformationTensor(0.0, 3, 3);
		{
			NumericVector fractionalDisplacement = crystalLattice().toFractionalCoordinate(displacement());
			additiveTransformationTensor[0] -= displacement() * (forceScalingMultiplier * fractionalDisplacement[0]);
			additiveTransformationTensor[1] -= displacement() * (forceScalingMultiplier * fractionalDisplacement[1]);
			additiveTransformationTensor[2] -= displacement() * (forceScalingMultiplier * fractionalDisplacement[2]);

			crystalLattice().addTransformationTensor(1.0, additiveTransformationTensor);
		}
	}
}

// Public methods
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// Protected methods

inline void SolidStatePhysics::Crystallization::Calculators::MinimumConstrainter::renewMinimumDistance(const double minimumDistance)
{
	m_minimumDistance = minimumDistance;

	std::shared_ptr<MinimumConstrainter> twinMinimumConstrainterPtr = std::dynamic_pointer_cast<MinimumConstrainter>(twinPtr());
	twinMinimumConstrainterPtr->m_minimumDistance = m_minimumDistance;
}

// Protected methods
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************

#endif // !SOLIDSTATEPHYSICS_CRYSTALLIZATION_CALCULATORS_MINIMUMCONSTRAINTER_H
