﻿#ifndef NUMERICALANALYSIS_LINEARALGEBRA_NUMERICMATRIX_H
#define NUMERICALANALYSIS_LINEARALGEBRA_NUMERICMATRIX_H

#include <valarray>
#include <vector>
#include <iterator>
#include <string>
#include <stdexcept>
#include <utility>
#include <memory>

#include <mkl_cblas.h>
#include <mkl_lapacke.h>
#include <mkl_trans.h>

#include "NumericMatrixTraits.h"


namespace NumericalAnalysis
{
	namespace LinearAlgebra
	{
		class NumericVector;
		class NumericVectorSlice;

		class NumericMatrix final
		{
			/*
			Column Major Layout for m_columnSize-by-m_rowSize matrix
			*/

		public:
			using size_type = std::size_t;
			using value_type = double;
			using pointer = value_type*;
			using const_pointer = const value_type*;

// **********************************************************************************************************************************************************************************************************************************************************************************************
		// Constructors and destructor

		public:
			NumericMatrix() = default;
			NumericMatrix(std::initializer_list<NumericVector>);
			explicit NumericMatrix(const std::vector<NumericVector>&);
			NumericMatrix(const size_type columnSize, const size_type rowSize);
			NumericMatrix(const value_type, const size_type columnSize, const size_type rowSize);

			NumericMatrix(const NumericMatrix&);
			NumericMatrix(NumericMatrix&&) noexcept = default;

			~NumericMatrix() {}

		// Constructors and destructor
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
		// Arithmetic functions

			NumericMatrix& add(const value_type scalingMultiplier, const NumericMatrix&) noexcept;
			NumericMatrix& add(const value_type scalingMultiplier, const NumericMatrix&, const value_type selfScalingMultiplier) noexcept;
			NumericMatrix& add(const value_type scalingMultiplier, const NumericVector&, const NumericVector&) noexcept;

		// Arithmetic functions
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
		// Unary operators

			NumericMatrix& operator=(const value_type scalar) noexcept;
			NumericMatrix& operator=(const NumericMatrix&) noexcept;
			NumericMatrix& operator=(NumericMatrix&&) noexcept = default;

			NumericMatrix& operator+=(const value_type scalar) noexcept;
			NumericMatrix& operator+=(const NumericMatrix&) noexcept;
			NumericMatrix& operator-=(const value_type scalar) noexcept;
			NumericMatrix& operator-=(const NumericMatrix&) noexcept;
			NumericMatrix& operator*=(const value_type scalingMultiplier) noexcept;
			NumericMatrix& operator/=(const value_type scalingDivider);

		// Unary operators
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
		// Access functions

			value_type operator()(const size_type rowIndex, const size_type columnIndex) const noexcept;
			value_type& operator()(const size_type rowIndex, const size_type columnIndex) noexcept;

			value_type at(const size_type rowIndex, const size_type columnIndex) const;
			value_type& at(const size_type rowIndex, const size_type columnIndex);

			pointer begin() noexcept;
			const_pointer begin() const noexcept;
			const_pointer cbegin() const noexcept;
			pointer end() noexcept;
			const_pointer end() const noexcept;
			const_pointer cend() const noexcept;

		// Access functions
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
		// Slice functions

			NumericVector operator[](const size_type columnIndex) const noexcept;
			NumericVectorSlice operator[](const size_type columnIndex) noexcept;

			NumericVector columnVector(const size_type columnIndex) const;
			NumericVectorSlice columnVectorSlice(const size_type columnIndex);

			NumericVector rowVector(const size_type rowIndex) const;
			NumericVectorSlice rowVectorSlice(const size_type rowIndex);

			NumericVector replicateAsNumericVector() const noexcept;
			NumericVector migrateToNumericVector() noexcept;
			NumericVectorSlice asNumericVectorSlice() noexcept;

		// Slice functions
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
		// Utility

			size_type columnSize() const noexcept;
			size_type rowSize() const noexcept;
			size_type size() const noexcept;
			void resize(const size_type columnSize, const size_type rowSize, const value_type value = value_type{});
			void swap(NumericMatrix&) noexcept;
			bool empty() const noexcept;
			void clear() noexcept;

		// Utility
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
		// Customize functions

			void insertColumn(const NumericVector& numericColumnVectors, const size_type insertedColumnIndex);
			void insertColumns(const NumericMatrix& numericMatrix, const size_type insertedColumnIndex);
			void pushfrontColumn(const NumericVector& numericColumnVector);
			void pushfrontColumns(const NumericMatrix& numericMatrix);
			void pushbackColumn(const NumericVector& numericColumnVector);
			void pushbackColumns(const NumericMatrix& numericMatrix);
						
			void insertRow(const NumericVector& numericRowVectors, const size_type insertedRowIndex);
			void insertRows(const NumericMatrix& numericMatrix, const size_type insertedRowIndex);

			void eraseColumns(std::initializer_list<size_type> columnIndices);
			void eraseRows(std::initializer_list<size_type> rowIndices);

			bool invert(); // If the inversion is failed, all the data of the matrix is erased.
			void transpose(const value_type scalingMultiplier = 1.0) noexcept;
			void conjugate(const value_type scalingMultiplier = 1.0) noexcept;
			void conjugateTranspose(const value_type scalingMultiplier = 1.0) noexcept;

			NumericMatrix inverseMatrix() const;

		// Customize functions
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
		// Friends and private methods

			friend class NumericVector;

			private:
				void insertColumn(const size_type insertedColumnIndex, const std::valarray<value_type>&, const size_type additionalRowSize);
				void insertRow(const size_type insertedRowIndex, const std::valarray<value_type>&, const size_type additionalColumnSize);

		// Friends and private methods
// **********************************************************************************************************************************************************************************************************************************************************************************************

		private:
			std::valarray<value_type> m_numericMatrix;
			size_type m_columnSize{ 0 };
			size_type m_rowSize{ 0 };
		};
	}
}


// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// Arithmetic functions

inline NumericalAnalysis::LinearAlgebra::NumericMatrix& NumericalAnalysis::LinearAlgebra::NumericMatrix::add(const value_type scalingMultiplier, const NumericMatrix& numericMatrix) noexcept
{
	cblas_daxpy(m_numericMatrix.size(), scalingMultiplier, numericMatrix.begin(), 1, std::begin(m_numericMatrix), 1);
	return *this;
}

inline NumericalAnalysis::LinearAlgebra::NumericMatrix& NumericalAnalysis::LinearAlgebra::NumericMatrix::add(const value_type scalingMultiplier, const NumericMatrix& numericMatrix, const value_type selfScalingMultiplier) noexcept
{
	cblas_daxpby(m_numericMatrix.size(), scalingMultiplier, numericMatrix.begin(), 1, selfScalingMultiplier, std::begin(m_numericMatrix), 1);
	return *this;
}

// Arithmetic functions
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// Unary operators

inline NumericalAnalysis::LinearAlgebra::NumericMatrix& NumericalAnalysis::LinearAlgebra::NumericMatrix::operator=(const value_type scalar) noexcept
{
	m_numericMatrix = scalar;
	return *this;
}

inline NumericalAnalysis::LinearAlgebra::NumericMatrix& NumericalAnalysis::LinearAlgebra::NumericMatrix::operator=(const NumericMatrix& numericMatrix) noexcept
{
	m_numericMatrix.resize(numericMatrix.size());
	cblas_dcopy(m_numericMatrix.size(), numericMatrix.begin(), 1, std::begin(m_numericMatrix), 1);
	m_columnSize = numericMatrix.m_columnSize;
	m_rowSize = numericMatrix.m_rowSize;
	return *this;
}

inline NumericalAnalysis::LinearAlgebra::NumericMatrix& NumericalAnalysis::LinearAlgebra::NumericMatrix::operator+=(const value_type scalar) noexcept
{
	m_numericMatrix += scalar;
	return *this;
}

inline NumericalAnalysis::LinearAlgebra::NumericMatrix& NumericalAnalysis::LinearAlgebra::NumericMatrix::operator+=(const NumericMatrix& numericMatrix) noexcept
{
	return add(1.0, numericMatrix);
}

inline NumericalAnalysis::LinearAlgebra::NumericMatrix& NumericalAnalysis::LinearAlgebra::NumericMatrix::operator-=(const value_type scalar) noexcept
{
	m_numericMatrix -= scalar;
	return *this;
}

inline NumericalAnalysis::LinearAlgebra::NumericMatrix& NumericalAnalysis::LinearAlgebra::NumericMatrix::operator-=(const NumericMatrix& numericMatrix) noexcept
{
	return add(-1.0, numericMatrix);
}

inline NumericalAnalysis::LinearAlgebra::NumericMatrix& NumericalAnalysis::LinearAlgebra::NumericMatrix::operator*=(const value_type scalingMultiplier) noexcept
{
	MKL_Dimatcopy('C', 'N', m_columnSize, m_rowSize, scalingMultiplier, std::begin(m_numericMatrix), m_columnSize, m_columnSize);
	return *this;
}

inline NumericalAnalysis::LinearAlgebra::NumericMatrix& NumericalAnalysis::LinearAlgebra::NumericMatrix::operator/=(const value_type scalingDivider)
{
	if (std::isnormal(scalingDivider))
	{
		cblas_dscal(m_numericMatrix.size(), (1.0 / scalingDivider), std::begin(m_numericMatrix), 1);
		return *this;
	}

	else
	{
		std::string errorMessage = "------------------------------------------------------\n\n";
		errorMessage += "Error   : Invalid argument\n";
		errorMessage += "Location: The function of \"operator/=\" in the class of NumericalAnalysis::LinearAlgebra::NumericMatrix.\n";
		errorMessage += "Message : scalingDivider is not a normalized value.\n\n";
		errorMessage += "------------------------------------------------------\n\n";
		throw std::invalid_argument(errorMessage);
	}
}

// Unary operators
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// Access functions

inline NumericalAnalysis::LinearAlgebra::NumericMatrix::value_type NumericalAnalysis::LinearAlgebra::NumericMatrix::operator()(const size_type rowIndex, const size_type columnIndex) const noexcept
{
	return m_numericMatrix[(columnIndex * m_columnSize) + rowIndex];
}

inline NumericalAnalysis::LinearAlgebra::NumericMatrix::value_type& NumericalAnalysis::LinearAlgebra::NumericMatrix::operator()(const size_type rowIndex, const size_type columnIndex) noexcept
{
	return m_numericMatrix[(columnIndex * m_columnSize) + rowIndex];
}

inline NumericalAnalysis::LinearAlgebra::NumericMatrix::value_type NumericalAnalysis::LinearAlgebra::NumericMatrix::at(const size_type rowIndex, const size_type columnIndex) const
{
	if (rowIndex < m_columnSize && columnIndex < m_rowSize)
	{
		return m_numericMatrix[(columnIndex * m_columnSize) + rowIndex];
	}

	else
	{
		std::string errorMessage = "------------------------------------------------------\n\n";
		errorMessage += "Error   : Out of range\n";
		errorMessage += "Location: The function of \"at\" in the class of NumericalAnalysis::LinearAlgebra::NumericMatrix.\n";
		errorMessage += "Message : rowIndex and/or columnIndex are out of range.\n\n";
		errorMessage += "------------------------------------------------------\n\n";
		throw std::out_of_range(errorMessage);
	}
}

inline NumericalAnalysis::LinearAlgebra::NumericMatrix::value_type& NumericalAnalysis::LinearAlgebra::NumericMatrix::at(const size_type rowIndex, const size_type columnIndex)
{
	if (rowIndex < m_columnSize && columnIndex < m_rowSize)
	{
		return m_numericMatrix[(columnIndex * m_columnSize) + rowIndex];
	}

	else
	{
		std::string errorMessage = "------------------------------------------------------\n\n";
		errorMessage += "Error   : Out of range\n";
		errorMessage += "Location: The function of \"at\" in the class of NumericalAnalysis::LinearAlgebra::NumericMatrix.\n";
		errorMessage += "Message : rowIndex and/or columnIndex are out of range.\n\n";
		errorMessage += "------------------------------------------------------\n\n";
		throw std::out_of_range(errorMessage);
	}
}

inline NumericalAnalysis::LinearAlgebra::NumericMatrix::pointer NumericalAnalysis::LinearAlgebra::NumericMatrix::begin() noexcept
{
	return std::begin(m_numericMatrix);
}

inline NumericalAnalysis::LinearAlgebra::NumericMatrix::const_pointer NumericalAnalysis::LinearAlgebra::NumericMatrix::begin() const noexcept
{
	return std::begin(m_numericMatrix);
}

inline NumericalAnalysis::LinearAlgebra::NumericMatrix::const_pointer NumericalAnalysis::LinearAlgebra::NumericMatrix::cbegin() const noexcept
{
	return std::begin(m_numericMatrix);
}

inline NumericalAnalysis::LinearAlgebra::NumericMatrix::pointer NumericalAnalysis::LinearAlgebra::NumericMatrix::end() noexcept
{
	return std::end(m_numericMatrix);
}

inline NumericalAnalysis::LinearAlgebra::NumericMatrix::const_pointer NumericalAnalysis::LinearAlgebra::NumericMatrix::end() const noexcept
{
	return std::end(m_numericMatrix);
}

inline NumericalAnalysis::LinearAlgebra::NumericMatrix::const_pointer NumericalAnalysis::LinearAlgebra::NumericMatrix::cend() const noexcept
{
	return std::end(m_numericMatrix);
}

// Access functions
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// Utility

inline NumericalAnalysis::LinearAlgebra::NumericMatrix::size_type NumericalAnalysis::LinearAlgebra::NumericMatrix::columnSize() const noexcept
{
	return m_columnSize;
}

inline NumericalAnalysis::LinearAlgebra::NumericMatrix::size_type NumericalAnalysis::LinearAlgebra::NumericMatrix::rowSize() const noexcept
{
	return m_rowSize;
}

inline NumericalAnalysis::LinearAlgebra::NumericMatrix::size_type NumericalAnalysis::LinearAlgebra::NumericMatrix::size() const noexcept
{
	return (m_columnSize * m_rowSize);
}

inline void NumericalAnalysis::LinearAlgebra::NumericMatrix::resize(const size_type columnSize, const size_type rowSize, const value_type value)
{
	m_columnSize = columnSize;
	m_rowSize = rowSize;
	m_numericMatrix.resize((m_columnSize * m_rowSize), value);
}

inline void NumericalAnalysis::LinearAlgebra::NumericMatrix::swap(NumericMatrix& numericMatrix) noexcept
{
	std::swap(m_columnSize, numericMatrix.m_columnSize);
	std::swap(m_rowSize, numericMatrix.m_rowSize);
	m_numericMatrix.swap(numericMatrix.m_numericMatrix);
}

inline bool NumericalAnalysis::LinearAlgebra::NumericMatrix::empty() const noexcept
{
	return !(static_cast<bool>(m_numericMatrix.size()));
}

inline void NumericalAnalysis::LinearAlgebra::NumericMatrix::clear() noexcept
{
	std::valarray<value_type> va;
	m_numericMatrix.swap(va);
	m_columnSize = 0;
	m_rowSize = 0;
}

// Utility
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// Customize functions

inline void NumericalAnalysis::LinearAlgebra::NumericMatrix::pushfrontColumn(const NumericVector& numericColumnVector)
{
	insertColumn(numericColumnVector, 0);
}

inline void NumericalAnalysis::LinearAlgebra::NumericMatrix::pushfrontColumns(const NumericMatrix& numericMatrix)
{
	insertColumns(numericMatrix, 0);
}

inline void NumericalAnalysis::LinearAlgebra::NumericMatrix::pushbackColumn(const NumericVector& numericColumnVector)
{
	insertColumn(numericColumnVector, m_rowSize);
}

inline void NumericalAnalysis::LinearAlgebra::NumericMatrix::pushbackColumns(const NumericMatrix& numericMatrix)
{
	insertColumns(numericMatrix, m_rowSize);
}

inline bool NumericalAnalysis::LinearAlgebra::NumericMatrix::invert()
{
	if (m_columnSize == m_rowSize)
	{
		std::unique_ptr<lapack_int> ipiv{ new lapack_int[m_columnSize] };

		if (LAPACKE_dgetrf(LAPACK_COL_MAJOR, m_columnSize, m_columnSize, std::begin(m_numericMatrix), m_columnSize, ipiv.get()) == 0)
		{
			if (LAPACKE_dgetri(LAPACK_COL_MAJOR, m_columnSize, std::begin(m_numericMatrix), m_columnSize, ipiv.get()) == 0)
				return true;
		}
	}

	clear();
	return false;
}

inline void NumericalAnalysis::LinearAlgebra::NumericMatrix::transpose(const value_type scalingMultiplier) noexcept
{
	MKL_Dimatcopy('C', 'T', m_columnSize, m_rowSize, scalingMultiplier, std::begin(m_numericMatrix), m_columnSize, m_rowSize);
	std::swap(m_columnSize, m_rowSize);
}

inline void NumericalAnalysis::LinearAlgebra::NumericMatrix::conjugate(const value_type scalingMultiplier) noexcept
{
	MKL_Dimatcopy('C', 'R', m_columnSize, m_rowSize, scalingMultiplier, std::begin(m_numericMatrix), m_columnSize, m_columnSize);
	std::swap(m_columnSize, m_rowSize);
}

inline void NumericalAnalysis::LinearAlgebra::NumericMatrix::conjugateTranspose(const value_type scalingMultiplier) noexcept
{
	MKL_Dimatcopy('C', 'C', m_columnSize, m_rowSize, scalingMultiplier, std::begin(m_numericMatrix), m_columnSize, m_rowSize);
	std::swap(m_columnSize, m_rowSize);
}

inline NumericalAnalysis::LinearAlgebra::NumericMatrix NumericalAnalysis::LinearAlgebra::NumericMatrix::inverseMatrix() const
{
	NumericMatrix resultMatrix;
	resultMatrix.m_numericMatrix = m_numericMatrix;
	resultMatrix.m_columnSize = m_columnSize;
	resultMatrix.m_rowSize = m_rowSize;

	if (resultMatrix.invert())
		return resultMatrix;

	else
	{
		std::string errorMessage = "------------------------------------------------------\n\n";
		errorMessage += "Error   : Logic error\n";
		errorMessage += "Location: The member function of \"inverseMatrix\" in the class of NumericalAnalysis::LinearAlgebra::NumericMatrix.\n";
		errorMessage += "Message : Could not create inverse matrix.\n\n";
		errorMessage += "------------------------------------------------------\n\n";
		throw std::logic_error(errorMessage);
	}
}

// Customize functions
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************


#endif // !NUMERICALANALYSIS_LINEARALGEBRA_NUMERICMATRIX_H
