#ifndef SOLIDSTATEPHYSICS_CRYSTALLIZATION_COMPONENTS_PERIODICLIQUID_H
#define SOLIDSTATEPHYSICS_CRYSTALLIZATION_COMPONENTS_PERIODICLIQUID_H

#include <stdexcept>
#include <string>
#include <type_traits>
#include <list>
#include <memory>

#include "CrystalStructure.h"

#include "IntelligentAtom.h"
#include "ChemicalBond.h"
#include "IonicConstrainter.h"
#include "MetallicConstrainter.h"
#include "PolyhedralConstrainter.h"
#include "SecondNeighborConstrainter.h"


namespace SolidStatePhysics
{
	namespace Crystallization
	{
		namespace Components
		{
			class PeriodicLiquid :public SolidStatePhysics::Materials::CrystalStructure<IntelligentAtom>
			{
				using CrystalLattice = SolidStatePhysics::Materials::CrystalLattice;
				using CrystalStructure = SolidStatePhysics::Materials::CrystalStructure<IntelligentAtom>;
				using AtomicAdaptor = SolidStatePhysics::Crystallization::Calculators::TranslatingAdaptor<IntelligentAtom>;

// **********************************************************************************************************************************************************************************************************************************************************************************************
			// Constructor and destructor

			public:
				PeriodicLiquid();
				PeriodicLiquid(const CrystalLattice& crystalLattice, const AtomicArrangement& atomicArrangement = AtomicArrangement{}, const std::string& name = std::string{});
				PeriodicLiquid(const CrystalLattice& crystalLattice, AtomicArrangement&& atomicArrangement, const std::string& name = std::string{});

				template <typename A, typename AA>
				explicit PeriodicLiquid(const SolidStatePhysics::Materials::CrystalStructure<A, AA>&);

				virtual ~PeriodicLiquid() = default;

			// Constructor and destructor
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
			// Public methods

				virtual void initialize() noexcept;
				virtual void update();
				virtual void renew();
				void clear() noexcept override;

				size_type chemicalBonding() const noexcept;
				void applyInteratomicForces();

				static double getDefaultPressure() noexcept;
				void resetPressure(const double);

			// Public methods
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
			// Customize methods

				void renewCrystalLatticeRandomly(const double minLength, const double maxLength) override;

				void normalize() noexcept override;
				void normalizeCrystalLattice() noexcept override;
				void normalizeFractionalCoordinates() noexcept override;

				void link(const std::shared_ptr<ChemicalBond>&);
				void link(const std::shared_ptr<IonicConstrainter>&);
				void link(const std::shared_ptr<MetallicConstrainter>&);
				void link(const std::shared_ptr<PolyhedralConstrainter>&);
				void link(const std::shared_ptr<SecondNeighborConstrainter>&);

				void clearChemicalBonds() noexcept;
				void clearIonicConstrainters() noexcept;
				void clearMetallicConstrainters() noexcept;
				void clearPolyhedralConstrainters() noexcept;
				void clearSecondNeighborConstrainters() noexcept;

			// Customize methods
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
			// Protected methods

			protected:
				void applyPressure();
				void cleanGeometricalConstrainters();
				void convertIntoBasicConstraints();

				bool hasInfeasibleChemicalBonds(const double maxBondingLengthErrorRate);
				bool hasInfeasibleConstrainters(const double maxBondingLengthErrorRate);

				std::list<AtomicAdaptor::LatticePoint> getVisibleLatticePoints(const IntelligentAtom& originAtom, const IntelligentAtom& targetAtom) const noexcept;

			// Protected methods
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
			// Private methods

			private:
				template <typename C>
				void applyForces(std::list<std::weak_ptr<C>>& constraintersPtrs);

				template <typename C>
				void updateConstrainters(std::list<std::weak_ptr<C>>& constraintersPtrs);

				template <typename C>
				void renewConstrainters(std::list<std::weak_ptr<C>>& constraintersPtrs);

				template <typename C>
				void eraseExpiredConstrainters(std::list<std::weak_ptr<C>>& constraintersPtrs);

				template <typename C>
				void convertIntoBasicConstraints(std::list<std::weak_ptr<C>>& constraintersPtrs);

				template <typename C>
				bool hasInfeasibleConstrainters(std::list<std::weak_ptr<C>>& constraintersPtrs, const double maxBondingLengthErrorRate);

			// Private methods
// **********************************************************************************************************************************************************************************************************************************************************************************************

			protected:
				std::list<std::weak_ptr<ChemicalBond>> m_chemicalBondsPtrs;
				std::list<std::weak_ptr<IonicConstrainter>> m_ionicConstraintersPtrs;
				std::list<std::weak_ptr<MetallicConstrainter>> m_metallicConstraintersPtrs;
				std::list<std::weak_ptr<PolyhedralConstrainter>> m_polyhedralConstraintersPtrs;
				std::list<std::weak_ptr<SecondNeighborConstrainter>> m_secondNeighborConstraintersPtrs;

			private:
				double m_pressure{ 1.0 };


			private:
				PeriodicLiquid(const PeriodicLiquid&) = delete;
				PeriodicLiquid(PeriodicLiquid&&) noexcept = delete;
				PeriodicLiquid& operator=(const PeriodicLiquid&) = delete;
				PeriodicLiquid& operator=(PeriodicLiquid&&) noexcept = delete;
			};
		}
	}
}

// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// Constructor

template <typename A, typename AA>
inline SolidStatePhysics::Crystallization::Components::PeriodicLiquid::PeriodicLiquid(const SolidStatePhysics::Materials::CrystalStructure<A, AA>& crystalStructure)
	: CrystalStructure{ crystalStructure }
	, m_chemicalBondsPtrs{}
	, m_ionicConstraintersPtrs{}
	, m_metallicConstraintersPtrs{}
	, m_polyhedralConstraintersPtrs{}
	, m_secondNeighborConstraintersPtrs{}
{
}

// Constructor
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// Public methods

inline void SolidStatePhysics::Crystallization::Components::PeriodicLiquid::initialize() noexcept
{
	clearChemicalBonds();
	clearIonicConstrainters();
	clearMetallicConstrainters();
	clearPolyhedralConstrainters();
	clearSecondNeighborConstrainters();

	normalizeFractionalCoordinates();

	for (auto& intelligentAtom : atomicArrangement())
		intelligentAtom.initialize();
}

inline void SolidStatePhysics::Crystallization::Components::PeriodicLiquid::update()
{
	updateConstrainters(m_chemicalBondsPtrs);
	updateConstrainters(m_ionicConstraintersPtrs);
	updateConstrainters(m_metallicConstraintersPtrs);
	updateConstrainters(m_polyhedralConstraintersPtrs);
	updateConstrainters(m_secondNeighborConstraintersPtrs);
}

inline void SolidStatePhysics::Crystallization::Components::PeriodicLiquid::renew()
{
	renewConstrainters(m_chemicalBondsPtrs);
	renewConstrainters(m_ionicConstraintersPtrs);
	renewConstrainters(m_metallicConstraintersPtrs);
	renewConstrainters(m_polyhedralConstraintersPtrs);
	renewConstrainters(m_secondNeighborConstraintersPtrs);
}

inline void SolidStatePhysics::Crystallization::Components::PeriodicLiquid::clear() noexcept
{
	clearChemicalBonds();
	clearIonicConstrainters();
	clearMetallicConstrainters();
	clearPolyhedralConstrainters();
	clearSecondNeighborConstrainters();

	CrystalStructure::clear();
}

inline void SolidStatePhysics::Crystallization::Components::PeriodicLiquid::applyInteratomicForces()
{
	applyForces(m_chemicalBondsPtrs);
	applyForces(m_ionicConstraintersPtrs);
	applyForces(m_metallicConstraintersPtrs);
	applyForces(m_polyhedralConstraintersPtrs);
	applyForces(m_secondNeighborConstraintersPtrs);
}

inline SolidStatePhysics::Crystallization::Components::PeriodicLiquid::size_type SolidStatePhysics::Crystallization::Components::PeriodicLiquid::chemicalBonding() const noexcept
{
	return m_chemicalBondsPtrs.size();
}

inline double SolidStatePhysics::Crystallization::Components::PeriodicLiquid::getDefaultPressure() noexcept
{
	return 1.0;
}

inline void SolidStatePhysics::Crystallization::Components::PeriodicLiquid::resetPressure(const double pressure)
{
	if (pressure < 0.0)
	{
		std::string errorMessage = "------------------------------------------------------\n\n";
		errorMessage += "Error   : Logic error\n";
		errorMessage += "Location: The member function of \"getStructuralIndex\" in the class of SolidStatePhysics::Crystallization::Analysis::StructuralDictionary.\n";
		errorMessage += "Message : \"pressure\" is less than zero.\n\n";
		errorMessage += "------------------------------------------------------\n\n";
		throw std::logic_error(errorMessage);
	}

	else
		m_pressure = pressure;
}

// Public methods
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// Customize methods

inline void SolidStatePhysics::Crystallization::Components::PeriodicLiquid::normalize() noexcept
{
	clearChemicalBonds();
	clearIonicConstrainters();
	clearMetallicConstrainters();
	clearPolyhedralConstrainters();
	clearSecondNeighborConstrainters();

	CrystalStructure::normalize();
}

inline void SolidStatePhysics::Crystallization::Components::PeriodicLiquid::normalizeCrystalLattice() noexcept
{
	clearChemicalBonds();
	clearIonicConstrainters();
	clearMetallicConstrainters();
	clearPolyhedralConstrainters();
	clearSecondNeighborConstrainters();

	CrystalStructure::normalizeCrystalLattice();
}

inline void SolidStatePhysics::Crystallization::Components::PeriodicLiquid::normalizeFractionalCoordinates() noexcept
{
	clearChemicalBonds();
	clearIonicConstrainters();
	clearMetallicConstrainters();
	clearPolyhedralConstrainters();
	clearSecondNeighborConstrainters();

	CrystalStructure::normalizeFractionalCoordinates();
}

inline void SolidStatePhysics::Crystallization::Components::PeriodicLiquid::link(const std::shared_ptr<ChemicalBond>& chemicalBondPtr)
{
	m_chemicalBondsPtrs.push_back(std::weak_ptr<ChemicalBond>{ chemicalBondPtr });
}

inline void SolidStatePhysics::Crystallization::Components::PeriodicLiquid::link(const std::shared_ptr<IonicConstrainter>& ionicConstrainterPtr)
{
	m_ionicConstraintersPtrs.push_back(std::weak_ptr<IonicConstrainter>{ ionicConstrainterPtr });
}

inline void SolidStatePhysics::Crystallization::Components::PeriodicLiquid::link(const std::shared_ptr<MetallicConstrainter>& metallicConstrainterPtr)
{
	m_metallicConstraintersPtrs.push_back(std::weak_ptr<MetallicConstrainter>{ metallicConstrainterPtr });
}

inline void SolidStatePhysics::Crystallization::Components::PeriodicLiquid::link(const std::shared_ptr<PolyhedralConstrainter>& polyhedralConstrainterPtr)
{
	m_polyhedralConstraintersPtrs.push_back(std::weak_ptr<PolyhedralConstrainter>{ polyhedralConstrainterPtr });
}

inline void SolidStatePhysics::Crystallization::Components::PeriodicLiquid::link(const std::shared_ptr<SecondNeighborConstrainter>& secondNeighborConstrainterPtr)
{
	m_secondNeighborConstraintersPtrs.push_back(std::weak_ptr<SecondNeighborConstrainter>{ secondNeighborConstrainterPtr });
}

inline void SolidStatePhysics::Crystallization::Components::PeriodicLiquid::clearChemicalBonds() noexcept
{
	for (auto& atom : atomicArrangement())
		atom.clearOwingChemicalBonds();

	m_chemicalBondsPtrs.clear();
}

inline void SolidStatePhysics::Crystallization::Components::PeriodicLiquid::clearIonicConstrainters() noexcept
{
	for (auto& atom : atomicArrangement())
		atom.clearOwingIonicConstrainters();

	m_ionicConstraintersPtrs.clear();
}

inline void SolidStatePhysics::Crystallization::Components::PeriodicLiquid::clearMetallicConstrainters() noexcept
{
	for (auto& atom : atomicArrangement())
		atom.clearOwingMetallicConstrainters();

	m_metallicConstraintersPtrs.clear();
}

inline void SolidStatePhysics::Crystallization::Components::PeriodicLiquid::clearPolyhedralConstrainters() noexcept
{
	for (auto& atom : atomicArrangement())
		atom.clearOwingPolyhedralConstrainters();

	m_polyhedralConstraintersPtrs.clear();
}

inline void SolidStatePhysics::Crystallization::Components::PeriodicLiquid::clearSecondNeighborConstrainters() noexcept
{
	for (auto& atom : atomicArrangement())
		atom.clearOwingSecondNeighborConstrainters();

	m_secondNeighborConstraintersPtrs.clear();
}

// Customize methods
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// Protected methods

inline void SolidStatePhysics::Crystallization::Components::PeriodicLiquid::applyPressure()
{
	const NumericMatrix& latticeVectors = crystalLattice().primitiveLatticeVectors();
	NumericMatrix transformationTensor(3, 3);
	{
		transformationTensor(0, 0) = latticeVectors(2, 1) * latticeVectors(1, 2) - latticeVectors(1, 1) * latticeVectors(2, 2);
		transformationTensor(1, 0) = latticeVectors(0, 1) * latticeVectors(2, 2) - latticeVectors(2, 1) * latticeVectors(0, 2);
		transformationTensor(2, 0) = latticeVectors(1, 1) * latticeVectors(0, 2) - latticeVectors(0, 1) * latticeVectors(1, 2);

		transformationTensor(0, 1) = latticeVectors(2, 2) * latticeVectors(1, 0) - latticeVectors(1, 2) * latticeVectors(2, 0);
		transformationTensor(1, 1) = latticeVectors(0, 2) * latticeVectors(2, 0) - latticeVectors(2, 2) * latticeVectors(0, 0);
		transformationTensor(2, 1) = latticeVectors(1, 2) * latticeVectors(0, 0) - latticeVectors(0, 2) * latticeVectors(1, 0);

		transformationTensor(0, 2) = latticeVectors(2, 0) * latticeVectors(1, 1) - latticeVectors(1, 0) * latticeVectors(2, 1);
		transformationTensor(1, 2) = latticeVectors(0, 0) * latticeVectors(2, 1) - latticeVectors(2, 0) * latticeVectors(0, 1);
		transformationTensor(2, 2) = latticeVectors(1, 0) * latticeVectors(0, 1) - latticeVectors(0, 0) * latticeVectors(1, 1);
	}

	crystalLattice().addTransformationTensor(m_pressure, transformationTensor);
}

inline void SolidStatePhysics::Crystallization::Components::PeriodicLiquid::cleanGeometricalConstrainters()
{
	for (auto& atom : atomicArrangement())
		atom.eraseInactiveConstraintOwners();

	eraseExpiredConstrainters(m_chemicalBondsPtrs);
	eraseExpiredConstrainters(m_ionicConstraintersPtrs);
	eraseExpiredConstrainters(m_metallicConstraintersPtrs);
	eraseExpiredConstrainters(m_polyhedralConstraintersPtrs);
	eraseExpiredConstrainters(m_secondNeighborConstraintersPtrs);
}

inline void SolidStatePhysics::Crystallization::Components::PeriodicLiquid::convertIntoBasicConstraints()
{
	convertIntoBasicConstraints(m_chemicalBondsPtrs);
	convertIntoBasicConstraints(m_polyhedralConstraintersPtrs);
	convertIntoBasicConstraints(m_secondNeighborConstraintersPtrs);
}

inline bool SolidStatePhysics::Crystallization::Components::PeriodicLiquid::hasInfeasibleChemicalBonds(const double maxBondingLengthErrorRate)
{
	return hasInfeasibleConstrainters(m_chemicalBondsPtrs, maxBondingLengthErrorRate);
}

inline bool SolidStatePhysics::Crystallization::Components::PeriodicLiquid::hasInfeasibleConstrainters(const double maxBondingLengthErrorRate)
{
	if (hasInfeasibleConstrainters(m_chemicalBondsPtrs, maxBondingLengthErrorRate))
		return true;

	else
	{
		if (hasInfeasibleConstrainters(m_ionicConstraintersPtrs, maxBondingLengthErrorRate))
			return true;

		else
		{
			if (hasInfeasibleConstrainters(m_metallicConstraintersPtrs, maxBondingLengthErrorRate))
				return true;

			else
			{
				if (hasInfeasibleConstrainters(m_polyhedralConstraintersPtrs, maxBondingLengthErrorRate))
					return true;

				else
				{
					if (hasInfeasibleConstrainters(m_secondNeighborConstraintersPtrs, maxBondingLengthErrorRate))
						return true;

					else
					{
						return false;
					}
				}
			}
		}
	}
}

// Protected methods
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// Private methods

template <typename C>
inline void SolidStatePhysics::Crystallization::Components::PeriodicLiquid::applyForces(std::list<std::weak_ptr<C>>& constraintersPtrs)
{
	for (auto iter = constraintersPtrs.begin(); iter != constraintersPtrs.end();)
	{
		if (std::shared_ptr<C> constrainterPtr = iter->lock())
		{
			constrainterPtr->applyForce();
			++iter;
		}

		else
			iter = constraintersPtrs.erase(iter);
	}
}

template <typename C>
inline void SolidStatePhysics::Crystallization::Components::PeriodicLiquid::updateConstrainters(std::list<std::weak_ptr<C>>& constraintersPtrs)
{
	for (auto iter = constraintersPtrs.begin(); iter != constraintersPtrs.end();)
	{
		if (std::shared_ptr<C> constrainterPtr = iter->lock())
		{
			constrainterPtr->update();
			++iter;
		}

		else
			iter = constraintersPtrs.erase(iter);
	}
}

template <typename C>
inline void SolidStatePhysics::Crystallization::Components::PeriodicLiquid::renewConstrainters(std::list<std::weak_ptr<C>>& constraintersPtrs)
{
	for (auto iter = constraintersPtrs.begin(); iter != constraintersPtrs.end();)
	{
		if (std::shared_ptr<C> constrainterPtr = iter->lock())
		{
			constrainterPtr->renew();
			++iter;
		}

		else
			iter = constraintersPtrs.erase(iter);
	}
}

template <typename C>
inline void SolidStatePhysics::Crystallization::Components::PeriodicLiquid::eraseExpiredConstrainters(std::list<std::weak_ptr<C>>& constraintersPtrs)
{
	for (auto iter = constraintersPtrs.begin(); iter != constraintersPtrs.end();)
	{
		if (std::shared_ptr<C> constrainterPtr = iter->lock())
			++iter;
		else
			iter = constraintersPtrs.erase(iter);
	}
}

template <typename C>
inline void SolidStatePhysics::Crystallization::Components::PeriodicLiquid::convertIntoBasicConstraints(std::list<std::weak_ptr<C>>& constraintersPtrs)
{
	static_assert(!(std::is_same<C, IonicConstrainter>::value), "\"C\" is \"IonicConstrainter\".");
	static_assert(!(std::is_same<C, MetallicConstrainter>::value), "\"C\" is \"MetallicConstrainter\".");

	for (auto iter = constraintersPtrs.begin(); iter != constraintersPtrs.end(); iter = constraintersPtrs.erase(iter))
	{
		if (std::shared_ptr<C> constrainterPtr = iter->lock())
		{
			if (constrainterPtr->ownerAtom().isCation() && constrainterPtr->attachedAtom().entity().isCation())
			{
				if (!(constrainterPtr->ownerAtom().hasMetallicConstrainter(constrainterPtr->attachedAtom())))
					constrainterPtr->ownerAtom().createMetallicConstrainter(constrainterPtr->attachedAtom(), *this);
			}

			else
			{
				if (!(constrainterPtr->ownerAtom().hasIonicConstrainter(constrainterPtr->attachedAtom())))
					constrainterPtr->ownerAtom().createIonicConstrainter(constrainterPtr->attachedAtom(), *this);
			}

			C::erase(std::move(constrainterPtr));
		}
	}
}

template <typename C>
inline bool SolidStatePhysics::Crystallization::Components::PeriodicLiquid::hasInfeasibleConstrainters(std::list<std::weak_ptr<C>>& constraintersPtrs, const double maxBondingLengthErrorRate)
{
	for (auto iter = constraintersPtrs.begin(); iter != constraintersPtrs.end();)
	{
		if (std::shared_ptr<C> constrainterPtr = iter->lock())
		{
			if (constrainterPtr->isFeasible(maxBondingLengthErrorRate))
				++iter;
			else
				return true;
		}

		else
			iter = constraintersPtrs.erase(iter);
	}

	return false;
}

// Private methods
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************

#endif // !SOLIDSTATEPHYSICS_CRYSTALLIZATION_COMPONENTS_PERIODICLIQUID_H
