#ifndef SOLIDSTATEPHYSICS_CRYSTALLIZATION_ANALYSIS_STRUCTURALDICTIONARY_H
#define SOLIDSTATEPHYSICS_CRYSTALLIZATION_ANALYSIS_STRUCTURALDICTIONARY_H

#include <string>
#include <filesystem>
#include <stdexcept>
#include <map>
#include <set>
#include <mutex>

#include "FinalSimulationsReporter.h"


namespace SolidStatePhysics
{
	namespace Crystallization
	{
		namespace Analysis
		{
			class StructuralDictionary
			{
				using size_type = std::size_t;
				using FinalSimulationsReporter = SolidStatePhysics::Crystallization::Simulators::FinalSimulationsReporter;

// **********************************************************************************************************************************************************************************************************************************************************************************************
			// Constructors, destructor, and operators

			public:
				StructuralDictionary() = default;
				StructuralDictionary(const std::filesystem::path& chemicalConstituentsDirectoryPath, const size_type maxGeometricalOptimizing);
				virtual ~StructuralDictionary() = default;

				StructuralDictionary(const StructuralDictionary&) = default;
				StructuralDictionary(StructuralDictionary&&) noexcept = default;
				StructuralDictionary& operator=(const StructuralDictionary&) = default;
				StructuralDictionary& operator=(StructuralDictionary&&) noexcept = default;

			// Constructors, destructor, and operators
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
			// Access methods

				const std::filesystem::path& chemicalConstituentsDirectoryPath() const noexcept;
				size_type crystallizedSamplesCounts() const noexcept;
				size_type failedSamplesCounts() const noexcept;
				size_type exceptionSamplesCounts() const noexcept;
				size_type getStructuralIndex(const std::string& structuralFingerprint) const;
				FinalSimulationsReporter& getFinalSimulationsReporter(const size_type structuralIndex) const;

				bool contains(const std::string& structuralFingerprint) const noexcept;
				bool contains(const size_type structuralIndex) const noexcept;

				void addCrystallizedSamplesCount() noexcept;
				void addFailedSamplesCount() noexcept;
				void addExceptionSamplesCount() noexcept;
				void resetCrystallizedSamplesCount(const size_type crystallizedSamples = 0);
				void registerCrystalStructure(const std::string& structuralFingerprint, const std::map<std::string, size_type>& atomicEnvironments);

			// Access methods
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
			// Access methods

				void initialize() noexcept;
				void clear() noexcept;

				std::string writeFinalReport() const noexcept;
				void createFinalSimulationsReportsPerStructureIfNecessary();

			// Public methods
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
			// Static methods

				static bool needFinalSimulationsReportsPerStructure() noexcept;
				static void resetFinalSimulationsReportsPerStructureNecessity(const bool needFinalSimulationsReportsPerStructure);

			// Static methods
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
			// Private methods

			private:
				std::string getAtomicEnvironmentsName(const std::map<std::string, size_type>& atomicEnvironments);
				void registerAtomicEnvironmentsCount(const size_type structuralIndex, const size_type atomicEnvironmentsCount) noexcept;
				void registerAtomicEnvironmentsName(const size_type structuralIndex, const std::map<std::string, size_type>& atomicEnvironments) noexcept;

				size_type getMaxAtomicEnvironmentsCount() const noexcept;
				std::set<size_type> getStructuralIndices(const size_type atomicEnvironmentsCount) const noexcept;
				std::map<std::string, size_type> countStructuralIndex(const std::set<size_type>& structuralIndices) const;

			// Private methods
// **********************************************************************************************************************************************************************************************************************************************************************************************

			private:
				std::filesystem::path m_chemicalConstituentsDirectoryPath;
				size_type m_maxGeometricalOptimizing;

				static size_type m_crystallizedSamplesCounts;
				static size_type m_failedSamplesCounts;
				static size_type m_exceptionSamplesCounts;
				static std::map<std::string, size_type> m_structuralIndices; // Key is the fingerprint of the structure
				static std::map<size_type, FinalSimulationsReporter> m_finalSimulationsReporters; // Key is the structural index
				static std::map<size_type, size_type> m_atomicEnvironmentsCounts; // Key is the structural index.
				static std::map<size_type, std::string> m_atomicEnvironmentsNames; // Key is the structural index.

				static bool m_needFinalSimulationsReportsPerStructure;
				static std::mutex m_dataMutex;
			};
		}
	}
}

// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// Access methods

inline const std::filesystem::path& SolidStatePhysics::Crystallization::Analysis::StructuralDictionary::chemicalConstituentsDirectoryPath() const noexcept
{
	return m_chemicalConstituentsDirectoryPath;
}

inline SolidStatePhysics::Crystallization::Analysis::StructuralDictionary::size_type SolidStatePhysics::Crystallization::Analysis::StructuralDictionary::crystallizedSamplesCounts() const noexcept
{
	return m_crystallizedSamplesCounts;
}

inline SolidStatePhysics::Crystallization::Analysis::StructuralDictionary::size_type SolidStatePhysics::Crystallization::Analysis::StructuralDictionary::failedSamplesCounts() const noexcept
{
	return m_failedSamplesCounts;
}

inline SolidStatePhysics::Crystallization::Analysis::StructuralDictionary::size_type SolidStatePhysics::Crystallization::Analysis::StructuralDictionary::exceptionSamplesCounts() const noexcept
{
	return m_exceptionSamplesCounts;
}

inline SolidStatePhysics::Crystallization::Analysis::StructuralDictionary::size_type SolidStatePhysics::Crystallization::Analysis::StructuralDictionary::getStructuralIndex(const std::string& structuralFingerprint) const
{
	auto iter = m_structuralIndices.find(structuralFingerprint);

	if (iter != m_structuralIndices.end())
		return iter->second;

	else
	{
		std::string errorMessage = "------------------------------------------------------\n\n";
		errorMessage += "Error   : Logic error\n";
		errorMessage += "Location: The member function of \"getStructuralIndex\" in the class of SolidStatePhysics::Crystallization::Analysis::StructuralDictionary.\n";
		errorMessage += "Message : Could not find the coresponding structural index.\n\n";
		errorMessage += "------------------------------------------------------\n\n";
		throw std::logic_error(errorMessage);
	}
}

inline SolidStatePhysics::Crystallization::Simulators::FinalSimulationsReporter& SolidStatePhysics::Crystallization::Analysis::StructuralDictionary::getFinalSimulationsReporter(const size_type structuralIndex) const
{
	if (m_needFinalSimulationsReportsPerStructure)
	{
		auto iter = m_finalSimulationsReporters.find(structuralIndex);

		if (iter != m_finalSimulationsReporters.end())
			return iter->second;

		else
		{
			std::string errorMessage = "------------------------------------------------------\n\n";
			errorMessage += "Error   : Logic error\n";
			errorMessage += "Location: The member function of \"getFinalSimulationsReporter\" in the class of SolidStatePhysics::Crystallization::Analysis::StructuralDictionary.\n";
			errorMessage += "Message : Could not find the coresponding final simulations reporter.\n\n";
			errorMessage += "------------------------------------------------------\n\n";
			throw std::logic_error(errorMessage);
		}
	}

	else
	{
		std::string errorMessage = "------------------------------------------------------\n\n";
		errorMessage += "Error   : Logic error\n";
		errorMessage += "Location: The member function of \"getFinalSimulationsReporter\" in the class of SolidStatePhysics::Crystallization::Analysis::StructuralDictionary.\n";
		errorMessage += "Message : Could not access it since this does not create the final simulation reports per structure.\n\n";
		errorMessage += "------------------------------------------------------\n\n";
		throw std::logic_error(errorMessage);
	}
}

inline bool SolidStatePhysics::Crystallization::Analysis::StructuralDictionary::contains(const std::string& structuralFingerprint) const noexcept
{
	auto iter = m_structuralIndices.find(structuralFingerprint);
	{
		if (iter != m_structuralIndices.end())
			return true;
		else
			return false;
	}
}

inline bool SolidStatePhysics::Crystallization::Analysis::StructuralDictionary::contains(const size_type structuralIndex) const noexcept
{
	auto iter = m_atomicEnvironmentsNames.find(structuralIndex);
	{
		if (iter != m_atomicEnvironmentsNames.end())
			return true;
		else
			return false;
	}
}

inline void SolidStatePhysics::Crystallization::Analysis::StructuralDictionary::addCrystallizedSamplesCount() noexcept
{
	std::lock_guard<std::mutex> guard{ m_dataMutex };
	++m_crystallizedSamplesCounts;
}

inline void SolidStatePhysics::Crystallization::Analysis::StructuralDictionary::addFailedSamplesCount() noexcept
{
	std::lock_guard<std::mutex> guard{ m_dataMutex };
	++m_failedSamplesCounts;
}

inline void SolidStatePhysics::Crystallization::Analysis::StructuralDictionary::addExceptionSamplesCount() noexcept
{
	std::lock_guard<std::mutex> guard{ m_dataMutex };
	++m_exceptionSamplesCounts;
}

inline void SolidStatePhysics::Crystallization::Analysis::StructuralDictionary::resetCrystallizedSamplesCount(const size_type crystallizedSamples)
{
	std::lock_guard<std::mutex> guard{ m_dataMutex };
	m_crystallizedSamplesCounts = crystallizedSamples;
}

inline void SolidStatePhysics::Crystallization::Analysis::StructuralDictionary::registerCrystalStructure(const std::string& structuralFingerprint, const std::map<std::string, size_type>& atomicEnvironments)
{
	std::lock_guard<std::mutex> guard{ m_dataMutex };

	size_type structuralIndex = m_structuralIndices.size();
	m_structuralIndices.emplace(structuralFingerprint, structuralIndex);

	size_type atomicEnvironmentsCount = 0;
	{
		for (const auto& symbolAndNum : atomicEnvironments)
			atomicEnvironmentsCount += symbolAndNum.second;
	}
	registerAtomicEnvironmentsCount(structuralIndex, atomicEnvironmentsCount);
	registerAtomicEnvironmentsName(structuralIndex, atomicEnvironments);


	if (m_needFinalSimulationsReportsPerStructure)
	{
		std::filesystem::path reportDirectoryPath{ m_chemicalConstituentsDirectoryPath };
		{
			std::string cationicEnvironmentsCountDirectoryName;
			{
				cationicEnvironmentsCountDirectoryName += std::to_string(atomicEnvironmentsCount);
				cationicEnvironmentsCountDirectoryName += "_AtomicEnvironments";
			}
			std::string structureName{ "Structure_" };
			structureName += std::to_string(structuralIndex);

			reportDirectoryPath /= cationicEnvironmentsCountDirectoryName;
			reportDirectoryPath /= getAtomicEnvironmentsName(atomicEnvironments);
			reportDirectoryPath /= structureName;
		}


		m_finalSimulationsReporters.emplace(structuralIndex, FinalSimulationsReporter{ reportDirectoryPath, m_maxGeometricalOptimizing });
	}
}

// Access methods
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// Static methods

inline bool SolidStatePhysics::Crystallization::Analysis::StructuralDictionary::needFinalSimulationsReportsPerStructure() noexcept
{
	return m_needFinalSimulationsReportsPerStructure;
}

inline void SolidStatePhysics::Crystallization::Analysis::StructuralDictionary::resetFinalSimulationsReportsPerStructureNecessity(const bool needFinalSimulationsReportsPerStructure)
{
	std::lock_guard<std::mutex> guard{ m_dataMutex };
	m_needFinalSimulationsReportsPerStructure = needFinalSimulationsReportsPerStructure;
}

// Static methods
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************

#endif // !SOLIDSTATEPHYSICS_CRYSTALLIZATION_ANALYSIS_STRUCTURALDICTIONARY_H
