#include "Crystallizer.h"

#include "DatFileOwner.h"
#include "CifFileOwner.h"

#include "GeometricalOptimizer.h"

using namespace SolidStatePhysics::Crystallization::Simulators;

// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// Constructor

Crystallizer::Crystallizer(const CrystallizationPlan& crystallizationPlan, const MonitoringPlan& monitoringPlan)
	: m_crystallizationPlan{ crystallizationPlan }
	, m_monitoringPlan{ monitoringPlan }
	, m_elapsedTime{ 0 }
	, m_localOptimizationCount{ 0 }
	, m_remodellingCount{ 0 }
	, m_workDirectoryPath{}
	, m_chemicalBondingRecorder{}
{
}

Crystallizer::Crystallizer(CrystallizationPlan&& crystallizationPlan, MonitoringPlan&& monitoringPlan)
	: m_crystallizationPlan{ std::move(crystallizationPlan) }
	, m_monitoringPlan{ std::move(monitoringPlan) }
	, m_elapsedTime{ 0 }
	, m_localOptimizationCount{ 0 }
	, m_remodellingCount{ 0 }
	, m_workDirectoryPath{}
	, m_chemicalBondingRecorder{}
{
}

// Constructor
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// Public methods

void Crystallizer::run(IntelligentLiquid& intelligentLiquid)
{
	runGeometricalOptimizers(intelligentLiquid);
	overCreateDatFile(intelligentLiquid);
	overCreateCifFile(intelligentLiquid);
}

// Public methods
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// Private methods

void Crystallizer::runGeometricalOptimizers(IntelligentLiquid& intelligentLiquid)
{
	intelligentLiquid.renewIdealChemicalBonding();
	intelligentLiquid.resetMaxBondingLengthErrorRate(m_crystallizationPlan.maxBondingLengthErrorRate());
	intelligentLiquid.resetMinPackingFraction(m_crystallizationPlan.minPackingFraction());

	GeometricalOptimizer globalOptimizer{ m_crystallizationPlan.maxGeometricalOptimizing(), m_crystallizationPlan.globalOptimizationVariables(), m_monitoringPlan };
	GeometricalOptimizer localOptimizer{ m_crystallizationPlan.maxGeometricalOptimizing(), m_crystallizationPlan.localOptimizationVariables(), m_monitoringPlan };
	globalOptimizer.initialize(intelligentLiquid, m_workDirectoryPath);


	while (globalOptimizer.canOptimize())
	{
		globalOptimizer.optimize(intelligentLiquid, m_elapsedTime, m_localOptimizationCount, m_remodellingCount);

		localOptimizer.initialize(intelligentLiquid, m_workDirectoryPath);
		localOptimizer.optimize(intelligentLiquid, m_elapsedTime, m_localOptimizationCount, m_remodellingCount);
		localOptimizer.finalize();
		++m_localOptimizationCount;


		if (globalOptimizer.canOptimize())
		{
			if (intelligentLiquid.isDenseCrystalStructure())
			{
				if (intelligentLiquid.hasNonbondingAtoms())
				{
					globalOptimizer.anneal(intelligentLiquid, m_elapsedTime);
					continue;
				}

				else
				{
					intelligentLiquid.renewGeometricalConstrainters();

					if (intelligentLiquid.isDefinitelyCrystallized())
						break;

					else
					{
						globalOptimizer.anneal(intelligentLiquid, m_elapsedTime);
						continue;
					}
				}
			}

			else
				break;
		}
	}

	globalOptimizer.finalize();
}

void Crystallizer::overCreateDatFile(const IntelligentLiquid& intelligentLiquid)
{
	std::filesystem::path datFilePath = m_workDirectoryPath;
	{
		std::string filename = intelligentLiquid.name();
		filename += ".dat";
		datFilePath /= filename;
	}

	SolidStatePhysics::Materials::IO::DatFileOwner datFileOwner(datFilePath);
	datFileOwner.overCreateFile(intelligentLiquid);
}

void Crystallizer::overCreateCifFile(const IntelligentLiquid& intelligentLiquid)
{
	std::filesystem::path cifFilePath = m_workDirectoryPath;
	{
		std::string filename = intelligentLiquid.name();
		filename += ".cif";
		cifFilePath /= filename;
	}

	SolidStatePhysics::Materials::IO::CifFileOwner cifFileOwner(cifFilePath);
	cifFileOwner.overCreateFile(intelligentLiquid);
}

void Crystallizer::overCreateMdFile(const IntelligentLiquid& intelligentLiquid)
{
	std::filesystem::path mdFilePath = m_workDirectoryPath;
	{
		std::string filename = intelligentLiquid.name();
		filename += ".md";
		mdFilePath /= filename;
	}

	SolidStatePhysics::Materials::IO::MdFileOwner mdFileOwner(mdFilePath);
	mdFileOwner.tryCreateEmptyFile();
}

void Crystallizer::linkChemicalBondingRecorder()
{
	std::filesystem::path chemicalBondingRecorderPath = m_workDirectoryPath;
	chemicalBondingRecorderPath /= "chemicalBondingRecord.txt";

	m_chemicalBondingRecorder = ChemicalBondingRecorder{ chemicalBondingRecorderPath };

	if (!(m_chemicalBondingRecorder.isValid()))
		m_chemicalBondingRecorder.tryCreateEmptyFile();
}

// Private methods
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
