#include "GeometricalOptimizer.h"

using namespace SolidStatePhysics::Crystallization::Simulators;


// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// Constructor

GeometricalOptimizer::GeometricalOptimizer(const size_type maxGeometricalOptimizing, const CrystallizationVariables& crystallizationVariables, const MonitoringPlan& monitoringPlan)
	: m_maxGeometricalOptimizing{ maxGeometricalOptimizing }
	, m_crystallizationVariables{ crystallizationVariables }
	, m_monitoringPlan{ monitoringPlan }
{
}

// Constructor
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// Public methods

void GeometricalOptimizer::initialize(const IntelligentLiquid& intelligentLiquid, const std::filesystem::path& workDirectoryPath) const
{
	m_shouldKeepRunning = true;
	m_time = 0;
	m_elapsedTime = 0;
	m_localOptimizationCount = 0;
	m_remodellingCount = 0;
	m_geometricalConstraintsTimer = 0;
	m_linkedPolyhedraTimer = 0;
	m_crystalLatticeTimer = 0;
	m_thermalTransformationTimer = 0;
	m_mdRecordTimer = 0;
	m_chemicalBondingRecordTimer = 0;
	m_workDirectoryPath = workDirectoryPath;
	{
		if (!(std::filesystem::is_directory(m_workDirectoryPath)))
		{
			finalize();

			std::string errorMessage = "------------------------------------------------------\n\n";
			errorMessage += "Error   : Invalid argument\n";
			errorMessage += "Location: The member function of \"initialize\" in the class of SolidStatePhysics::Crystallization::UnitOperations::UnitOperator.\n";
			errorMessage += "Message : \"m_workDirectoryPath\" is not a path to a directory.\n\n";
			errorMessage += "------------------------------------------------------\n\n";
			throw std::invalid_argument(errorMessage);
		}
	}

	if (m_monitoringPlan.needMdRecord())
		linkMdFile(intelligentLiquid);

	if (m_monitoringPlan.needChemicalBondingRecord())
		linkChemicalBondingRecorder();
}

void GeometricalOptimizer::finalize() const noexcept
{
	m_shouldKeepRunning = false;
	m_time = 0;
	m_elapsedTime = 0;
	m_localOptimizationCount = 0;
	m_remodellingCount = 0;
	m_geometricalConstraintsTimer = 0;
	m_linkedPolyhedraTimer = 0;
	m_crystalLatticeTimer = 0;
	m_thermalTransformationTimer = 0;
	m_mdRecordTimer = 0;
	m_chemicalBondingRecordTimer = 0;
	m_workDirectoryPath.clear();
	m_mdFileOwner.releaseFile();
	m_chemicalBondingRecorder.releaseFile();
}

void GeometricalOptimizer::anneal(IntelligentLiquid& intelligentLiquid, size_type& initialElapsedTime, const double maxDisplacement, size_type necessaryAnnealing) const
{
	m_shouldKeepRunning = true;
	m_elapsedTime = initialElapsedTime;
	m_chemicalBondingRecorder.resetCurrentTimeCounts(initialElapsedTime);
	{
		if(maxDisplacement < 0.0)
			intelligentLiquid.dynamicsController().resetMaxDisplacement(m_crystallizationVariables.initialMaxDisplacement());
		else
			intelligentLiquid.dynamicsController().resetMaxDisplacement(maxDisplacement);
	}
	intelligentLiquid.dynamicsController().resetCoolingRate(1.0);
	intelligentLiquid.dynamicsController().resetMaxLatticeTransformation(intelligentLiquid.atomicArrangement());
	{
		if(necessaryAnnealing == 0)
			necessaryAnnealing = 5 * m_crystallizationVariables.geometricalConstraintersLifetime();
	}

	initializeGeometricalConstraintersIfNecessary(intelligentLiquid);


	for (size_type annealingIndex = 0; annealingIndex < necessaryAnnealing; ++annealingIndex)
	{
		applyGeometricalOptimization(intelligentLiquid);
		manageTime();

		if (!(manageGeometricalConstrainters(intelligentLiquid)))
			manageLinkedPolyhedra(intelligentLiquid);

		manageCrystalLattice(intelligentLiquid);
		manageThermalTransformation(intelligentLiquid);
		manageMdRecorder(intelligentLiquid);
		manageChemicalBondingRecorder(intelligentLiquid);

		if (!m_shouldKeepRunning)
			break;
	}

	initialElapsedTime = m_elapsedTime;

	if (m_monitoringPlan.needChemicalBondingRecord())
		m_chemicalBondingRecorder.finishRecording();
}

void GeometricalOptimizer::optimize(IntelligentLiquid& intelligentLiquid, size_type& initialElapsedTime, size_type& initialLocalOptimizationCount, size_type& remodellingCount) const
{
	m_shouldKeepRunning = true;
	m_elapsedTime = initialElapsedTime;
	m_chemicalBondingRecorder.resetCurrentTimeCounts(initialElapsedTime);
	m_localOptimizationCount = initialLocalOptimizationCount;
	m_remodellingCount = remodellingCount;

	intelligentLiquid.dynamicsController().resetMaxDisplacement(m_crystallizationVariables.initialMaxDisplacement());
	intelligentLiquid.dynamicsController().resetCoolingRate(std::pow((m_crystallizationVariables.finalMaxDisplacement() / m_crystallizationVariables.initialMaxDisplacement()), (1.0 / static_cast<double>(m_crystallizationVariables.necessaryGeometricalOptimizing()))));
	intelligentLiquid.dynamicsController().resetMaxLatticeTransformation(intelligentLiquid.atomicArrangement());

	initializeGeometricalConstraintersIfNecessary(intelligentLiquid);


	while (m_shouldKeepRunning)
	{
		applyGeometricalOptimization(intelligentLiquid);
		manageTime();

		if (!(manageGeometricalConstrainters(intelligentLiquid)))
		{
			if (manageLinkedPolyhedra(intelligentLiquid))
			{
				if (intelligentLiquid.isProbablyCrystallized())
					m_shouldKeepRunning = false;
			}
		}

		manageCrystalLattice(intelligentLiquid);
		manageThermalTransformation(intelligentLiquid);
		manageMdRecorder(intelligentLiquid);
		manageChemicalBondingRecorder(intelligentLiquid);
	}

	initialElapsedTime = m_elapsedTime;
	initialLocalOptimizationCount = m_localOptimizationCount;
	remodellingCount = m_remodellingCount;

	if (m_monitoringPlan.needChemicalBondingRecord())
		m_chemicalBondingRecorder.finishRecording();
}

// Public methods
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// Private methods

void GeometricalOptimizer::linkMdFile(const IntelligentLiquid& intelligentLiquid) const
{
	std::filesystem::path mdFilePath = m_workDirectoryPath;
	{
		std::string filename = intelligentLiquid.name();
		filename += ".md";
		mdFilePath /= filename;
	}

	m_mdFileOwner = SolidStatePhysics::Materials::IO::MdFileOwner{ mdFilePath };

	if (!(m_mdFileOwner.isValid()))
		m_mdFileOwner.tryCreateEmptyFile();
}

void GeometricalOptimizer::linkChemicalBondingRecorder() const
{
	std::filesystem::path chemicalBondingRecorderPath = m_workDirectoryPath;
	chemicalBondingRecorderPath /= "chemicalBondingRecord.txt";

	m_chemicalBondingRecorder = ChemicalBondingRecorder{ chemicalBondingRecorderPath };

	if (!(m_chemicalBondingRecorder.isValid()))
		m_chemicalBondingRecorder.tryCreateEmptyFile();
}

// Private methods
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
