#include "IntelligentLiquid.h"

#include <set>
#include <limits>
#include <utility>
#include <cmath>

#include "AbstractChemicalElement.h"

using namespace SolidStatePhysics::Crystallization::Components;

// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// Constructors

IntelligentLiquid::IntelligentLiquid()
	: PeriodicLiquid{}
	, m_idealChemicalBonding{ 0 }
	, m_dynamicsController{}
	, m_maxBondingLengthErrorRate{ 0.0 }
	, m_minPackingFraction{ 0.0 }
{
}

IntelligentLiquid::IntelligentLiquid(const CrystalLattice& crystalLattice, const AtomicArrangement& atomicArrangement, const std::string& name)
	: PeriodicLiquid{ crystalLattice, atomicArrangement, name }
	, m_idealChemicalBonding{ 0 }
	, m_dynamicsController{}
	, m_maxBondingLengthErrorRate{ 0.0 }
	, m_minPackingFraction{ 0.0 }
{
}

IntelligentLiquid::IntelligentLiquid(const CrystalLattice& crystalLattice, AtomicArrangement&& atomicArrangement, const std::string& name)
	: PeriodicLiquid{ crystalLattice, std::move(atomicArrangement), name }
	, m_idealChemicalBonding{ 0 }
	, m_dynamicsController{}
	, m_maxBondingLengthErrorRate{ 0.0 }
	, m_minPackingFraction{ 0.0 }
{
}

// Constructors
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// Utility

IntelligentLiquid::size_type IntelligentLiquid::countAtomicEnvironments() const noexcept
{
	std::set<std::string> fingerprints;
	{
		for (const auto& atom : atomicArrangement())
		{
			std::string thisFingerprint = atom.getFingerprint(m_maxBondingLengthErrorRate);
			auto iter = fingerprints.find(thisFingerprint);

			if (iter == fingerprints.end())
				fingerprints.emplace(thisFingerprint);
		}
	}

	return fingerprints.size();
}

std::map<std::string, IntelligentLiquid::size_type> IntelligentLiquid::getAtomicEnvironments() const noexcept
{
	std::map<size_type, std::set<std::string>> fingerprintsPerElement;
	{
		for (const auto& atom : atomicArrangement())
		{
			std::map<size_type, std::set<std::string>>::iterator elementIter = fingerprintsPerElement.find(atom.atomicNumber());
			std::string thisFingerprint = atom.getFingerprint(m_maxBondingLengthErrorRate);

			if (elementIter == fingerprintsPerElement.end())
			{
				std::set<std::string> vec{ thisFingerprint };
				fingerprintsPerElement.emplace(atom.atomicNumber(), std::move(vec));
			}

			else
			{
				auto fingerprintIter = elementIter->second.find(thisFingerprint);

				if (fingerprintIter == elementIter->second.end())
					elementIter->second.emplace(thisFingerprint);
			}
		}
	}

	std::map<std::string, size_type> atomicEnvironmentsPerElement;
	{
		for (const auto& fingerprints : fingerprintsPerElement)
		{
			using AbstractChemicalElement = SolidStatePhysics::Crystallization::Calculators::AbstractChemicalElement;
			atomicEnvironmentsPerElement.emplace(AbstractChemicalElement::getElementSymbol(fingerprints.first), fingerprints.second.size());
		}
	}

	return atomicEnvironmentsPerElement;
}

std::string IntelligentLiquid::writeFingerprint() const noexcept
{
	std::vector<std::pair<const IntelligentAtom*, std::string>> m_atomPtrs;
	{
		for (const auto& atom : atomicArrangement())
		{
			std::string fingerPrint = atom.getFingerprint(m_maxBondingLengthErrorRate);
			m_atomPtrs.push_back(std::make_pair(&atom, fingerPrint));
		}

		std::sort(m_atomPtrs.begin(), m_atomPtrs.end(), [](const std::pair<const IntelligentAtom*, std::string>& iaa, const std::pair<const IntelligentAtom*, std::string>& iab) { return (iaa.second < iab.second); });
	}

	std::string fingerprintString;
	{
		fingerprintString += "NOA_";
		fingerprintString += std::to_string(atomicArrangement().size());
		fingerprintString += "___";

		for (auto atomPtr : m_atomPtrs)
		{
			fingerprintString += atomPtr.second;
			fingerprintString += "___";
		}

		fingerprintString.pop_back();
		fingerprintString.pop_back();
		fingerprintString.pop_back();
	}

	return fingerprintString;
}

// Utility
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// Private methods

void IntelligentLiquid::createBasicConstrainters()
{
	for (auto iter = atomicArrangement().begin(); iter != atomicArrangement().end(); ++iter)
	{
		for (auto anotherIter = std::next(iter); anotherIter != atomicArrangement().end(); ++anotherIter)
		{
			for (const auto& latticePoint : getVisibleLatticePoints(*iter, *anotherIter))
			{
				AtomicAdaptor attachedAtom{ *anotherIter, crystalLattice(), latticePoint };

				if (iter->isCation() && anotherIter->isCation())
				{
					if (MetallicConstrainter::shouldCreate(*iter, attachedAtom))
					{
						if (!(iter->hasMetallicConstrainter(attachedAtom)))
							iter->createMetallicConstrainter(attachedAtom, *this);
					}
				}

				else
				{
					if (IonicConstrainter::shouldCreate(*iter, attachedAtom))
					{
						if (!(iter->hasIonicConstrainter(attachedAtom)))
							iter->createIonicConstrainter(attachedAtom, *this);
					}
				}
			}
		}

		for (const auto& latticePoint : getVisibleLatticePoints(*iter, *iter))
		{
			if (latticePoint[0] != 0 || latticePoint[1] != 0 || latticePoint[2] != 0)
			{
				AtomicAdaptor attachedAtom{ *iter, crystalLattice(), latticePoint };

				if (iter->isCation())
				{
					if (MetallicConstrainter::shouldCreate(*iter, attachedAtom))
					{
						if (!(iter->hasMetallicConstrainter(attachedAtom)))
							iter->createMetallicConstrainter(attachedAtom, *this);
					}
				}

				else
				{
					if (IonicConstrainter::shouldCreate(*iter, attachedAtom))
					{
						if (!(iter->hasIonicConstrainter(attachedAtom)))
							iter->createIonicConstrainter(attachedAtom, *this);
					}
				}
			}
		}
	}
}

void IntelligentLiquid::createChemicalBonds()
{
	std::vector<IntelligentAtom*> m_atomPtrs;
	{
		for (auto& atom : atomicArrangement())
		{
			if (atom.isCation())
				m_atomPtrs.push_back(&atom);
		}
	}

	while (!(m_atomPtrs.empty()))
	{
		std::sort(m_atomPtrs.begin(), m_atomPtrs.end(), [](const IntelligentAtom* iaa, const IntelligentAtom* iab) { return (iaa->bondingMotivation().second < iab->bondingMotivation().second); });

		if (!(m_atomPtrs.back()->configureChemicalBonds(*this)))
		{
			m_atomPtrs.pop_back();
		}
	}
}

void IntelligentLiquid::createSecondNeighborConstrainters()
{
	for (auto iter = m_ionicConstraintersPtrs.begin(); iter != m_ionicConstraintersPtrs.end();)
	{
		if (std::shared_ptr<IonicConstrainter> ionicConstrainterPtr = iter->lock())
		{
			if (ChemicalBond::shouldCreate(std::static_pointer_cast<SolidStatePhysics::Crystallization::Calculators::InteratomicConnector>(ionicConstrainterPtr)))
			{
				if (!(ionicConstrainterPtr->ownerAtom().hasSecondNeighborConstrainter(ionicConstrainterPtr->attachedAtom())))
				{
					ionicConstrainterPtr->ownerAtom().createSecondNeighborConstrainter(ionicConstrainterPtr->attachedAtom(), *this);
					IonicConstrainter::erase(std::move(ionicConstrainterPtr));
				}
			}

			++iter;
		}

		else
			iter = m_ionicConstraintersPtrs.erase(iter);
	}
}

void IntelligentLiquid::createPolyhedralConstrainters()
{
	for (auto& intelligentAtom : atomicArrangement())
		intelligentAtom.configurePolyhedralConstrainters(*this);

	for (auto iter = m_polyhedralConstraintersPtrs.begin(); iter != m_polyhedralConstraintersPtrs.end();)
	{
		if (std::shared_ptr<PolyhedralConstrainter> polyhedralConstrainterPtr = iter->lock())
		{
			polyhedralConstrainterPtr->removeExcessCommonBridgingAtoms(*this);
			++iter;
		}

		else
			iter = m_polyhedralConstraintersPtrs.erase(iter);
	}
}

// Private methods
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
