#include "MATEVENT.h"

#include <stdexcept>
#include <string>
#include <iostream>
#include <sstream>

#include "Receptionist.h"

#include "LogWriter.h"
#include "SystemPathsRepository.h"
#include "ThreadController.h"
#include "MpiController.h"

#include "MaterialsExplorer.h"

using namespace ControlCenter::Window;

// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// Constructor

MATEVENT::MATEVENT(int argc, char* argv[])
	: m_mpiNumber{ ControlCenter::System::MpiController::thisMpiNumber() }
	, m_orderSheetPath{}
	, m_demandCustomerSupport{ false }
	, m_clock{}
{
	ControlCenter::System::MpiController::resetThisMpiNumber(0);
	ControlCenter::System::MpiController::resetTotalMpiCount(1);

	for (int i = 1; i < argc; ++i)
	{
		std::string argvString(argv[i]);

		if (argvString == "-osp")
		{
			if (argc <= i + 1)
				break;

			m_orderSheetPath = std::filesystem::path{ argv[++i] };
		}

		else if (argvString == "-mpin")
		{
			if (argc <= i + 1)
				break;

			else
			{
				int thisMpiNumber = 0;
				{
					std::stringstream thisMpiNumberStream{ argv[++i] };
					thisMpiNumberStream >> thisMpiNumber;
				}

				ControlCenter::System::MpiController::resetThisMpiNumber(static_cast<std::size_t>(thisMpiNumber));
			}
		}

		else if (argvString == "-node")
		{
			if (argc <= i + 1)
				break;

			else
			{
				int totalMpiCount = 0;
				{
					std::stringstream totalMpiCountStream{ argv[++i] };
					totalMpiCountStream >> totalMpiCount;
				}

				ControlCenter::System::MpiController::resetTotalMpiCount(static_cast<std::size_t>(totalMpiCount));
			}
		}

		else if (argvString == "--help")
			m_demandCustomerSupport = true;

		else
			continue;
	}


	if (!(ControlCenter::System::MpiController::hasValidMpiConstants()))
	{
		std::string errorMessage = "------------------------------------------------------\n\n";
		errorMessage += "Error   : Runtime error\n";
		errorMessage += "Location: The constructor of the class of ControlCenter::Window::MATEVENT.\n";
		errorMessage += "Message : Could not read the valid index of the MPI and/or the valid number of the nodes.\n\n";
		errorMessage += "------------------------------------------------------\n\n";
		throw std::runtime_error(errorMessage);
	}
}

// Constructor
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// Public method

int MATEVENT::execute()
{
	if (needCustomerSupport())
		executeCustomerSupport();

	else
	{
		try
		{
			OrderForm orderForm = readOrderForm();
			setEnvironmentVariables(orderForm);
			
			if (orderForm.shouldGenerateMaterials())
			{
				SolidStatePhysics::Crystallization::Simulators::CrystallizationPlan crystallizationPlan = orderForm.crystallizationPlan();
				SolidStatePhysics::Crystallization::Simulators::MonitoringPlan monitoringPlan = orderForm.monitoringPlan();
				ControlCenter::Administering::RequestedFeedstocks requestedFeedstocks = orderForm.requestedFeedstocks();
				orderForm.clear();
			
				ControlCenter::Administering::MaterialsExplorer materialsExplorer{ std::move(crystallizationPlan), std::move(monitoringPlan) };
				materialsExplorer.executePeriodicSolidsGeneration(std::move(requestedFeedstocks));
			}

			else if (orderForm.shouldReheatMaterials())
			{
				SolidStatePhysics::Crystallization::Simulators::CrystallizationPlan crystallizationPlan = orderForm.crystallizationPlan();
				SolidStatePhysics::Crystallization::Simulators::MonitoringPlan monitoringPlan = orderForm.monitoringPlan();
				orderForm.clear();

				ControlCenter::Administering::MaterialsExplorer materialsExplorer{ std::move(crystallizationPlan), std::move(monitoringPlan) };
				materialsExplorer.executePeriodicSolidsReheating();
			}

			else
			{
				std::string errorMessage = "------------------------------------------------------\n\n";
				errorMessage += "Error   : Runtime error\n";
				errorMessage += "Location: The member function of \"execute\" in the class of ControlCenter::Window::MATEVENT.\n";
				errorMessage += "Message : \"Execution.Mode\" is invalid.\n\n";
				errorMessage += "------------------------------------------------------\n\n";
				throw std::runtime_error(errorMessage);
			}
		}

		catch (const std::exception& e)
		{
			std::string errorMessage;
			{
				errorMessage += "Process_";
				errorMessage += std::to_string(ControlCenter::System::MpiController::thisMpiNumber());
				errorMessage += '\n';
				errorMessage += e.what();
			}

			ControlCenter::System::IO::LogWriter logWriter;
			logWriter.report(errorMessage);
			throw;
		}
	}

	return 0;
}

// Public method
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// Private methods

bool MATEVENT::needCustomerSupport() const
{
	if (m_demandCustomerSupport)
		return true;

	else
	{
		if (m_orderSheetPath.empty())
		{
			std::string stdoutMessage = "------------------------------------------------------\n\n";
			stdoutMessage += "Error   : Runtime error\n";
			stdoutMessage += "Location: The member function of \"needCustomerSupport\" in the class of ControlCenter::Window::MATEVENT.\n";
			stdoutMessage += "Message : \"m_orderSheetPath\" is empty. --> This function will send the manual for the command-line arguments.\n\n";
			stdoutMessage += "------------------------------------------------------\n\n";
			std::cout << stdoutMessage << std::flush;

			return true;
		}

		else
			return false;
	}
}

void MATEVENT::executeCustomerSupport() const
{
	Receptionist receptionist;
	receptionist.sendCommandLineArgumentsManual();
}

void MATEVENT::setEnvironmentVariables(const OrderForm& orderForm) const
{
	std::string orderName = orderForm.orderName();
	ControlCenter::System::ThreadController::resetMaxThreading(orderForm.maxThreading());
	ControlCenter::System::ThreadController::resetMklThreading(orderForm.mklThreading());

	ControlCenter::System::IO::SystemPathsRepository::resetOutputDirectoryPath(orderForm.outputDirectoryPath());
	ControlCenter::System::IO::SystemPathsRepository::resetWorkDirectoryPath(orderForm.workDirectoryPath());
	ControlCenter::System::IO::SystemPathsRepository::tryInitializeOutputDirectory();
	ControlCenter::System::IO::SystemPathsRepository::tryInitializeWorkDirectory();

	std::filesystem::path logFilePath = orderForm.outputDirectoryPath();
	logFilePath /= "log.txt";
	ControlCenter::System::IO::LogWriter logWriter(logFilePath);
}

OrderForm MATEVENT::readOrderForm() const
{
	try
	{
		Receptionist receptionist;
		return receptionist.createOrderForm(m_orderSheetPath);
	}

	catch (const std::exception& e)
	{
		std::string errorMessage = e.what();
		errorMessage += "------------------------------------------------------\n\n";
		errorMessage += "Error   : Exception\n";
		errorMessage += "Location: The member function of \"readOrderForm\" in the class of ControlCenter::Window::MATEVENT.\n";
		errorMessage += "Message : This function catches the logic_error. There are problems in the order sheet.\n\n";
		errorMessage += "------------------------------------------------------\n\n";

		std::cout << errorMessage << std::endl;
		throw std::logic_error(errorMessage);
	}
}

void MATEVENT::reportElapsedTime()
{
	std::string reportMessage = "Elapsed time  :  ";
	reportMessage += m_clock.tellElapsedTime();

	ControlCenter::System::IO::LogWriter logWriter;
	logWriter.report(reportMessage);

	m_clock.terminateStopWatch();
}

// Private method
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
