#include "MaterialsExplorer.h"

#include <thread>
#include <regex>

#include "Clock.h"
#include "ThreadController.h"
#include "MpiController.h"
#include "SystemPathsRepository.h"
#include "DirectoryManipulator.h"
#include "FileOwner.h"

#include "StructuralDictionary.h"
#include "FinalSimulationsReporter.h"

#include "MinimumConstrainter.h"
#include "MinMaxConstrainter.h"
#include "PeriodicLiquid.h"

#include "GeneratePeriodicSolids.h"
#include "ReheatPeriodicSolids.h"

#include "PolyhedralConstrainter.h"

using namespace ControlCenter::Administering;

// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// Constructors

MaterialsExplorer::MaterialsExplorer(const CrystallizationPlan& crystallizationPlan, const MonitoringPlan& monitoringPlan)
	: m_crystallizationPlan{ crystallizationPlan }
	, m_monitoringPlan{ monitoringPlan }
{
}

MaterialsExplorer::MaterialsExplorer(CrystallizationPlan&& crystallizationPlan, MonitoringPlan&& monitoringPlan)
	: m_crystallizationPlan{ std::move(crystallizationPlan) }
	, m_monitoringPlan{ std::move(monitoringPlan) }
{
}

// Constructors
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// Private methods

void MaterialsExplorer::initializeGeometricalConstraints()
{
	SolidStatePhysics::Crystallization::Components::PolyhedralConstrainter::resetVertexSharingCationicPairs(m_crystallizationPlan.vertexSharingCationicPairs());
	SolidStatePhysics::Crystallization::Components::PolyhedralConstrainter::resetEdgeSharingCationicPairs(m_crystallizationPlan.edgeSharingCationicPairs());
	SolidStatePhysics::Crystallization::Components::PolyhedralConstrainter::resetFaceSharingCationicPairs(m_crystallizationPlan.faceSharingCationicPairs());
}

// Private methods
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// Private methods for generating

void MaterialsExplorer::operatePeriodicSolidsGenerator()
{
	SolidStatePhysics::Crystallization::Calculators::MinimumConstrainter::resetRepulsiveForceConstant(m_crystallizationPlan.repulsiveForceConstant());
	SolidStatePhysics::Crystallization::Calculators::MinMaxConstrainter::resetRepulsiveForceConstant(m_crystallizationPlan.repulsiveForceConstant());
	SolidStatePhysics::Crystallization::Calculators::MinMaxConstrainter::resetAttractiveForceConstant(m_crystallizationPlan.attractiveForceConstant());


	for (const auto& chemicalConstituents : m_targetChemicalConstituents)
	{
		ControlCenter::System::Clock clock;
		clock.startStopWatch();

		tryInitializeChemicalSystemDirectory(chemicalConstituents.getChemicalSystemName());
		std::filesystem::path chemicalConstituentsDirectoryPath = m_chemicalSystemDirectoryPath;
		chemicalConstituentsDirectoryPath /= chemicalConstituents.name();
		ControlCenter::System::IO::DirectoryManipulator::overCreateDirectory(chemicalConstituentsDirectoryPath);

		SolidStatePhysics::Crystallization::Analysis::StructuralDictionary structuralDictionary{ chemicalConstituentsDirectoryPath, m_crystallizationPlan.maxGeometricalOptimizing() };
		SolidStatePhysics::Crystallization::Analysis::StructuralDictionary::resetFinalSimulationsReportsPerStructureNecessity(m_monitoringPlan.needFinalSimulationsReportsPerStructure());
		structuralDictionary.initialize();
		initializeFinalSimulationsReporter(chemicalConstituentsDirectoryPath);
		GeneratePeriodicSolids::resetSampleIndex();


		std::vector<GeneratePeriodicSolids> generatePeriodicSolids;
		{
			for (std::size_t rep = 0; rep < ControlCenter::System::ThreadController::maxThreading(); ++rep)
				generatePeriodicSolids.push_back(GeneratePeriodicSolids{ rep, m_crystallizationPlan, m_monitoringPlan });
		}

		std::vector<std::thread> operatingSystems;
		{
			for (std::size_t rep = 0; rep < ControlCenter::System::ThreadController::maxThreading(); ++rep)
				operatingSystems.push_back(std::thread{ generatePeriodicSolids[rep], chemicalConstituentsDirectoryPath, chemicalConstituents, m_feedstocksPerChemicalConstituents });
		}

		for (auto& operatingSystem : operatingSystems)
			operatingSystem.join();


		GeneratePeriodicSolids::reportResults(structuralDictionary, clock.tellElapsedTime());
		structuralDictionary.createFinalSimulationsReportsPerStructureIfNecessary();
		structuralDictionary.clear();

		clock.terminateStopWatch();
	}
}

void MaterialsExplorer::initializeFinalSimulationsReporter(const std::filesystem::path& chemicalConstituentsDirectoryPath)
{
	using SolidStatePhysics::Crystallization::Simulators::FinalSimulationsReporter;
	GeneratePeriodicSolids::resetFinalSimulationsReporter(FinalSimulationsReporter{ chemicalConstituentsDirectoryPath, m_crystallizationPlan.maxGeometricalOptimizing() });
}

void MaterialsExplorer::assignChemicalConstituentsToMpiProcess(const RequestedFeedstocks& requestedFeedstocks)
{
	const size_type thisMpiNumber = ControlCenter::System::MpiController::thisMpiNumber();
	const size_type totalMpiCount = ControlCenter::System::MpiController::totalMpiCount();

	for (size_type index = thisMpiNumber; index < requestedFeedstocks.targetChemicalConstituents().size(); index += totalMpiCount)
		m_targetChemicalConstituents.push_back(requestedFeedstocks.targetChemicalConstituents().at(index));
}

void MaterialsExplorer::tryInitializeChemicalSystemDirectory(const std::string& chemicalSystemName)
{
	m_chemicalSystemDirectoryPath = ControlCenter::System::IO::SystemPathsRepository::workDirectoryPath();
	m_chemicalSystemDirectoryPath /= chemicalSystemName;

	ControlCenter::System::IO::DirectoryManipulator::tryInitializeDirectory(m_chemicalSystemDirectoryPath);
}

// Private methods for generating
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// Private methods for reheating

void MaterialsExplorer::operatePeriodicSolidsReheater()
{
	std::vector<ReheatPeriodicSolids> reheatPeriodicSolids;
	{
		for (std::size_t rep = 0; rep < ControlCenter::System::ThreadController::maxThreading(); ++rep)
			reheatPeriodicSolids.push_back(ReheatPeriodicSolids{ rep, m_crystallizationPlan, m_monitoringPlan });
	}

	std::vector<std::thread> operatingSystems;
	{
		for (std::size_t rep = 0; rep < reheatPeriodicSolids.size(); ++rep)
			operatingSystems.push_back(std::thread{ reheatPeriodicSolids[rep] });
	}

	for (auto& operatingSystem : operatingSystems)
		operatingSystem.join();
}

void MaterialsExplorer::assignReheatingChemicalConstituentsDirectoriesToMpiProcess()
{
	std::vector<std::filesystem::path> allReheatingChemicalConstituentsDirectories;
	{
		std::regex parentDirectoryNamePat("([[:alpha:]]+[\\_]{1}[\\d]+[\\_]{1})*([[:alpha:]]+[\\_]{1}[\\d]+){1}");
		std::regex directoryNamePat("Sample[\\_]{1}([\\d]+){1}");

		for (const std::filesystem::directory_entry& directoryEntry : std::filesystem::recursive_directory_iterator(ControlCenter::System::IO::SystemPathsRepository::workDirectoryPath()))
		{
			std::filesystem::path directoryPath{ directoryEntry.path() };

			if (std::filesystem::is_directory(directoryPath))
			{
				std::smatch parentSm;
				std::string parentDirectoryName = getDirectoryName(directoryPath.parent_path());

				if (std::regex_match(parentDirectoryName, parentSm, parentDirectoryNamePat))
				{
					std::smatch directorySm;
					std::string directoryName = getDirectoryName(directoryPath);

					if (std::regex_match(directoryName, directorySm, directoryNamePat))
					{
						if (hasDatFile(directoryPath))
							allReheatingChemicalConstituentsDirectories.push_back(directoryPath);
					}
				}
			}
		}
	}

	m_reheatingChemicalConstituentsDirectories.clear();
	const size_type thisMpiNumber = ControlCenter::System::MpiController::thisMpiNumber();
	const size_type totalMpiCount = ControlCenter::System::MpiController::totalMpiCount();

	for (size_type index = thisMpiNumber; index < allReheatingChemicalConstituentsDirectories.size(); index += totalMpiCount)
		m_reheatingChemicalConstituentsDirectories.push_back(allReheatingChemicalConstituentsDirectories.at(index));
}

bool MaterialsExplorer::hasDatFile(const std::filesystem::path& workDirectoryPath)
{
	for (const std::filesystem::directory_entry& fileEntry : std::filesystem::directory_iterator(workDirectoryPath))
	{
		std::filesystem::path filePath = fileEntry.path();

		if (std::filesystem::is_regular_file(filePath))
		{
			if (filePath.extension().string() == ".dat")
				return true;
			else
				return false;
		}
	}

	return false;
}

void MaterialsExplorer::createReheatingChemicalConstituentsDirectoriesFile()
{
	ControlCenter::System::IO::FileOwner workDirectoryFile{ ReheatPeriodicSolids::getReheatingChemicalConstituentsDirectoriesFilePath() };
	workDirectoryFile.overCreateEmptyFile();

	std::string message;
	{
		for (const auto& directoryPath : m_reheatingChemicalConstituentsDirectories)
		{
			message += directoryPath.generic_string();
			message += "\n";
		}
	}

	workDirectoryFile.write(message);
}

void MaterialsExplorer::removeReheatingChemicalConstituentsDirectoriesFile()
{
	ControlCenter::System::IO::FileOwner workDirectoryFile{ ReheatPeriodicSolids::getReheatingChemicalConstituentsDirectoriesFilePath() };
	workDirectoryFile.remove();
}

// Private methods for reheating
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
