#include "ReheatPeriodicSolids.h"

#include <regex>

#include "ThreadController.h"
#include "MpiController.h"
#include "SystemPathsRepository.h"
#include "DirectoryManipulator.h"
#include "LogWriter.h"
#include "FileOwner.h"

#include "DatFileOwner.h"

#include "Crystallizer.h"
#include "IntelligentLiquid.h"

using namespace ControlCenter::Administering;

// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// Constructors

std::mutex ReheatPeriodicSolids::m_reheatMutex;

ReheatPeriodicSolids::ReheatPeriodicSolids(const size_type threadIndex)
	: m_threadIndex{ threadIndex }
	, m_crystallizationPlan{}
{
}

ReheatPeriodicSolids::ReheatPeriodicSolids(const size_type threadIndex, const CrystallizationPlan& crystallizationPlan, const MonitoringPlan& monitoringPlan)
	: m_threadIndex{ threadIndex }
	, m_crystallizationPlan{ crystallizationPlan }
	, m_monitoringPlan{ monitoringPlan }
{
}

ReheatPeriodicSolids::ReheatPeriodicSolids(const size_type threadIndex, CrystallizationPlan&& crystallizationPlan, MonitoringPlan&& monitoringPlan)
	: m_threadIndex{ threadIndex }
	, m_crystallizationPlan{ std::move(crystallizationPlan) }
	, m_monitoringPlan{ std::move(monitoringPlan) }
{
}

// Constructors
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// Public methods

void ReheatPeriodicSolids::operator()()
{
	while (true)
	{
		try
		{
			std::filesystem::path workDirectoryPath = assignWorkDirectory();

			if (workDirectoryPath.empty())
				break;
			else
				operateCrystallizer(workDirectoryPath);
		}

		catch (const std::exception& e)
		{
			std::string errorMessage;
			{
				errorMessage += "Thread_";
				errorMessage += std::to_string(m_threadIndex);
				errorMessage += " of ";
				errorMessage += "Process_";
				errorMessage += std::to_string(ControlCenter::System::MpiController::thisMpiNumber());
				errorMessage += '\n';
				errorMessage += e.what();
			}

			ControlCenter::System::IO::LogWriter logWriter;
			logWriter.report(errorMessage);
		}
	}
}

std::filesystem::path ReheatPeriodicSolids::getReheatingChemicalConstituentsDirectoriesFilePath() noexcept
{
	std::filesystem::path filePath = ControlCenter::System::IO::SystemPathsRepository::workDirectoryPath();
	{
		std::string filename = "workDirectories_";
		filename += std::to_string(ControlCenter::System::MpiController::thisMpiNumber());
		filename += ".txt";

		filePath /= filename;
	}

	return filePath;
}

// Public methods
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// Private methods

void ReheatPeriodicSolids::operateCrystallizer(const std::filesystem::path& workDirectoryPath)
{
	SolidStatePhysics::Crystallization::Components::IntelligentLiquid intelligentLiquid;
	{
		for (const std::filesystem::directory_entry& fileEntry : std::filesystem::directory_iterator(workDirectoryPath))
		{
			std::filesystem::path filePath = fileEntry.path();

			if (std::filesystem::is_regular_file(filePath))
			{
				if (filePath.extension().string() == ".dat")
				{
					SolidStatePhysics::Materials::IO::DatFileOwner datFileOwner{ filePath };
					datFileOwner.readDatFile(intelligentLiquid);
					break;
				}

				else
					continue;
			}
		}
	}

	initializeWorkDirectory(workDirectoryPath);

	SolidStatePhysics::Crystallization::Simulators::Crystallizer crystallizer{ m_crystallizationPlan, m_monitoringPlan };
	crystallizer.initialize(workDirectoryPath);
	crystallizer.run(intelligentLiquid);
	crystallizer.finalize();
}

std::filesystem::path ReheatPeriodicSolids::assignWorkDirectory()
{
	std::lock_guard<std::mutex> threadLock{ m_reheatMutex };

	std::filesystem::path workDirectoryPath;
	{
		ControlCenter::System::IO::FileOwner fileOwner{ getReheatingChemicalConstituentsDirectoriesFilePath() };

		workDirectoryPath = std::filesystem::path{ fileOwner.readLine() };
		fileOwner.clearLine();
	}

	return workDirectoryPath;
}

void ReheatPeriodicSolids::initializeWorkDirectory(const std::filesystem::path& workDirectoryPath)
{
	std::filesystem::path inputDirectoryPath = workDirectoryPath;
	inputDirectoryPath /= "Input";
	ControlCenter::System::IO::DirectoryManipulator::overCreateDirectory(inputDirectoryPath);
	ControlCenter::System::IO::DirectoryManipulator::overTransmitAllFiles(workDirectoryPath, inputDirectoryPath);
}

// Private methods
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
