#include "StructuralDictionary.h"

#include <iterator>

using namespace SolidStatePhysics::Crystallization::Analysis;

// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// Constructor

StructuralDictionary::size_type StructuralDictionary::m_crystallizedSamplesCounts{ 0 };
StructuralDictionary::size_type StructuralDictionary::m_failedSamplesCounts{ 0 };
StructuralDictionary::size_type StructuralDictionary::m_exceptionSamplesCounts{ 0 };
std::map<std::string, StructuralDictionary::size_type> StructuralDictionary::m_structuralIndices;
std::map<StructuralDictionary::size_type, SolidStatePhysics::Crystallization::Simulators::FinalSimulationsReporter> StructuralDictionary::m_finalSimulationsReporters;
std::map<StructuralDictionary::size_type, StructuralDictionary::size_type> StructuralDictionary::m_atomicEnvironmentsCounts;
std::map<StructuralDictionary::size_type, std::string> StructuralDictionary::m_atomicEnvironmentsNames;
bool StructuralDictionary::m_needFinalSimulationsReportsPerStructure{ false };
std::mutex StructuralDictionary::m_dataMutex;


StructuralDictionary::StructuralDictionary(const std::filesystem::path& chemicalConstituentsDirectoryPath, const size_type maxGeometricalOptimizing)
	: m_chemicalConstituentsDirectoryPath{ chemicalConstituentsDirectoryPath }
	, m_maxGeometricalOptimizing{ maxGeometricalOptimizing }
{
}

// Constructor
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// Public methods

void StructuralDictionary::initialize() noexcept
{
	std::lock_guard<std::mutex> guard{ m_dataMutex };
	m_crystallizedSamplesCounts = 0;
	m_failedSamplesCounts = 0;
	m_exceptionSamplesCounts = 0;
	m_structuralIndices.clear();
	m_finalSimulationsReporters.clear();
	m_atomicEnvironmentsCounts.clear();
	m_atomicEnvironmentsNames.clear();
}

void StructuralDictionary::clear() noexcept
{
	std::lock_guard<std::mutex> guard{ m_dataMutex };
	m_crystallizedSamplesCounts = 0;
	m_failedSamplesCounts = 0;
	m_exceptionSamplesCounts = 0;
	m_structuralIndices.clear();
	m_finalSimulationsReporters.clear();
	m_atomicEnvironmentsCounts.clear();
	m_atomicEnvironmentsNames.clear();
}

std::string StructuralDictionary::writeFinalReport() const noexcept
{
	std::string reportString;
	{
		reportString += "-----------------------------------------------------------------------------------------------------------------------\n";
		reportString += "-----------------------------------------------------------------------------------------------------------------------\n\n";
		reportString += "The number of crystallized samples :  ";
		reportString += std::to_string(m_crystallizedSamplesCounts);
		reportString += "\n";
		reportString += "The number of failed samples       :  ";
		reportString += std::to_string(m_failedSamplesCounts);
		reportString += "\n";
		reportString += "The number of eception     samples :  ";
		reportString += std::to_string(m_exceptionSamplesCounts);
		reportString += "\n";
		reportString += "The kinds of crystal structure     :  ";
		reportString += std::to_string(m_structuralIndices.size());
		reportString += "\n\n";
		reportString += "-----------------------------------------------------------------------------------------------------------------------\n";
		reportString += "-----------------------------------------------------------------------------------------------------------------------\n";

		const size_type maxAtomicEnvironmentsCount = getMaxAtomicEnvironmentsCount();
		{
			for (size_type atomicEnvironmentsCount = 1; atomicEnvironmentsCount <= maxAtomicEnvironmentsCount; ++atomicEnvironmentsCount)
			{
				std::set<size_type> structuralIndices = getStructuralIndices(atomicEnvironmentsCount);

				if (!(structuralIndices.empty()))
				{
					reportString += "#\n# ";
					reportString += std::to_string(atomicEnvironmentsCount);
					reportString += " atomic environments\n";
					reportString += "#\n\n";

					reportString += "The kinds of crystal structure     :  ";
					reportString += std::to_string(structuralIndices.size());
					reportString += "\n\n";

					std::map<std::string, size_type> atomicEnvironmentsNamesDistribution = countStructuralIndex(structuralIndices);
					{
						for (const auto& nameAndCount : atomicEnvironmentsNamesDistribution)
						{
							reportString += nameAndCount.first;
							reportString += " :  ";
							reportString += std::to_string(nameAndCount.second);
							reportString += "\n";
						}
					}

					reportString += "\n\n";
					reportString += "-----------------------------------------------------------------------------------------------------------------------\n";
					reportString += "-----------------------------------------------------------------------------------------------------------------------\n";
				}
			}
		}
	}

	return reportString;
}

void StructuralDictionary::createFinalSimulationsReportsPerStructureIfNecessary()
{
	if (m_needFinalSimulationsReportsPerStructure)
	{
		for (const auto& finalSimulationsReporter : m_finalSimulationsReporters)
			finalSimulationsReporter.second.createDistributionFiles();
	}
}

// Public methods
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// Private methods

std::string StructuralDictionary::getAtomicEnvironmentsName(const std::map<std::string, size_type>& cationicEnvironments)
{
	std::string atomicEnvironmentsName;
	{
		for (const auto& symbolAndNum : cationicEnvironments)
		{
			atomicEnvironmentsName += symbolAndNum.first;
			atomicEnvironmentsName += "-";
			atomicEnvironmentsName += std::to_string(symbolAndNum.second);
			atomicEnvironmentsName += "_";
		}

		atomicEnvironmentsName.pop_back();
	}

	return atomicEnvironmentsName;
}

void StructuralDictionary::registerAtomicEnvironmentsCount(const size_type structuralIndex, const size_type atomicEnvironmentsCount) noexcept
{
	auto iter = m_atomicEnvironmentsCounts.find(structuralIndex);

	if (iter == m_atomicEnvironmentsCounts.end())
		m_atomicEnvironmentsCounts.emplace(structuralIndex, atomicEnvironmentsCount);
}

void StructuralDictionary::registerAtomicEnvironmentsName(const size_type structuralIndex, const std::map<std::string, size_type>& atomicEnvironments) noexcept
{
	auto iter = m_atomicEnvironmentsNames.find(structuralIndex);

	if (iter == m_atomicEnvironmentsNames.end())
		m_atomicEnvironmentsNames.emplace(structuralIndex, getAtomicEnvironmentsName(atomicEnvironments));
}

StructuralDictionary::size_type StructuralDictionary::getMaxAtomicEnvironmentsCount() const noexcept
{
	size_type maxAtomicEnvironmentsCount = 0;
	{
		for (const auto indexAndNum : m_atomicEnvironmentsCounts)
		{
			if (maxAtomicEnvironmentsCount < indexAndNum.second)
				maxAtomicEnvironmentsCount = indexAndNum.second;
		}
	}

	return maxAtomicEnvironmentsCount;
}

std::set<StructuralDictionary::size_type> StructuralDictionary::getStructuralIndices(const size_type atomicEnvironmentsCount) const noexcept
{
	std::set<size_type> structuralIndices;
	{
		for (const auto& indexAndCount : m_atomicEnvironmentsCounts)
		{
			if (indexAndCount.second == atomicEnvironmentsCount)
				structuralIndices.emplace(indexAndCount.first);
		}
	}

	return structuralIndices;
}

std::map<std::string, StructuralDictionary::size_type> StructuralDictionary::countStructuralIndex(const std::set<size_type>& structuralIndices) const
{
	std::map<std::string, size_type> atomicEnvironmentsNamesDistribution;
	{
		try
		{
			for (const auto& structuralIndex : structuralIndices)
			{
				std::string atomicEnvironmentsName = m_atomicEnvironmentsNames.at(structuralIndex);

				auto iter = atomicEnvironmentsNamesDistribution.find(atomicEnvironmentsName);
				{
					if (iter == atomicEnvironmentsNamesDistribution.end())
						atomicEnvironmentsNamesDistribution.emplace(atomicEnvironmentsName, 1);
					else
						iter->second += 1;
				}
			}
		}

		catch (const std::exception& e)
		{
			std::string errorMessage = e.what();
			errorMessage += "------------------------------------------------------\n\n";
			errorMessage += "Error   : Logic error\n";
			errorMessage += "Location: The member function of \"countStructuralIndex\" in the class of SolidStatePhysics::Crystallization::Analysis::StructuralDictionary.\n";
			errorMessage += "Message : \"structuralIndex\" is invalid key for \"m_atomicEnvironmentsNames\".\n\n";
			errorMessage += "------------------------------------------------------\n\n";
			throw std::logic_error(errorMessage);
		}
	}

	return atomicEnvironmentsNamesDistribution;
}

// Private methods
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
